package com.deepclone.lw.beans.admin;


import java.net.InetAddress;
import java.util.List;
import java.util.Set;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import com.deepclone.lw.cmd.ObjectNameError;
import com.deepclone.lw.cmd.admin.AdminOverviewResponse;
import com.deepclone.lw.cmd.admin.adata.AdminOverview;
import com.deepclone.lw.cmd.admin.adata.Administrator;
import com.deepclone.lw.cmd.admin.adata.Privileges;
import com.deepclone.lw.cmd.admin.bans.BanRequest;
import com.deepclone.lw.cmd.admin.bans.BanType;
import com.deepclone.lw.cmd.admin.bans.BansSummaryResponse;
import com.deepclone.lw.cmd.admin.bans.ListBansResponse;
import com.deepclone.lw.cmd.admin.bans.RequestBanResponse;
import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.cmd.admin.su.AddAdministratorResponse;
import com.deepclone.lw.cmd.admin.su.AddAdministratorResponse.AddressError;
import com.deepclone.lw.interfaces.acm.PasswordProhibitedException;
import com.deepclone.lw.interfaces.admin.AdminDAO;
import com.deepclone.lw.interfaces.admin.AdminDAOException;
import com.deepclone.lw.interfaces.admin.Administration;
import com.deepclone.lw.interfaces.admin.BanMailData;
import com.deepclone.lw.interfaces.admin.BansDAO;
import com.deepclone.lw.interfaces.admin.IpBan;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.mailer.MailData;
import com.deepclone.lw.interfaces.mailer.Mailer;
import com.deepclone.lw.sqld.admin.AdminConnection;
import com.deepclone.lw.sqld.admin.AdminRecord;
import com.deepclone.lw.utils.DigestHelper;
import com.deepclone.lw.utils.EmailAddress;
import com.deepclone.lw.utils.Password;



@Transactional
public class AdministrationBean
		implements Administration
{

	private SystemLogger sysLog;
	private AdminDAO adminDao;
	private IpBan ipBan;
	private BansDAO bansDao;
	private Mailer mailer;


	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.sysLog = logger.getSystemLogger( "Administration" );
	}


	@Autowired( required = true )
	public void setAdminDao( AdminDAO adminDao )
	{
		this.adminDao = adminDao;
	}


	@Autowired( required = true )
	public void setIpBan( IpBan ipBan )
	{
		this.ipBan = ipBan;
	}


	@Autowired( required = true )
	public void setBansDao( BansDAO bansDao )
	{
		this.bansDao = bansDao;
	}


	@Autowired( required = true )
	public void setMailer( Mailer mailer )
	{
		this.mailer = mailer;
	}


	@Override
	public AdminRecord getAdmin( int id )
	{
		return this.adminDao.getAdmin( id );
	}


	@Override
	public AdminRecord login( EmailAddress address , String challenge , String sha1Hash , String md5Hash ,
			InetAddress ipAddress )
	{
		// Check IP ban
		if ( this.ipBan.isBanned( ipAddress ) ) {
			return null;
		}

		// Find admin record
		AdminRecord record = this.adminDao.getAdmin( address );
		if ( record == null ) {
			this.ipBan.increaseBanCounter( ipAddress );
			this.sysLog.log( LogLevel.WARNING ,
					"failed login attempt from unknown address '" + address.getAddress( ) + "'" ).flush( );
			return null;
		}

		// Make sure the administrator is active
		if ( !record.isActive( ) ) {
			this.ipBan.increaseBanCounter( ipAddress );
			this.adminDao.logConnectionAttempt( record.getId( ) , AdminConnection.INACTIVE , ipAddress );
			return null;
		}

		// Check password
		String eSha1 = DigestHelper.digest( "sha-1" , challenge + " " + record.getpSha1( ) );
		String eMd5 = DigestHelper.digest( "md5" , challenge + " " + record.getpMd5( ) );
		if ( eSha1.equals( sha1Hash ) && eMd5.equals( md5Hash ) ) {
			this.adminDao.logConnectionAttempt( record.getId( ) , AdminConnection.SUCCESS , ipAddress );
			return record;
		}

		this.ipBan.increaseBanCounter( ipAddress );
		this.adminDao.logConnectionAttempt( record.getId( ) , AdminConnection.PASSWORD , ipAddress );
		return null;
	}


	@Override
	public void logout( int adminId )
	{
		this.adminDao.logDisconnection( adminId );
	}


	@Override
	public boolean setPassword( int id , String challenge , String sha1Auth , String md5Auth , Password password )
			throws PasswordProhibitedException
	{
		AdminRecord admin = this.getAdmin( id );

		// Check challenge response
		String eSha1 = DigestHelper.digest( "sha-1" , challenge + " " + admin.getpSha1( ) );
		String eMd5 = DigestHelper.digest( "md5" , challenge + " " + admin.getpMd5( ) );
		if ( ! ( eSha1.equals( sha1Auth ) && eMd5.equals( md5Auth ) ) ) {
			return false;
		}

		// Try setting the password
		if ( !this.adminDao.setPassword( id , password.getSha1( ) , password.getMd5( ) ) ) {
			throw new PasswordProhibitedException( );
		}

		return true;
	}


	@Override
	public List< AdminRecord > listAdministrators( )
	{
		return this.adminDao.listAdministrators( );
	}


	@Override
	public AddAdministratorResponse createAdmin( Administrator creator , String address , String name , int privileges )
	{
		int errCode;
		try {
			this.adminDao.createAdmin( address , name , privileges );
			return new AddAdministratorResponse( creator , false );
		} catch ( AdminDAOException e ) {
			errCode = e.getErrorCode( );
		}

		AddressError aError;
		ObjectNameError nError;
		switch ( errCode ) {
			case 1:
				aError = AddressError.NOT_FOUND;
				nError = null;
				break;
			case 2:
				aError = AddressError.STATUS;
				nError = null;
				break;
			case 3:
				aError = null;
				nError = ObjectNameError.UNAVAILABLE;
				break;
			case 4:
				aError = AddressError.ALREADY_ADMIN;
				nError = null;
				break;
			default:
				throw new RuntimeException( "unknown error code " + errCode );
		}

		return new AddAdministratorResponse( creator , aError , address , nError , name , false , privileges );
	}


	@Override
	public AdminRecord resetPassword( Administrator admin , int identifier )
	{
		if ( this.adminDao.resetPassword( identifier , admin.getId( ) ) ) {
			return this.adminDao.getAdmin( identifier );
		}
		return null;
	}


	@Override
	public AdminRecord setPrivileges( Administrator admin , int identifier , Set< Privileges > privileges )
	{
		int privs = 0;
		for ( Privileges p : privileges ) {
			privs = p.grant( privs );
		}
		if ( this.adminDao.setPrivileges( identifier , admin.getId( ) , privs ) ) {
			return this.adminDao.getAdmin( identifier );
		}
		return null;
	}


	@Override
	public BansSummaryResponse getBansSummary( Administrator admin )
	{
		return new BansSummaryResponse( admin , this.bansDao.getSummary( ) );
	}


	@Override
	public ListBansResponse getBans( Administrator admin , BanType type )
	{
		List< BanRequest > bans;
		switch ( type ) {
			case PENDING:
				bans = this.bansDao.getPending( );
				break;
			case ARCHIVED:
				bans = this.bansDao.getArchived( );
				break;
			case VALIDATED:
				bans = this.bansDao.getActive( );
				break;
			default:
				throw new RuntimeException( "unknown ban type " + type );
		}
		return new ListBansResponse( admin , type , bans );
	}


	@Override
	public RequestBanResponse requestBan( Administrator admin , String user , boolean empire , String reason )
	{
		int errCode;
		if ( empire ) {
			errCode = this.bansDao.requestBan( admin.getId( ) , user , reason );
		} else {
			EmailAddress address = new EmailAddress( user );
			if ( !address.isValid( ) ) {
				errCode = 1;
			} else {
				errCode = this.bansDao.requestBan( admin.getId( ) , address , reason );
			}
		}

		if ( errCode == 0 ) {
			return new RequestBanResponse( admin , false );
		}

		RequestBanResponse.Error error;
		switch ( errCode ) {
			case 1:
				error = RequestBanResponse.Error.NOT_FOUND;
				break;
			case 2:
				error = RequestBanResponse.Error.BANNED;
				break;
			default:
				throw new RuntimeException( "unknown error code " + errCode );
		}

		return new RequestBanResponse( admin , error , user , empire , reason );
	}


	@Override
	public void rejectBan( Administrator admin , int id , String reason )
	{
		this.bansDao.rejectBan( admin.getId( ) , id , reason );
	}


	@Override
	public void confirmBan( Administrator admin , int id )
	{
		BanMailData result = this.bansDao.validateBan( admin.getId( ) , id );
		if ( result == null ) {
			return;
		}

		try {
			MailData mail = this.mailer.createMail( result.language , "bannedMail" , result.address );
			mail.setData( "reason" , result.reason );
			mail.queue( );
		} catch ( Exception e ) {
			throw new RuntimeException( e );
		}
	}


	@Override
	public void liftBan( Administrator admin , int id )
	{
		BanMailData result = this.bansDao.liftBan( admin.getId( ) , id );
		if ( result == null ) {
			return;
		}

		try {
			MailData mail = this.mailer.createMail( result.language , "banLiftedMail" , result.address );
			mail.queue( );
		} catch ( Exception e ) {
			throw new RuntimeException( e );
		}
	}


	@Override
	public AdminOverviewResponse getOverview( Administrator admin )
	{
		AdminOverview overview = this.adminDao.getOverview( admin.getId( ) );
		if ( !admin.hasPrivilege( Privileges.NAME ) ) {
			overview.setPendingNames( null );
		}
		if ( !admin.hasPrivilege( Privileges.BANH ) ) {
			overview.setPendingBans( null );
		}
		if ( !admin.hasPrivilege( Privileges.BUGT ) ) {
			overview.setPendingBugs( null );
			overview.setOpenBugs( null );
			overview.setUpdatedBugs( null );
		}
		return new AdminOverviewResponse( admin , overview );
	}
}
