package com.deepclone.lw.beans.prefs;


import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cmd.admin.adata.Administrator;
import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.cmd.player.gdata.account.PrefCategory;
import com.deepclone.lw.cmd.player.gdata.account.PrefChoice;
import com.deepclone.lw.cmd.player.gdata.account.PrefType;
import com.deepclone.lw.cmd.player.gdata.account.PrefValue;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.i18n.TranslationException;
import com.deepclone.lw.interfaces.i18n.Translator;
import com.deepclone.lw.interfaces.prefs.PreferenceDefinitionException;
import com.deepclone.lw.interfaces.prefs.PreferenceDefinitions;
import com.deepclone.lw.interfaces.prefs.PreferenceType;
import com.deepclone.lw.interfaces.prefs.PreferenceTypesRegistry;
import com.deepclone.lw.utils.StoredProc;



public class PreferenceDefinitionsBean
		implements PreferenceDefinitions
{
	private static class PrefDefRecord
	{
		final String groupName;
		final String groupDisplay;
		final String name;
		final String dName;
		final String dDescription;
		final String type;
		final String value;


		public PrefDefRecord( ResultSet rs )
				throws SQLException
		{
			this.groupName = rs.getString( "group_name" );
			this.groupDisplay = rs.getString( "group_display" );
			this.name = rs.getString( "name" );
			this.dName = rs.getString( "d_name" );
			this.dDescription = rs.getString( "d_desc" );
			this.type = rs.getString( "java_type" );
			this.value = rs.getString( "default_value" );
		}
	}

	private TransactionTemplate tTemplate;
	private PreferenceTypesRegistry registry;
	private SystemLogger logger;
	private Translator translator;

	private JdbcTemplate dTemplate;

	private StoredProc uocGroup;
	private StoredProc uocDef;
	private StoredProc fSetDefault;

	private final RowMapper< PrefDefRecord > mPrefDef;


	public PreferenceDefinitionsBean( )
	{
		this.mPrefDef = new RowMapper< PrefDefRecord >( ) {
			@Override
			public PrefDefRecord mapRow( ResultSet rs , int rowNum )
					throws SQLException
			{
				return new PrefDefRecord( rs );
			}
		};
	}


	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );

		this.uocGroup = new StoredProc( dataSource , "defs" , "uoc_preference_group" );
		this.uocGroup.addParameter( "name" , Types.VARCHAR );
		this.uocGroup.addParameter( "display" , Types.VARCHAR );
		this.uocGroup.addOutput( "success" , Types.BOOLEAN );

		this.uocDef = new StoredProc( dataSource , "defs" , "uoc_preference" );
		this.uocDef.addParameter( "g_name" , Types.VARCHAR );
		this.uocDef.addParameter( "p_name" , Types.VARCHAR );
		this.uocDef.addParameter( "d_name" , Types.VARCHAR );
		this.uocDef.addParameter( "d_desc" , Types.VARCHAR );
		this.uocDef.addParameter( "j_type" , Types.VARCHAR );
		this.uocDef.addParameter( "d_val" , Types.VARCHAR );
		this.uocDef.addOutput( "err_code" , Types.INTEGER );

		this.fSetDefault = new StoredProc( dataSource , "defs" , "set_preference_default" );
		this.fSetDefault.addParameter( "admin_id" , Types.INTEGER );
		this.fSetDefault.addParameter( "pref_name" , Types.VARCHAR );
		this.fSetDefault.addParameter( "default_value" , Types.VARCHAR );
	}


	@Autowired( required = true )
	public void setTypesRegistry( PreferenceTypesRegistry registry )
	{
		this.registry = registry;
	}


	@Autowired( required = true )
	public void setTransactionManager( PlatformTransactionManager transactionManager )
	{
		this.tTemplate = new TransactionTemplate( transactionManager );
	}


	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.logger = logger.getSystemLogger( "PreferenceDefinitions" );
	}


	@Autowired( required = true )
	public void setTranslator( Translator translator )
	{
		this.translator = translator;
	}


	@Override
	public void registerGroup( final String name , final String displayName )
			throws PreferenceDefinitionException
	{
		boolean success = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {
			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				return (Boolean) uocGroup.execute( name , displayName ).get( "success" );
			}
		} );
		if ( !success ) {
			throw new PreferenceDefinitionException( PreferenceDefinitionException.Error.MISSING_STRING );
		}
		this.logger.log( LogLevel.DEBUG , "registered preference group '" + name + "'" );
	}


	@Override
	public void registerPreference( final String name , final String group , final String displayName ,
			final String displayDescription , final Object defaultValue )
			throws PreferenceDefinitionException
	{
		final Class< ? > type = defaultValue.getClass( );
		final PreferenceType< ? > pType = this.registry.getType( type );

		int errCode = this.tTemplate.execute( new TransactionCallback< Integer >( ) {
			@Override
			public Integer doInTransaction( TransactionStatus status )
			{
				return (Integer) uocDef.execute( group , name , displayName , displayDescription ,
						type.getSimpleName( ) , pType.convert( defaultValue ) ).get( "err_code" );
			}
		} );

		switch ( errCode ) {
			case 0:
				break;
			case 1:
				throw new PreferenceDefinitionException( PreferenceDefinitionException.Error.MISSING_GROUP );
			case 2:
				throw new PreferenceDefinitionException( PreferenceDefinitionException.Error.MISSING_STRING );
			case 3:
				throw new PreferenceDefinitionException( PreferenceDefinitionException.Error.INVALID_TYPE );
			default:
				throw new RuntimeException( "unknown error code " + errCode );
		}

		this.logger.log( LogLevel.DEBUG , "registered preference '" + name + "' in group '" + group + "'" );
	}


	private String translate( String strName )
	{
		try {
			return this.translator.translate( "en" , strName );
		} catch ( TranslationException e ) {
			throw new RuntimeException( e );
		}
	}


	@Override
	@Transactional
	public List< PrefCategory > getDefaults( )
	{
		List< PrefCategory > categories = new LinkedList< PrefCategory >( );
		PrefCategory currentCategory = null;
		String catName = "";
		for ( PrefDefRecord def : this.dTemplate.query( "SELECT * FROM defs.preferences_view" , this.mPrefDef ) ) {
			// Make sure we're modifying the right category
			if ( currentCategory == null || !def.groupName.equals( catName ) ) {
				currentCategory = new PrefCategory( );
				catName = def.groupName;
				currentCategory.setName( this.translate( def.groupDisplay ) );
				categories.add( currentCategory );
			}

			// Add value
			PrefValue value = new PrefValue( );
			value.setId( def.name );
			value.setName( this.translate( def.dName ) );
			value.setDescription( this.translate( def.dDescription ) );
			value.setValue( def.value );

			PreferenceType< ? > dType = this.registry.getType( def.type );
			if ( dType.getType( ) == String.class ) {
				value.setType( PrefType.STRING );
			} else if ( dType.getType( ) == Integer.class ) {
				value.setType( PrefType.INTEGER );
			} else if ( dType.getType( ) == Boolean.class ) {
				value.setType( PrefType.BOOLEAN );
			} else {
				Map< String , String > choices = dType.getChoices( );
				List< PrefChoice > cList = new LinkedList< PrefChoice >( );
				for ( Map.Entry< String , String > entry : choices.entrySet( ) ) {
					cList.add( new PrefChoice( entry.getKey( ) , this.translate( entry.getValue( ) ) ) );
				}
				value.setType( PrefType.CHOICE );
				value.setChoices( cList );
			}

			currentCategory.addValue( value );
		}

		return categories;
	}


	@Override
	@Transactional
	public void setDefault( Administrator admin , String preference , String value )
	{
		String sql = "SELECT * FROM defs.preferences_view WHERE name = ?";
		PrefDefRecord def;
		try {
			def = this.dTemplate.queryForObject( sql , this.mPrefDef , preference );
		} catch ( EmptyResultDataAccessException e ) {
			return;
		}

		PreferenceType< ? > pType = this.registry.getType( def.type );
		Object inObject = pType.valueOf( value );
		if ( inObject != null ) {
			String dbValue = pType.convert( inObject );
			this.fSetDefault.execute( admin.getId( ) , preference , dbValue );
		}
	}

}
