package com.deepclone.lw.beans.bt.es;


import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;

import com.deepclone.lw.beans.bt.es.data.AccountInformation;
import com.deepclone.lw.beans.bt.es.data.AllianceInformation;
import com.deepclone.lw.beans.bt.es.data.BuildingsInformation;
import com.deepclone.lw.beans.bt.es.data.DebugInformation;
import com.deepclone.lw.beans.bt.es.data.EmpireInformation;
import com.deepclone.lw.beans.bt.es.data.EmpireResourceInformation;
import com.deepclone.lw.beans.bt.es.data.FleetInformation;
import com.deepclone.lw.beans.bt.es.data.MovementInformation;
import com.deepclone.lw.beans.bt.es.data.PlanetInformation;
import com.deepclone.lw.beans.bt.es.data.QueueInformation;
import com.deepclone.lw.beans.bt.es.data.QueueItemInformation;
import com.deepclone.lw.beans.bt.es.data.ResearchGraphInformation;
import com.deepclone.lw.beans.bt.es.data.ResearchInformation;
import com.deepclone.lw.beans.bt.es.data.ResearchLineInformation;
import com.deepclone.lw.beans.bt.es.data.ResourceDeltaInformation;
import com.deepclone.lw.beans.bt.es.data.ResourceProviderInformation;
import com.deepclone.lw.beans.bt.es.data.ShipsInformation;
import com.deepclone.lw.beans.bt.es.data.SystemInformation;
import com.deepclone.lw.interfaces.bt.EmpireSummary;
import com.thoughtworks.xstream.XStream;



/**
 * Empire XML summary generator
 * 
 * <p>
 * This component is used by the bug tracking system to generate XML dumps of an empire's state. It
 * reads the necessary data from the database, convert it to XML dump record classes then serialises
 * the whole thing through XStream.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class EmpireSummaryBean
		implements EmpireSummary
{
	/** Beginning of all dump SQL queries */
	private static final String SQL_START = "SELECT * FROM bugs.dump_";

	/** Ending of (almost) all dump SQL queries */
	private static final String SQL_END = "_view WHERE empire_id = ?";

	/** SQL query that accesses the main empire dump view */
	private static final String Q_EMPIRE = SQL_START + "main" + SQL_END;

	/** SQL query that accesses the resources dump view */
	private static final String Q_RESOURCES = SQL_START + "emp_resources" + SQL_END;

	/** SQL query that accesses the research dump view */
	private static final String Q_RESEARCH = SQL_START + "research" + SQL_END;

	/** SQL query that accesses the empire's planets dump view */
	private static final String Q_PLANETS = SQL_START + "planets" + SQL_END;

	/** SQL query that accesses the planetary resources dump view */
	private static final String Q_PLANET_RESOURCES = SQL_START + "planet_resources" + SQL_END;

	/** SQL query that accesses the build queues dump view */
	private static final String Q_BUILD_QUEUES = SQL_START + "queues" + SQL_END + " ORDER BY queue_order";

	/** SQL query that accesses the buildings dump view */
	private static final String Q_BUILDINGS = SQL_START + "buildings" + SQL_END;

	/** SQL query that accesses the fleets dump view */
	private static final String Q_FLEETS = SQL_START + "fleets" + SQL_END;

	/** SQL query that accesses the ships dump view */
	private static final String Q_SHIPS = SQL_START + "ships" + SQL_END;

	/** JDBC access interface */
	private JdbcTemplate dTemplate;

	/** XStream instance used to serialise XML dumps */
	private final XStream xStream;

	/** Top-level row mapper */
	private final DebugInformationMapper mMainInfo;

	/** Empire resources row mapper */
	private final EmpireResourceInformationMapper mResources;

	/** Empire research row mapper */
	private final ResearchInformationMapper mResearch;

	/** Empire-owned planet row mapper */
	private final PlanetInformationMapper mPlanet;

	/** Planet resources row mapper */
	private final ResourceRowMapper mPlanetResources;

	/** Planet construction queue item row mapper */
	private final QueueItemInformationMapper mQueueItem;

	/** Planet buildings information mapper */
	private final BuildingsInformationMapper mBuildings;

	/** Fleet information row mapper */
	private final FleetInformationMapper mFleet;

	/** Fleet ships row mapper */
	private final ShipsInformationMapper mShips;


	/**
	 * Set up the {@link XStream} instance that generates the dumps as well as the necessary row
	 * mappers.
	 */
	public EmpireSummaryBean( )
	{
		this.xStream = new XStream( );
		this.xStream.processAnnotations( new Class< ? >[] {
				AccountInformation.class , AllianceInformation.class , BuildingsInformation.class ,
				DebugInformation.class , EmpireInformation.class , FleetInformation.class , MovementInformation.class ,
				PlanetInformation.class , QueueInformation.class , QueueItemInformation.class ,
				ResearchLineInformation.class , ResearchGraphInformation.class , ResourceDeltaInformation.class ,
				ResourceProviderInformation.class , ShipsInformation.class , SystemInformation.class
		} );

		this.mMainInfo = new DebugInformationMapper( );
		this.mResources = new EmpireResourceInformationMapper( );
		this.mResearch = new ResearchInformationMapper( );
		this.mPlanet = new PlanetInformationMapper( );
		this.mPlanetResources = new ResourceRowMapper( );
		this.mQueueItem = new QueueItemInformationMapper( );
		this.mBuildings = new BuildingsInformationMapper( );
		this.mFleet = new FleetInformationMapper( );
		this.mShips = new ShipsInformationMapper( );
	}


	/**
	 * Dependency injector that sets the data source
	 * 
	 * @param dataSource
	 *            the data source
	 */
	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );
	}


	/**
	 * Generate an empire's XML summary
	 * 
	 * <p>
	 * Retrieve all information from the database, convert that information into XML dump record
	 * instances through row mappers, then convert the top-level record into XML using the
	 * component's {@link XStream} instance.
	 */
	@Override
	public String getSummary( int empireId )
	{
		DebugInformation empireDump = this.dTemplate.queryForObject( Q_EMPIRE , this.mMainInfo , empireId );
		this.getResources( empireId , empireDump );
		this.getResearch( empireId , empireDump );

		this.getPlanets( empireId , empireDump );
		this.getFleets( empireId , empireDump );

		return this.xStream.toXML( empireDump );
	}


	/**
	 * Extract resources information
	 * 
	 * <p>
	 * Read the list of empire resources from the appropriate view and add the extracted entries to
	 * the empire's list of resources.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param empireDump
	 *            the top-level instance
	 */
	private void getResources( int empireId , DebugInformation empireDump )
	{
		List< EmpireResourceInformation > resources = empireDump.getEmpire( ).getResources( );
		for ( EmpireResourceInformation empRes : this.dTemplate.query( Q_RESOURCES , this.mResources , empireId ) ) {
			resources.add( empRes );
		}
	}


	/**
	 * Get research information
	 * 
	 * <p>
	 * Read all research-related information from the appropriate dump view, and add corresponding
	 * entries to the top-level instance.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param empireDump
	 *            the top-level instance
	 */
	private void getResearch( int empireId , DebugInformation empireDump )
	{
		for ( ResearchInformation ri : this.dTemplate.query( Q_RESEARCH , this.mResearch , empireId ) ) {
			empireDump.getResearch( ).add( ri );
		}
	}


	/**
	 * Get empire planets and related information
	 * 
	 * <p>
	 * Read the planet list and basic planet information from the appropriate dump view, then read
	 * all extra information (resources, build queues, buildings).
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param empireDump
	 *            the top-level instance
	 */
	private void getPlanets( int empireId , DebugInformation empireDump )
	{
		HashMap< Integer , PlanetInformation > planets = new HashMap< Integer , PlanetInformation >( );
		for ( PlanetInformation planet : this.dTemplate.query( Q_PLANETS , this.mPlanet , empireId ) ) {
			empireDump.getPlanets( ).add( planet );
			planets.put( planet.getId( ) , planet );
		}

		this.getPlanetResources( empireId , planets );
		this.getBuildQueues( empireId , planets );
		this.getBuildings( empireId , planets );
	}


	/**
	 * Add resources information to planets
	 * 
	 * <p>
	 * Access resources information for the empire's planets and add the records to each planet.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param planets
	 *            a map associating planet records with their identifier
	 */
	private void getPlanetResources( int empireId , Map< Integer , PlanetInformation > planets )
	{
		for ( PlanetResourceRow resRow : this.dTemplate.query( Q_PLANET_RESOURCES , this.mPlanetResources , empireId ) ) {
			PlanetInformation planet = planets.get( resRow.getPlanetId( ) );
			planet.getResourceDeltas( ).add( resRow.getDelta( ) );

			ResourceProviderInformation resProv = resRow.getProvider( );
			if ( resProv != null ) {
				planet.getResourceProviders( ).add( resProv );
			}
		}
	}


	/**
	 * Add build queue information to planet records
	 * 
	 * <p>
	 * Access the contents of build queues and add the records to the corresponding planets.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param planets
	 *            the map of planets by identifier
	 */
	private void getBuildQueues( int empireId , Map< Integer , PlanetInformation > planets )
	{
		for ( QueueItemInformation queueItem : this.dTemplate.query( Q_BUILD_QUEUES , this.mQueueItem , empireId ) ) {
			PlanetInformation planet = planets.get( queueItem.getPlanetId( ) );

			QueueInformation queue;
			if ( queueItem.isMilitary( ) ) {
				queue = planet.getMilitaryQueue( );
			} else {
				queue = planet.getCivilianQueue( );
			}

			queue.getItems( ).add( queueItem );
		}
	}


	/**
	 * Add buildings information to planet records
	 * 
	 * <p>
	 * Access the contents of the buildings dump view and add the building records to the
	 * corresponding planets.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param planets
	 *            the map of planets by identifier
	 */
	private void getBuildings( int empireId , Map< Integer , PlanetInformation > planets )
	{
		for ( BuildingsInformation bi : this.dTemplate.query( Q_BUILDINGS , this.mBuildings , empireId ) ) {
			planets.get( bi.getPlanetId( ) ).getBuildings( ).add( bi );
		}
	}


	/**
	 * Retrieve information about the empire's fleets
	 * 
	 * <p>
	 * Access the list of fleets, add the records to the top-level instance, then retrieve all
	 * necessary information.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param empireDump
	 *            the top-level empire dump
	 */
	private void getFleets( int empireId , DebugInformation empireDump )
	{
		Map< Long , FleetInformation > fleets = new HashMap< Long , FleetInformation >( );
		for ( FleetInformation fi : this.dTemplate.query( Q_FLEETS , this.mFleet , empireId ) ) {
			empireDump.getFleets( ).add( fi );
			fleets.put( fi.getId( ) , fi );
		}
		this.getShips( empireId , fleets );
	}


	/**
	 * Add ships information to fleet records
	 * 
	 * <p>
	 * Retrieve information about ships and add the records to the corresponding fleets.
	 * 
	 * @param empireId
	 *            the empire's identifier
	 * @param fleets
	 *            the map of fleets by identifier
	 */
	private void getShips( int empireId , Map< Long , FleetInformation > fleets )
	{
		for ( ShipsInformation si : this.dTemplate.query( Q_SHIPS , this.mShips , empireId ) ) {
			fleets.get( si.getFleetId( ) ).getShips( ).add( si );
		}
	}

}
