package com.deepclone.lw.beans.bt.es.data;


import java.io.Serializable;

import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;



/**
 * Resource provider XML record
 * 
 * <p>
 * This class represents the information about a resource provider that will be serialised to the
 * debugging XML dump when a player posts a bug report.
 * 
 * <p>
 * This record exists since version 2 (B6M2) of XML dumps.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@SuppressWarnings( "serial" )
public class ResourceProviderInformation
		implements Serializable
{

	/** Identifier of the resource */
	@XStreamAsAttribute
	private String resource;

	/** Maximal quantity in the provider */
	@XStreamAsAttribute
	@XStreamAlias( "max" )
	private Double maximalQuantity;

	/** Current quantity in the provider */
	@XStreamAsAttribute
	@XStreamAlias( "current" )
	private Double currentQuantity;

	/** Extraction difficulty */
	@XStreamAsAttribute
	private Double difficulty;

	/** Recovery rate */
	@XStreamAsAttribute
	private Double recovery;

	/** Planet-specific mining priority */
	@XStreamAlias( "mining-priority" )
	@XStreamAsAttribute
	private Integer miningPriority;


	/** @return the resource's identifier */
	public String getResource( )
	{
		return this.resource;
	}


	/**
	 * Set the resource's identifier
	 * 
	 * @param resource
	 *            the resource's identifier
	 * 
	 * @throws InvalidDumpContentsException
	 *             if the resource's identifier is <code>null</code>
	 */
	public void setResource( String resource )
			throws InvalidDumpContentsException
	{
		if ( resource == null ) {
			throw new InvalidDumpContentsException( this.getClass( ) , "resource" );
		}
		this.resource = resource;
	}


	/** @return the provider's total capacity */
	public Double getMaximalQuantity( )
	{
		return this.maximalQuantity;
	}


	/**
	 * Set the maximal quantity of resources in the provider
	 * 
	 * @param maximalQuantity
	 *            the provider's total capacity
	 */
	public void setMaximalQuantity( double maximalQuantity )
	{
		this.maximalQuantity = maximalQuantity;
	}


	/** @return the current quantity of resources in the provider */
	public Double getCurrentQuantity( )
	{
		return this.currentQuantity;
	}


	/**
	 * Set the current quantity of resources in the provider
	 * 
	 * @param currentQuantity
	 *            the current quantity of resources in the provider
	 */
	public void setCurrentQuantity( double currentQuantity )
	{
		this.currentQuantity = currentQuantity;
	}


	/** @return the provider's extraction difficulty */
	public Double getDifficulty( )
	{
		return this.difficulty;
	}


	/**
	 * Set the provider's extraction difficulty
	 * 
	 * @param difficulty
	 *            the provider's extraction difficulty
	 */
	public void setDifficulty( double difficulty )
	{
		this.difficulty = difficulty;
	}


	/** @return the provider's recovery rate */
	public Double getRecovery( )
	{
		return this.recovery;
	}


	/**
	 * Set the provider's recovery rate
	 * 
	 * @param recovery
	 *            the provider's recovery rate
	 */
	public void setRecovery( double recovery )
	{
		this.recovery = recovery;
	}


	/**
	 * @return the planet-specific mining priority, or <code>null</code> if the empire's global
	 *         settings are in used
	 */
	public Integer getMiningPriority( )
	{
		return this.miningPriority;
	}


	/**
	 * Set the planet-specific mining priority
	 * 
	 * @param miningPriority
	 *            the planet-specific mining priority
	 */
	public void setMiningPriority( int miningPriority )
	{
		this.miningPriority = miningPriority;
	}

}
