package com.deepclone.lw.beans.eventlog;


import java.text.SimpleDateFormat;
import java.util.LinkedList;
import java.util.List;

import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cmd.admin.adata.Privileges;
import com.deepclone.lw.cmd.admin.logs.ExceptionEntry;
import com.deepclone.lw.cmd.admin.logs.TraceEntry;
import com.deepclone.lw.interfaces.admin.AdminDAO;
import com.deepclone.lw.interfaces.eventlog.ExtendedLogEntry;
import com.deepclone.lw.interfaces.eventlog.LogReader;
import com.deepclone.lw.interfaces.mailer.MailData;
import com.deepclone.lw.interfaces.mailer.Mailer;
import com.deepclone.lw.sqld.admin.AdminRecord;



class AdminErrorMailTask
		implements Runnable
{
	private static class ErrorData
	{
		public final List< AdminRecord > administrators;
		public final List< ExtendedLogEntry > logEntries;


		public ErrorData( List< AdminRecord > administrators , List< ExtendedLogEntry > logEntries )
		{
			this.administrators = administrators;
			this.logEntries = logEntries;
		}

	}

	private final TransactionTemplate tTemplate;
	private final AdminDAO adminDao;
	private final LogReader logReader;
	private final Mailer mailer;


	public AdminErrorMailTask( AdminDAO adminDao , LogReader logReader , Mailer mailer , TransactionTemplate tTemplate )
	{
		this.adminDao = adminDao;
		this.logReader = logReader;
		this.mailer = mailer;
		this.tTemplate = tTemplate;
	}


	@Override
	public void run( )
	{
		// Get errors and administrators
		ErrorData errorData;
		errorData = this.tTemplate.execute( new TransactionCallback< ErrorData >( ) {
			@Override
			public ErrorData doInTransaction( TransactionStatus status )
			{
				return getErrorData( );
			}
		} );

		// No receiving administrators or no errors -> exit
		if ( errorData.administrators.isEmpty( ) || errorData.logEntries.isEmpty( ) ) {
			return;
		}

		// Generate message contents
		String message = this.buildMessage( errorData.logEntries );
		for ( AdminRecord admin : errorData.administrators ) {
			try {
				MailData mail = this.mailer.createMail( "en" , "adminErrorMail" , admin.getAddress( ) );
				mail.setData( "contents" , message );
				mail.queue( );
			} catch ( Exception e ) {
				throw new RuntimeException( e );
			}
		}
	}


	private ErrorData getErrorData( )
	{
		List< AdminRecord > admins = new LinkedList< AdminRecord >( );
		for ( AdminRecord admin : this.adminDao.listAdministrators( ) ) {
			if ( Privileges.BUGM.hasPrivilege( admin.getPrivileges( ) ) ) {
				admins.add( admin );
			}
		}
		return new ErrorData( admins , this.logReader.getErrorEntries( ) );
	}


	private String buildMessage( List< ExtendedLogEntry > logEntries )
	{
		StringBuilder builder = new StringBuilder( );
		SimpleDateFormat dFmt = new SimpleDateFormat( "yyyy-MM-dd" );
		SimpleDateFormat tFmt = new SimpleDateFormat( "HH:mm:ss" );

		for ( ExtendedLogEntry entry : logEntries ) {
			builder.append( "Date: " ).append( dFmt.format( entry.logEntry.getTimestamp( ) ) );
			builder.append( "\nTime: " ).append( tFmt.format( entry.logEntry.getTimestamp( ) ) );
			builder.append( "\nComponent: " ).append( entry.logEntry.getAbout( ) );
			builder.append( "\nMessage: " ).append( entry.logEntry.getEntry( ) ).append( "\n" );

			for ( ExceptionEntry eEntry : entry.exceptions ) {
				builder.append( "\n    " ).append( eEntry.getClassName( ) );
				if ( eEntry.getMessage( ) != null && !"".equals( eEntry ) ) {
					builder.append( ": " ).append( eEntry.getMessage( ) );
				}

				int i = 0;
				for ( TraceEntry tEntry : eEntry.getTrace( ) ) {
					builder.append( "\n        " );
					if ( i++ > 5 ) {
						builder.append( "..." );
						break;
					}
					builder.append( tEntry.getLocation( ) );
				}
			}

			builder.append( "\n\n\n" );
		}

		return builder.toString( );
	}

}
