package com.deepclone.lw.beans.eventlog;


import java.sql.Types;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.LinkedBlockingQueue;

import javax.sql.DataSource;

import org.springframework.dao.DataAccessResourceFailureException;
import org.springframework.transaction.CannotCreateTransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.sqld.sys.ExceptionLog;
import com.deepclone.lw.sqld.sys.StackTraceLog;
import com.deepclone.lw.sqld.sys.SystemLogEntry;
import com.deepclone.lw.utils.StoredProc;



/**
 * The log writer task is spawned by the log writer bean. Whenever log entries are pushed to the log
 * writer, they are added to a queue which this tasks continuously reads. If a queue item carrying
 * the null value is found, the queue is flushed and the task terminates.
 * 
 * @author tseeker
 */
class LogWriterTask
		implements Runnable
{

	/** The queue */
	private final LinkedBlockingQueue< EntryQueueItem > queue;

	/** Transaction management interface */
	private final TransactionTemplate tTemplate;

	private StoredProc fLog;
	private StoredProc fException;
	private StoredProc fTrace;


	/** Copies the required references */
	LogWriterTask( LinkedBlockingQueue< EntryQueueItem > queue , DataSource dataSource , TransactionTemplate tTemplate )
	{
		this.queue = queue;
		this.tTemplate = tTemplate;

		this.fLog = new StoredProc( dataSource , "sys" , "write_log" );
		this.fLog.addParameter( "component" , Types.VARCHAR );
		this.fLog.addParameter( "level" , "log_level" );
		this.fLog.addParameter( "message" , Types.VARCHAR );
		this.fLog.addOutput( "entry_id" , Types.BIGINT );

		this.fException = new StoredProc( dataSource , "sys" , "append_exception" );
		this.fException.addParameter( "log_id" , Types.BIGINT );
		this.fException.addParameter( "class_name" , Types.VARCHAR );
		this.fException.addParameter( "message" , Types.VARCHAR );
		this.fException.addOutput( "entry_id" , Types.BIGINT );

		this.fTrace = new StoredProc( dataSource , "sys" , "append_trace" );
		this.fTrace.addParameter( "exc_id" , Types.BIGINT );
		this.fTrace.addParameter( "location" , Types.VARCHAR );
		this.fTrace.addParameter( "file_name" , Types.VARCHAR );
		this.fTrace.addParameter( "line_number" , Types.INTEGER );
	}


	/**
	 * Implements the consumer side of the queue; when an item is found, all items are flushed and
	 * written to the DB in the same transaction.
	 */
	@Override
	public void run( )
	{
		boolean keepRunning = true;
		while ( keepRunning ) {
			LinkedList< EntryQueueItem > items;
			items = new LinkedList< EntryQueueItem >( );

			// Wait for the next queue entry
			EntryQueueItem item;
			try {
				item = this.queue.take( );
			} catch ( InterruptedException e ) {
				continue;
			}

			// Take everything else on the queue
			this.queue.drainTo( items );
			items.add( item );

			// Process the items obtained from the queue, watching for null values as they will
			// interrupt the task
			final LinkedList< SystemLogEntry > okItems = new LinkedList< SystemLogEntry >( );
			for ( EntryQueueItem i : items ) {
				if ( i.entry == null ) {
					keepRunning = false;
				} else {
					okItems.add( i.entry );
				}
			}

			// If there's nothing to write, restart the loop (that usually means we're being
			// terminated)
			if ( okItems.isEmpty( ) ) {
				continue;
			}

			// Try writing to the DB
			boolean writeSuccess = false;
			while ( ! writeSuccess ) {
				try {
					this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {
						@Override
						protected void doInTransactionWithoutResult( TransactionStatus status )
						{
							writeLogEntries( okItems );
						}
					} );
					writeSuccess = true;
				} catch ( DataAccessResourceFailureException e ) {
					// Do nothing
				} catch ( CannotCreateTransactionException e ) {
					// Do nothing
				}
				
				if (! writeSuccess ) {
					if ( !this.queue.isEmpty( ) ) {
						// If there is stuff in the queue, abort writing.
						// We might actually need to exit.
						break;
					}

					try {
						Thread.sleep( 2000 );
					} catch ( InterruptedException e1 ) {
						break;
					}
				}
			}
		}
	}


	private void writeLogEntries( List< SystemLogEntry > entries )
	{
		for ( SystemLogEntry entry : entries ) {
			Long id = (Long) this.fLog.execute( entry.getComponent( ) , entry.getLevel( ).toString( ) ,
					entry.getMessage( ) ).get( "entry_id" );

			for ( ExceptionLog exc : entry.getException( ) ) {
				Long eid = (Long) this.fException.execute( id , exc.getClassName( ) , exc.getMessage( ) ).get(
						"entry_id" );

				for ( StackTraceLog st : exc.getStack( ) ) {
					this.fTrace.execute( eid , st.getLocation( ) , st.getFileName( ) , st.getLine( ) );
				}
			}
		}
	}

}
