package com.deepclone.lw.beans.events;


import javax.sql.DataSource;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.interfaces.sys.Ticker;
import com.deepclone.lw.interfaces.sys.Ticker.Frequency;



/**
 * Events processor component
 * 
 * <p>
 * This component acquires the necessary dependencies to create an {@link EventProcessorTask}, does
 * so, and registers it with the ticker. It clears the reference to the task when destroyed.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
class EventProcessorBean
		implements InitializingBean , DisposableBean
{

	/** The ticker */
	private Ticker ticker;

	/** The Spring transaction template */
	private TransactionTemplate tTemplate;

	/** The database interface */
	private DataSource dataSource;

	/** The task */
	private Runnable task;


	/**
	 * Dependency injector that sets the ticker component
	 * 
	 * @param ticker
	 *            the ticker component
	 */
	@Autowired( required = true )
	public void setTicker( Ticker ticker )
	{
		this.ticker = ticker;
	}


	/**
	 * Dependency injector that sets the transaction manager
	 * 
	 * <p>
	 * Creates the transaction template from the transaction manager
	 * 
	 * @param tManager
	 *            the Spring transaction manager
	 */
	@Autowired( required = true )
	public void setTransactionManager( PlatformTransactionManager tManager )
	{
		this.tTemplate = new TransactionTemplate( tManager );
	}


	/**
	 * Dependency injector that sets the database interface
	 * 
	 * @param dataSource
	 *            the database interface
	 */
	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dataSource = dataSource;
	}


	/**
	 * Initialise and register the event processing task
	 * 
	 * <p>
	 * When the component's dependencies have been injected, this method creates the
	 * {@link EventProcessorTask} instance and registers it to the ticker.
	 */
	@Override
	public void afterPropertiesSet( )
	{
		this.task = new EventProcessorTask( this.tTemplate , this.dataSource );
		this.ticker.registerTask( Frequency.HIGH , "Events processor" , this.task );
	}


	/**
	 * Clear the task reference
	 * 
	 * <p>
	 * Remove the reference to the {@link EventProcessorTask} instance, allowing the ticker to clean
	 * up.
	 */
	@Override
	public void destroy( )
	{
		this.task = null;
	}
}
