package com.deepclone.lw.beans.events;


import javax.sql.DataSource;

import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.utils.StoredProc;



/**
 * Event processing task
 * 
 * <p>
 * This class implements the task responsible for regular database event processing calls. It is
 * registered as a high-frequency ticker task by {@link EventProcessorBean}.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class EventProcessorTask
		implements Runnable
{

	/** The Spring transaction template */
	private final TransactionTemplate tTemplate;

	/** The stored procedure that implements event processing */
	private final StoredProc eqProcess;


	/**
	 * Initialise the transaction template and stored procedure interface
	 * 
	 * @param tTemplate
	 *            the transaction template
	 * @param dataSource
	 *            the database interface
	 */
	public EventProcessorTask( TransactionTemplate tTemplate , DataSource dataSource )
	{
		this.tTemplate = tTemplate;
		this.eqProcess = new StoredProc( dataSource , "events" , "eq_process" );
		this.eqProcess.addOutput( "_result" , java.sql.Types.BOOLEAN );
	}


	/**
	 * When the task executes, it will call <code>events.eq_process()</code> repeatedly until the
	 * stored procedure returns <code>false</code>. Each call is made in a separate transaction.
	 */
	@Override
	public void run( )
	{
		boolean hasMore;
		do {
			hasMore = this.processBatch( );
		} while ( hasMore );
	}


	/**
	 * Process a batch of events
	 * 
	 * <p>
	 * This method spawns a transaction in which the <code>events.eq_process()</code> stored
	 * procedure is called.
	 * 
	 * @return <code>true</code> if events were processed, <code>false</code> if the queues were
	 *         empty.
	 */
	private boolean processBatch( )
	{
		return this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				return (Boolean) EventProcessorTask.this.eqProcess.execute( ).get( "_result" );
			}

		} );
	}
}
