package com.deepclone.lw.beans.i18n;


import java.util.Map;
import java.util.Set;

import com.deepclone.lw.interfaces.i18n.*;



/**
 * The implementation of the I18N administrative session uses the manager's shared {@link I18NData}
 * instance to modify the I18N database, logging all actions in the process. It also provides some
 * read access to languages and translations, but (as opposed to access through the
 * {@link Translator} bean) completely ignores language support requirements, allowing new or
 * incomplete languages to be managed.
 * 
 * @author tseeker
 */
class I18NAdministrationImpl
		implements I18NAdministration
{

	/** The shared I18N data instance */
	private I18NData data;
	private int administrator;


	/**
	 * Copies the various required references
	 * 
	 * @param data
	 *            shared I18N data instance
	 * @param administrator 
	 */
	I18NAdministrationImpl( I18NData data, int administrator )
	{
		this.data = data;
		this.administrator = administrator;
	}


	/* Documented in I18NAdministration interface */
	@Override
	public Set< String > getLanguages( )
	{
		this.data.readLock( ).lock( );
		try {
			return this.data.getLanguages( );
		} finally {
			this.data.readLock( ).unlock( );
		}
	}


	/* Documented in I18NAdministration interface */
	@Override
	public String getLanguageName( String language )
			throws UnknownLanguageException
	{
		String l;
		this.data.readLock( ).lock( );
		try {
			l = this.data.getLanguageName( language );
		} finally {
			this.data.readLock( ).unlock( );
		}

		if ( l == null ) {
			throw new UnknownLanguageException( language );
		}
		return l;
	}


	/* Documented in I18NAdministration interface */
	@Override
	public double getLanguageSupport( String language )
			throws UnknownLanguageException
	{
		int sCount;
		double lSize;

		this.data.readLock( ).lock( );
		try {
			if ( !this.data.hasLanguage( language ) ) {
				throw new UnknownLanguageException( language );
			}
			lSize = this.data.getLanguageSize( language );
			sCount = this.data.getStringsCount( );
		} finally {
			this.data.readLock( ).unlock( );
		}

		if ( sCount == 0 ) {
			return 0;
		}
		return lSize / (double) sCount;
	}


	/* Documented in I18NAdministration interface */
	@Override
	public void createLanguage( String language , String name )
			throws DuplicateLanguageException
	{
		this.data.writeLock( ).lock( );
		try {
			if ( !this.data.addLanguage( this.administrator , language , name ) ) {
				throw new DuplicateLanguageException( language );
			}
		} finally {
			this.data.writeLock( ).unlock( );
		}
	}


	/* Documented in I18NAdministration interface */
	@Override
	public void setLanguageName( String language , String name )
			throws UnknownLanguageException
	{
		String oldName;

		this.data.writeLock( ).lock( );
		try {
			oldName = this.data.updateLanguage( this.administrator , language , name );
			if ( oldName == null ) {
				throw new UnknownLanguageException( language );
			}
		} finally {
			this.data.writeLock( ).unlock( );
		}
	}


	@Override
	public Set< String > getStrings( )
	{
		this.data.readLock( ).lock( );
		try {
			return this.data.getStrings( );
		} finally {
			this.data.readLock( ).unlock( );
		}
	}


	/* Documented in I18NAdministration interface */
	@Override
	public String getTranslation( String language , String string )
			throws UnknownStringException , UnknownLanguageException
	{
		this.data.readLock( ).lock( );
		try {
			if ( !this.data.hasLanguage( language ) ) {
				throw new UnknownLanguageException( language );
			}
			if ( !this.data.hasString( string ) ) {
				throw new UnknownStringException( string );
			}
			return this.data.getTranslation( language , string );
		} finally {
			this.data.readLock( ).unlock( );
		}
	}


	@Override
	public boolean updateTranslation( String language , String string , String translation )
			throws UnknownStringException , UnknownLanguageException
	{
		String oldString;
		this.data.writeLock( ).lock( );
		try {
			if ( !this.data.hasLanguage( language ) ) {
				throw new UnknownLanguageException( language );
			}
			if ( !this.data.hasString( string ) ) {
				throw new UnknownStringException( string );
			}
			oldString = this.data.setTranslation( this.administrator , language , string , translation );
		} finally {
			this.data.writeLock( ).unlock( );
		}
		return ( oldString == null );
	}


	/* Documented in I18NAdministration interface */
	@Override
	public void createString( String string , Map< String , String > translations )
			throws DuplicateStringException , UnknownLanguageException , InvalidUpdateException
	{
		Set< String > languages;

		this.data.writeLock( ).lock( );
		try {
			// Check for duplicate string ID
			if ( this.data.hasString( string ) ) {
				throw new DuplicateStringException( string );
			}

			// Make sure all specified languages actually exist
			languages = this.data.getLanguages( );
			for ( String lId : translations.keySet( ) ) {
				if ( !languages.contains( lId ) ) {
					throw new UnknownLanguageException( lId );
				}
				languages.remove( lId );
			}

			// Make sure no 100% supported language becomes unsupported because a translation is
			// missing
			int sCount = this.data.getStringsCount( );
			for ( String lId : languages ) {
				if ( this.data.getLanguageSize( lId ) < sCount ) {
					languages.remove( lId );
				}
			}
			if ( !languages.isEmpty( ) ) {
				throw new InvalidUpdateException( languages );
			}

			// Create the string
			this.data.createString( this.administrator , string , translations );
		} finally {
			this.data.writeLock( ).unlock( );
		}
	}

}
