package com.deepclone.lw.beans.i18n;


import java.sql.Types;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.sql.DataSource;

import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.utils.StoredProc;



/**
 * This class is used by all parts of the I18N support system; it centralises all I18N-related data,
 * providing both read and write access. It implements the {@link ReadWriteLock} interface, although
 * only as a convenience.
 * 
 * @author tseeker
 */
class I18NData
		implements ReadWriteLock
{
	/** The instance's read/write lock */
	private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock( );

	/** Database interface */
	private final JdbcTemplate dTemplate;

	/** Transaction manager interface */
	private final TransactionTemplate tTemplate;

	/** String definitions, by identifier */
	private HashSet< String > strings;

	/** Languages by identifier. Each language is represented by a {@link LanguageStore} */
	private HashMap< String , LanguageStore > languages;

	private StoredProc fUocLanguage;
	private StoredProc fUocTranslation;


	/** Copies the required references then loads all data */
	I18NData( DataSource dataSource , TransactionTemplate tTemplate )
	{
		this.dTemplate = new JdbcTemplate( dataSource );
		this.tTemplate = tTemplate;

		this.fUocLanguage = new StoredProc( dataSource , "defs" , "uoc_language" );
		this.fUocLanguage.addParameter( "language" , Types.VARCHAR );
		this.fUocLanguage.addParameter( "name" , Types.VARCHAR );
		this.fUocLanguage.addParameter( "admin_id" , Types.INTEGER );

		this.fUocTranslation = new StoredProc( dataSource , "defs" , "uoc_translation" );
		this.fUocTranslation.addParameter( "language" , Types.VARCHAR );
		this.fUocTranslation.addParameter( "string_id" , Types.VARCHAR );
		this.fUocTranslation.addParameter( "contents" , Types.VARCHAR );
		this.fUocTranslation.addParameter( "admin_id" , Types.INTEGER );

		this.loadAll( );
	}


	/**
	 * (Re)initialises the strings and languages definition maps, then starts a loader transaction
	 * to fill them from the database's contents.
	 */
	void loadAll( )
	{
		this.strings = new HashSet< String >( );
		this.languages = new HashMap< String , LanguageStore >( );

		LoaderTransaction trans = new LoaderTransaction( this.dTemplate , this.strings , this.languages );
		this.tTemplate.execute( trans );
	}


	/** @return the set of known language identifiers */
	Set< String > getLanguages( )
	{
		return new HashSet< String >( this.languages.keySet( ) );
	}


	/**
	 * @param identifier
	 *            a language's identifier
	 * @return whether a language exists or not
	 */
	boolean hasLanguage( String identifier )
	{
		return this.languages.containsKey( identifier );
	}


	/**
	 * @param string
	 *            a string identifier
	 * @return whether such a string definition exists or not
	 */
	boolean hasString( String string )
	{
		return this.strings.contains( string );
	}


	/**
	 * @param identifier
	 *            a language's identifier
	 * @return the amount of translations defined for the specified language, or 0 if the language
	 *         does not exist
	 */
	int getLanguageSize( String identifier )
	{
		if ( !this.hasLanguage( identifier ) ) {
			return 0;
		}
		return this.languages.get( identifier ).getTranslationsCount( );
	}


	/** @return the amount of defined string identifiers */
	int getStringsCount( )
	{
		return this.strings.size( );
	}


	/**
	 * Checks whether a language is fully supported.
	 * 
	 * @param identifier
	 *            the language's identifier
	 * @return true if there's a translation for each string definition in the specified language.
	 */
	boolean isLanguageComplete( String identifier )
	{
		return this.getLanguageSize( identifier ) == this.getStringsCount( );
	}


	/** @return the set of defined string identifiers */
	Set< String > getStrings( )
	{
		return new HashSet< String >( this.strings );
	}


	String getLanguageName( String id )
	{
		if ( !this.hasLanguage( id ) ) {
			return null;
		}
		return this.languages.get( id ).getLanguageName( );
	}


	/**
	 * Creates a new language definition in the database.
	 * 
	 * @param id
	 *            the new language's identifier
	 * @param name
	 *            the new language's name
	 * @return <code>true</code> on success, <code>false</code> otherwise
	 */
	boolean addLanguage( int administrator , String id , String name )
	{
		if ( this.hasLanguage( id ) ) {
			return false;
		}
		this.uocLanguage( administrator , id , name );
		this.languages.put( id , new LanguageStore( id , name ) );
		return true;
	}


	/**
	 * Modifies the name of a language.
	 * 
	 * @param administrator
	 * 
	 * @param id
	 *            the language's identifier
	 * @param name
	 *            the language's new name
	 * @return null if the language does not exist, or the old name of the language if it does
	 */
	String updateLanguage( int administrator , String id , String name )
	{
		LanguageStore ls = this.languages.get( id );
		if ( ls == null ) {
			return null;
		}

		String oldName = ls.getLanguageName( );
		this.uocLanguage( administrator , id , name );
		ls.setLanguageName( name );
		return oldName;
	}


	private void uocLanguage( final int administrator , final String id , final String name )
	{
		this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {
			@Override
			protected void doInTransactionWithoutResult( TransactionStatus status )
			{
				fUocLanguage.execute( id , name , administrator );
			}
		} );
	}


	/**
	 * Accesses the {@link Translation} object for a given language/string identifier pair.
	 * 
	 * @param language
	 *            the language's identifier
	 * @param string
	 *            the string's identifier
	 * @return the translation
	 * @throws NullPointerException
	 *             if the language does not exist
	 */
	String getTranslation( String language , String string )
	{
		return this.languages.get( language ).getTranslation( string );
	}


	/**
	 * Access the store for some language
	 * 
	 * @param language
	 *            the language to access
	 * 
	 * @return the language store, or <code>null</code> if the language is not defined.
	 */
	LanguageStore getStore( String language )
	{
		return this.languages.get( language );
	}


	/**
	 * Sets or creates the translation for a given language/string identifier pair.
	 * 
	 * @param administrator
	 * 
	 * @param language
	 *            the language's identifier
	 * @param string
	 *            the string's identifier
	 * @param translation
	 *            the translated text for the string
	 * @return the old translated text if the translation existed or null if a new translation was
	 *         created
	 * @throws NullPointerException
	 *             if the language does not exist
	 * @throws IllegalArgumentException
	 *             if the string does not exist
	 */
	String setTranslation( final int administrator , final String language , final String string ,
			final String translation )
	{
		// Get existing translation
		LanguageStore store = this.languages.get( language );
		String old = store.getTranslation( string );
		if ( !this.strings.contains( string ) ) {
			throw new IllegalArgumentException( );
		}

		// Create or update the database record
		this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {
			@Override
			protected void doInTransactionWithoutResult( TransactionStatus status )
			{
				fUocTranslation.execute( language , string , translation , administrator );
			}
		} );

		// Store the new translation
		store.addTranslation( string , translation );
		return old;
	}


	/**
	 * Creates a string from scratch, storing initial translations along with it.
	 * 
	 * @param string
	 *            the new string's identifier
	 * @param translations
	 *            a language identifiers -> translated text map to use as the string's initial
	 *            translations.
	 * @throws ConstraintViolationException
	 *             if the string identifier already existed in the DB
	 * @throws NullPointerException
	 *             if one of the languages does not exist
	 * @throws IllegalArgumentException
	 *             if the string exists
	 */
	void createString( final int administrator , final String string , final Map< String , String > translations )
	{
		if ( this.strings.contains( string ) ) {
			throw new IllegalArgumentException( );
		}

		// Map language stores to translations
		Map< LanguageStore , String > nTrans = new HashMap< LanguageStore , String >( );
		for ( String lId : translations.keySet( ) ) {
			LanguageStore store = this.languages.get( lId );
			nTrans.put( store , translations.get( lId ) );
		}

		// Update the database
		this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {

			@Override
			protected void doInTransactionWithoutResult( TransactionStatus status )
			{
				for ( Map.Entry< String , String > entry : translations.entrySet( ) ) {
					fUocTranslation.execute( entry.getKey( ) , string , entry.getValue( ) , administrator );
				}
			}

		} );

		// Add the string definition to the local cache
		this.strings.add( string );

		// Add the various translations to each language's store
		for ( Map.Entry< LanguageStore , String > entry : nTrans.entrySet( ) ) {
			entry.getKey( ).addTranslation( string , entry.getValue( ) );
		}
	}


	/** @return the instance's read lock */
	@Override
	public Lock readLock( )
	{
		return this.lock.readLock( );
	}


	/** @return the instance's write lock */
	@Override
	public Lock writeLock( )
	{
		return this.lock.writeLock( );
	}

}
