package com.deepclone.lw.beans.mailer;


import java.util.concurrent.LinkedBlockingQueue;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.mail.MailSender;

import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.interfaces.eventlog.*;
import com.deepclone.lw.interfaces.i18n.*;
import com.deepclone.lw.interfaces.mailer.*;



/**
 * The mailer bean spawns a queue-handling task on initialisation. It is responsible for
 * initialising mail data instances, giving them access to the queue-handling task. When the
 * container destroys the bean, it inserts a terminator in the mail queue then waits for the task to
 * end.
 * 
 * @author tseeker
 */
public class MailerBean
		implements Mailer , InitializingBean , DisposableBean
{
	/** The default From: address for mails sent by the bean */
	private final static String defaultSender = "webmaster@legacyworlds.com";

	/** Translator bean */
	private Translator translator;

	/** Spring mail sender bean */
	private MailSender mailSender;

	/** System logger for the mailer bean */
	private SystemLogger logger;

	/** From: address to use when sending mails */
	private String mailFrom = defaultSender;

	/** Mail queue */
	private LinkedBlockingQueue< MailQueueItem > queue;

	/** Queue-handling task */
	private MailQueueTask task;


	/**
	 * Sets the translator bean (DI)
	 * 
	 * @param translator
	 *            the translator bean
	 */
	@Autowired( required = true )
	public void setTranslator( Translator translator )
	{
		this.translator = translator;
	}


	/**
	 * Sets the mail sender bean (DI)
	 * 
	 * @param mailSender
	 *            the mail sender bean
	 */
	@Autowired( required = true )
	public void setMailSender( MailSender mailSender )
	{
		this.mailSender = mailSender;
	}


	/**
	 * Initialises the system logger (DI)
	 * 
	 * @param logger
	 *            the logger bean
	 */
	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.logger = logger.getSystemLogger( "Mailer" );
	}


	/**
	 * Changes the From: address.
	 * 
	 * @param mailFrom
	 *            the new From: address
	 */
	public void setMailFrom( String mailFrom )
	{
		this.mailFrom = mailFrom;
	}


	/**
	 * Creates the mail queue and spawns the queue-handling task when the bean is initialised.
	 */
	@Override
	public void afterPropertiesSet( )
	{
		this.logger.log( LogLevel.INFO , "mailer task initialising" ).flush( );

		this.queue = new LinkedBlockingQueue< MailQueueItem >( );
		this.task = new MailQueueTask( this.logger , this.mailFrom , this.queue , this.mailSender );

		Thread t = new Thread( this.task );
		t.start( );
	}


	/**
	 * Terminates and waits for the queue-handling task when the bean is destroyed.
	 */
	@Override
	public void destroy( )
	{
		this.task.stop( );
		this.task = null;
		this.queue = null;
	}


	/* Documentation in Mailer interface */
	@Override
	public MailData createMail( String language , String contentsDef , String target )
			throws TranslationException
	{

		String template = this.translator.translate( language , contentsDef );
		return new MailDataImpl( this.task , target , template );
	}

}
