package com.deepclone.lw.beans.game.resources;


import java.sql.ResultSet;
import java.sql.SQLException;

import com.deepclone.lw.cmd.player.gdata.planets.PlanetResourceRecord;
import com.deepclone.lw.cmd.player.gdata.planets.ResourceProviderRecord;



/**
 * Row mapper for planet resources
 * 
 * <p>
 * This class maps rows obtained from the planet resources information stored procedure into
 * instances of {@link PlanetResourceRecord} which can be sent to a client.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class PlanetResourceMapper
		extends AbstractResourceMapper< PlanetResourceRecord >
{

	/**
	 * Map a row from <code>emp.get_planet_resources( )</code>
	 * 
	 * <p>
	 * This method extracts fields that are always present, generating the
	 * {@link PlanetResourceRecord} instance and setting it accordingly, then reads resource
	 * provider columns and adds a {@link ResourceProviderRecord} instance to the record if
	 * necessary.
	 */
	@Override
	public PlanetResourceRecord mapRow( ResultSet rs , int rowNum )
			throws SQLException
	{
		PlanetResourceRecord resource = getResourceFields( rs );
		getResourceProvider( rs , resource );
		return resource;
	}


	/**
	 * Map common fields into a {@link PlanetResourceRecord} instance
	 * 
	 * <p>
	 * This method creates the instance then reads descriptive fields, then extracts the income,
	 * upkeep and investment fields.
	 * 
	 * @param rs
	 *            the JDBC result set that contains the row being extracted
	 * 
	 * @return the new {@link PlanetResourceRecord} instance with its common fields set
	 * 
	 * @throws SQLException
	 *             if a SQLException is encountered getting column values
	 */
	private PlanetResourceRecord getResourceFields( ResultSet rs )
			throws SQLException
	{
		PlanetResourceRecord resource = new PlanetResourceRecord( );
		this.getResourceDescription( resource , rs );
		resource.setIncome( rs.getLong( "pres_income" ) );
		resource.setUpkeep( rs.getLong( "pres_upkeep" ) );
		resource.setInvested( rs.getLong( "pres_invested" ) );
		return resource;
	}


	/**
	 * Map resource provider fields if they are present.
	 * 
	 * <p>
	 * Check if the record includes resource provider information. If it does, extract the fields'
	 * values into a {@link ResourceProviderRecord} instance and add it to the
	 * {@link PlanetResourceRecord}.
	 * 
	 * @param rs
	 *            the JDBC result set that contains the row being extracted
	 * @param resource
	 *            the {@link PlanetResourceRecord} instance to add information to
	 * 
	 * @throws SQLException
	 *             if a SQLException is encountered getting column values
	 */
	private void getResourceProvider( ResultSet rs , PlanetResourceRecord resource )
			throws SQLException
	{
		long capacity = rs.getLong( "resprov_capacity" );
		if ( rs.wasNull( ) ) {
			return;
		}

		ResourceProviderRecord provider = new ResourceProviderRecord( );

		provider.setCapacity( capacity );
		provider.setQuantity( rs.getLong( "resprov_quantity" ) );
		provider.setDifficulty( rs.getInt( "resprov_difficulty" ) );
		provider.setPriority( rs.getInt( "mset_weight" ) );

		resource.setResourceProvider( provider );
	}

}
