package com.deepclone.lw.beans.game.resources;


import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;

import com.deepclone.lw.cmd.player.gdata.PlanetListResourceRecord;
import com.deepclone.lw.cmd.player.gdata.empire.EmpireResourceRecord;
import com.deepclone.lw.cmd.player.gdata.planets.PlanetResourceRecord;
import com.deepclone.lw.interfaces.game.resources.ResourcesInformationDAO;



/**
 * Resource information access component
 * 
 * <p>
 * This component's goal is to read information about resources from the database and return the
 * records in some usable format.
 * 
 * <p>
 * It does not contain any method that actually change the database.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
class ResourcesInformationDAOBean
		implements ResourcesInformationDAO
{

	/** SQL query that fetches a planet's resources information */
	private static final String Q_PLANET_RESOURCES = "SELECT * FROM emp.get_planet_resources( ? )";

	/** SQL query that fetches an empire's resources information */
	private static final String Q_EMPIRE_RESOURCES = "SELECT * FROM emp.resources_view WHERE empire_id = ?";

	/** SQL query that fetches the resource records for an empire's planet list */
	private static final String Q_PLIST_RESOURCES = "SELECT * FROM emp.plist_resources_view WHERE empire_id = ?";

	/** Row mapper for planet resources */
	private final PlanetResourceMapper mPlanetResource;

	/** Row mapper for empire resources */
	private final EmpireResourceMapper mEmpireResource;

	/** Row mapper for planet list records */
	private final PlanetListResourceRecordMapper mPlanetListRecord;

	/** Spring JDBC interface */
	private JdbcTemplate dTemplate;


	/** Initialise the necessary row mappers */
	public ResourcesInformationDAOBean( )
	{
		this.mPlanetResource = new PlanetResourceMapper( );
		this.mEmpireResource = new EmpireResourceMapper( );
		this.mPlanetListRecord = new PlanetListResourceRecordMapper( );
	}


	/**
	 * Dependency injector that sets the data source
	 * 
	 * @param dataSource
	 *            the data source
	 */
	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );
	}


	/**
	 * Run the planet resources information query and extract the data
	 * 
	 * <p>
	 * This implementation simply executes a query using the <code>emp.get_planet_resources()</code>
	 * stored procedure and maps the resulting rows into a list of {@link PlanetResourceRecord}
	 * instances.
	 */
	@Override
	public List< PlanetResourceRecord > getPlanetInformation( int planet )
	{
		return this.dTemplate.query( Q_PLANET_RESOURCES , this.mPlanetResource , planet );
	}


	/**
	 * Run the empire resources information query and extract the data
	 * 
	 * <p>
	 * This implementation executes a query on <code>emp.resources_view</code> for a specific
	 * empire, and maps the resulting rows into a list of {@link EmpireResourceRecord} instances.
	 */
	@Override
	public List< EmpireResourceRecord > getEmpireInformation( int empire )
	{
		return this.dTemplate.query( Q_EMPIRE_RESOURCES , this.mEmpireResource , empire );
	}


	/**
	 * Run the planet list resources query and extract the data
	 * 
	 * <p>
	 * This implementation queries <code>emp.plist_resources_view</code> to obtain the planet list's
	 * records, then maps the resulting rows using the intermediary {@link FullPlanetListRecord}
	 * class, and finally associates records to the corresponding planet identifiers.
	 */
	@Override
	public Map< Integer , List< PlanetListResourceRecord > > getPlanetListData( int empire )
	{
		Map< Integer , List< PlanetListResourceRecord > > result;
		result = new HashMap< Integer , List< PlanetListResourceRecord > >( );

		List< FullPlanetListRecord > queryOutput;
		queryOutput = this.dTemplate.query( Q_PLIST_RESOURCES , this.mPlanetListRecord , empire );

		for ( FullPlanetListRecord fullRecord : queryOutput ) {
			List< PlanetListResourceRecord > records = result.get( fullRecord.getPlanetId( ) );
			if ( records == null ) {
				records = new LinkedList< PlanetListResourceRecord >( );
				result.put( fullRecord.getPlanetId( ) , records );
			}
			records.add( fullRecord.getRecord( ) );
		}

		return result;
	}

}
