package com.deepclone.lw.beans.msgs;


import java.util.List;
import java.util.Map;

import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cmd.msgdata.Message;
import com.deepclone.lw.cmd.msgdata.MessageType;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.i18n.TranslationException;
import com.deepclone.lw.interfaces.i18n.Translator;
import com.deepclone.lw.interfaces.mailer.Mailer;
import com.deepclone.lw.interfaces.msg.MessageContentCache;
import com.deepclone.lw.interfaces.msg.MessageExtractor;
import com.deepclone.lw.interfaces.msg.MessageFormatRegistry;
import com.deepclone.lw.interfaces.msg.NotificationsDAO;
import com.deepclone.lw.sqld.msgs.InboxRecord;
import com.deepclone.lw.sqld.msgs.MessageDataRecord;
import com.deepclone.lw.sqld.msgs.NotificationsRecord;



abstract class MailTaskBase
		implements Runnable
{

	protected static class MessageData
	{
		public List< InboxRecord > messages;
		public Map< Long , MessageDataRecord > data;
	}

	protected final SystemLogger logger;
	protected final TransactionTemplate tTemplate;
	protected final NotificationsDAO notificationsDao;
	protected final MessageContentCache contents;
	protected final MessageFormatRegistry formats;
	protected final Translator translator;
	protected final Mailer mailer;
	protected final MessageCleanerBean cleaner;
	protected final boolean mode;


	public MailTaskBase( Logger logger , PlatformTransactionManager tManager , NotificationsDAO notificationsDao ,
			MessageContentCache contents , MessageFormatRegistry formats , Translator translator , Mailer mailer ,
			MessageCleanerBean cleaner , boolean mode )
	{
		this.logger = logger.getSystemLogger( this.getClass( ).getSimpleName( ) );
		this.tTemplate = new TransactionTemplate( tManager );
		this.notificationsDao = notificationsDao;
		this.contents = contents;
		this.formats = formats;
		this.translator = translator;
		this.mailer = mailer;
		this.cleaner = cleaner;
		this.mode = mode;
	}


	@Override
	public void run( )
	{
		List< NotificationsRecord > notifications = this.getNotifications( );
		for ( NotificationsRecord notification : notifications ) {
			this.notify( notification );
		}
	}


	private List< NotificationsRecord > getNotifications( )
	{
		return this.tTemplate.execute( new TransactionCallback< List< NotificationsRecord > >( ) {
			@Override
			public List< NotificationsRecord > doInTransaction( TransactionStatus status )
			{
				return notificationsDao.getNotificationRecords( mode );
			}
		} );
	}


	protected abstract void notify( NotificationsRecord notification );


	protected final MessageData getMessagesToSend( final int id , final long maxId , final boolean nPrivate ,
			final boolean nInternal , final boolean nAlliance , final boolean nAdmin )
	{
		return this.tTemplate.execute( new TransactionCallback< MessageData >( ) {
			@Override
			public MessageData doInTransaction( TransactionStatus status )
			{
				MessageData md = new MessageData( );
				md.messages = notificationsDao.listMessages( id , maxId , mode , nPrivate , nInternal , nAlliance ,
						nAdmin );
				if ( !md.messages.isEmpty( ) ) {
					md.data = contents.getContent( md.messages );
				}
				return md;
			}
		} );
	}


	protected final String translate( String string , String lang )
	{
		try {
			return this.translator.translate( lang , string );
		} catch ( TranslationException e ) {
			throw new RuntimeException( e );
		}
	}


	protected final Message readMessage( int empireId , String lang , InboxRecord record , MessageDataRecord content ,
			boolean needBody )
	{
		MessageExtractor formatter = this.formats.getFormatter( content ).getExtractor( record , content , lang );

		// Extract basic information
		Message message = new Message( );
		message.setTime( record.getReceived( ) );
		message.setSender( formatter.getSender( ) );
		message.setReceiver( record.getReceiverName( ) );
		message.setTitle( formatter.getSubject( ) );
		if ( needBody ) {
			message.setContents( formatter.getContents( ) );
		}
		message.setUnread( !record.isRead( ) );
		message.setGameTime( content.getTick( ) );

		// Handle message type
		if ( record.isInternal( ) ) {
			message.setType( MessageType.INTERNAL );
		} else {
			String type = record.getSenderType( );
			if ( "ALL".equals( record.getReceiverType( ) ) ) {
				message.setType( MessageType.ALLIANCE );
				message.setReceiver( "[" + message.getReceiver( ) + "]" );
			} else if ( "EMP".equals( type ) ) {
				message.setType( MessageType.EMPIRE );
			} else if ( "ADM".equals( type ) ) {
				message.setType( MessageType.ADMINISTRATOR );
			} else {
				throw new RuntimeException( "unknown message type " + type );
			}
		}

		return message;
	}

}
