package com.deepclone.lw.beans.sys;


import java.lang.ref.WeakReference;
import java.net.InetAddress;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;

import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.cmd.admin.users.SessionTerminationType;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.session.*;
import com.deepclone.lw.interfaces.sys.Ticker;
import com.deepclone.lw.session.*;
import com.deepclone.lw.utils.RandomStringGenerator;



/**
 * Session management bean implementation.
 * 
 * <p>
 * This class is the implementation of the session management bean. It is responsible for creating
 * sessions, using its collection of type definers, for accessing the sessions, and for making them
 * expire as required.
 * 
 * @author tseeker
 */
public class SessionManagerBean
		implements SessionManager , DisposableBean
{

	/** Session type definers, by name */
	private final Map< String , WeakReference< SessionTypeDefiner > > types = new HashMap< String , WeakReference< SessionTypeDefiner > >( );

	/** Sessions, by identifier */
	private final Map< String , ServerSessionData > sessions = new HashMap< String , ServerSessionData >( );

	/** Random string generator for session identifiers */
	private RandomStringGenerator sessionIDGenerator;

	/** Random string generator for authentication challenges */
	private RandomStringGenerator challengeGenerator;

	/** Session clean-up task, registered to the {@link Ticker} */
	private Runnable cleanupTask;

	/** System logger for the session manager */
	private SystemLogger logger;


	/**
	 * Sets the generator for session identifiers (DI)
	 * 
	 * @param rsg
	 *            reference to the random string generator to use for session identifiers
	 */
	@Autowired( required = true )
	@Qualifier( "sessionIdentifiers" )
	public void setSessionIDGenerator( RandomStringGenerator rsg )
	{
		this.sessionIDGenerator = rsg;
	}


	/**
	 * Sets the generator for authentication challenges (DI)
	 * 
	 * @param rsg
	 *            reference to the random string generator to use for authentication challenges
	 */
	@Autowired( required = true )
	@Qualifier( "authChallenges" )
	public void setChallangeGenerator( RandomStringGenerator rsg )
	{
		this.challengeGenerator = rsg;
	}


	/**
	 * Initialises the session clean-up task (DI)
	 * 
	 * @param ticker
	 *            reference to the ticker bean
	 */
	@Autowired( required = true )
	public void setTicker( Ticker ticker )
	{
		this.cleanupTask = new Runnable( ) {

			@Override
			public void run( )
			{
				purgeSessions( );
			}

		};
		ticker.registerTask( Ticker.Frequency.HIGH , "Sessions clean-up" , this.cleanupTask );
	}


	/**
	 * Initialises the system logger (DI)
	 * 
	 * @param logger
	 *            reference to the logger bean
	 */
	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.logger = logger.getSystemLogger( "SessionManager" );
	}


	/**
	 * Removes the reference to the clean-up task.
	 */
	@Override
	public void destroy( )
	{
		this.cleanupTask = null;
		synchronized ( this.sessions ) {
			for ( ServerSessionData data : this.sessions.values( ) ) {
				try {
					this.logger.log( LogLevel.DEBUG , "expiring '" + data.getIdentifier( ) + "'" ).flush( );
					data.handleTermination( SessionTerminationType.EXPIRED );
				} catch ( RuntimeException e ) {
					this.logger.log( LogLevel.ERROR , "error while terminating '" + data.getIdentifier( ) + "'" , e )
							.flush( );
				}
			}
		}
	}


	/**
	 * Finds the session type definer corresponding to a type name.
	 * 
	 * @param type
	 *            the name of the session type
	 * @return the session type definer, or <code>null</code> if it doesn't exist
	 */
	private SessionTypeDefiner getTypeDefiner( String type )
	{
		SessionTypeDefiner definer = null;

		synchronized ( this.types ) {
			WeakReference< SessionTypeDefiner > ref = this.types.get( type );
			if ( ref != null ) {
				definer = ref.get( );
				if ( definer == null ) {
					this.types.remove( type );
				}
			}
		}

		return definer;
	}


	/**
	 * Allocates a session identifier.
	 * 
	 * <p>
	 * This method generates an unique session identifier for the specified session, setting the
	 * session's identifier and storing it in the {@link #sessions} map.
	 * 
	 * @param data
	 *            the session that needs to be initialised
	 */
	private void allocateSessionIdentifier( ServerSessionData data )
	{
		synchronized ( this.sessions ) {
			String id;
			do {
				id = this.sessionIDGenerator.generate( );
			} while ( this.sessions.containsKey( id ) );
			data.setIdentifier( id );
			this.sessions.put( id , data );
		}
	}


	/**
	 * Retrieves a session from an identifier.
	 * 
	 * <p>
	 * This method attempts to retrieve a session from the {@link #sessions} map. If the session
	 * exists but is expired, it will be removed and the method will behave as if it didn't exist.
	 * 
	 * @param id
	 *            the session's identifier
	 * @return the session's data
	 * @throws SessionIdentifierException
	 *             if the session doesn't exist or has expired.
	 */
	private ServerSessionData getSession( String id )
			throws SessionIdentifierException
	{
		ServerSessionData data;
		synchronized ( this.sessions ) {
			data = this.sessions.get( id );
			if ( data == null ) {
				this.logger.log( LogLevel.INFO , "session '" + id + "' not found" ).flush( );
				throw new SessionIdentifierException( id );
			}
			if ( data.getExpirationDate( ).before( new Date( ) ) ) {
				this.logger.log( LogLevel.INFO , "session '" + id + "' found but expired" ).flush( );
				ServerSessionData session = this.sessions.remove( id );
				session.handleTermination( SessionTerminationType.EXPIRED );
				throw new SessionIdentifierException( id );
			}
		}
		return data;
	}


	/**
	 * Removes expired sessions.
	 * 
	 * <p>
	 * This method goes through all registered sessions, removing them from the {@link #sessions}
	 * map if they have expired.
	 */
	private void purgeSessions( )
	{
		Date now = new Date( );
		List< ServerSessionData > toRemove = new LinkedList< ServerSessionData >( );
		synchronized ( this.sessions ) {
			for ( ServerSessionData session : this.sessions.values( ) ) {
				if ( session.getExpirationDate( ).before( now ) ) {
					toRemove.add( session );
				}
			}

			for ( ServerSessionData session : toRemove ) {
				this.sessions.remove( session.getIdentifier( ) );
			}
		}

		for ( ServerSessionData session : toRemove ) {
			this.handleTermination( session , SessionTerminationType.EXPIRED );
		}

		if ( !toRemove.isEmpty( ) ) {
			this.logger.log( LogLevel.TRACE , toRemove.size( ) + " session(s) removed" ).flush( );
		}
	}


	private void handleTermination( ServerSessionData session , SessionTerminationType reason )
	{
		try {
			session.handleTermination( reason );
		} catch ( RuntimeException e ) {
			this.logger.log( LogLevel.ERROR , "error while terminating '" + session.getIdentifier( ) + "'" , e )
					.flush( );
		}
	}


	/* Documented in SessionManager interface */
	@Override
	public void registerSessionType( SessionTypeDefiner definer )
	{
		synchronized ( this.types ) {
			this.types.put( definer.getName( ) , new WeakReference< SessionTypeDefiner >( definer ) );
		}
		this.logger.log( LogLevel.INFO , "registered session type definer '" + definer.getName( ) + "'" ).flush( );
	}


	/* Documented in SessionAccessor interface */
	@Override
	public SessionReference create( String type , String client , InetAddress address )
			throws SessionInternalException
	{
		SessionTypeDefiner definer = this.getTypeDefiner( type );
		if ( definer == null ) {
			this.logger.log( LogLevel.WARNING , "no such session type '" + type + "'" ).flush( );
			return null;
		}

		ServerSessionData data;
		try {
			data = new ServerSessionData( this.challengeGenerator , definer , client , address );
		} catch ( RuntimeException e ) {
			this.logger.log( LogLevel.ERROR , "error while creating session with type '" + type + "'" , e ).flush( );
			throw new SessionInternalException( false , e );
		}

		this.allocateSessionIdentifier( data );
		this.logger.log( LogLevel.INFO , "session " + data.getIdentifier( ) + " created (type " + type + ")" ).flush( );

		return data.toSessionReference( );
	}


	/* Documented in SessionAccessor interface */
	@Override
	public SessionReference authenticate( String session , String identifier , String sha1Hash , String md5Hash )
			throws SessionIdentifierException , SessionStateException , SessionInternalException
	{
		ServerSessionData data = this.getSession( session );
		this.logger.log( LogLevel.TRACE , "attempting to authenticate session '" + session + "'" ).flush( );

		try {
			data.authenticate( identifier , sha1Hash , md5Hash );
		} catch ( SessionIdentifierException e ) {
			synchronized ( this.sessions ) {
				this.sessions.remove( session );
			}
			throw e;
		} catch ( RuntimeException e ) {
			this.logger.log( LogLevel.ERROR , "error while authenticating session '" + session + "'" , e ).flush( );
			throw new SessionInternalException( false , e );
		}

		if ( data.getTerminationType( ) != null ) {
			synchronized ( this.sessions ) {
				this.logger.log( LogLevel.INFO , "terminated '" + session + "' during authentication" ).flush( );
				this.sessions.remove( session );
			}
		}

		return data.toSessionReference( );
	}


	/* Documented in SessionAccessor interface */
	@Override
	public SessionResponse executeCommand( String session , Command command )
			throws SessionIdentifierException , SessionStateException , SessionCommandException ,
			SessionInternalException
	{
		ServerSessionData data = this.getSession( session );
		this.logger.log( LogLevel.DEBUG , "executing command " + command.getClass( ) + " on '" + session + "'" )
				.flush( );

		CommandResponse response;
		try {
			response = data.execute( command );
		} catch ( SessionIdentifierException e ) {
			synchronized ( this.sessions ) {
				this.sessions.remove( session );
			}
			throw e;
		} catch ( RuntimeException e ) {
			this.logger.log( LogLevel.ERROR , "error while executing command on '" + session + "'" , e ).flush( );
			throw new SessionInternalException( false , e );
		}

		if ( data.getTerminationType( ) != null ) {
			this.logger.log( LogLevel.INFO , "terminated '" + session + "' during command execution" ).flush( );
			synchronized ( this.sessions ) {
				this.sessions.remove( session );
			}
		}

		return new SessionResponse( data.toSessionReference( ) , response );
	}


	/* Documented in SessionAccessor interface */
	@Override
	public void terminate( String session )
			throws SessionIdentifierException , SessionInternalException
	{
		ServerSessionData data;
		synchronized ( this.sessions ) {
			data = this.sessions.remove( session );
		}
		if ( data == null ) {
			throw new SessionIdentifierException( session );
		} else if ( data.getExpirationDate( ).before( new Date( ) ) ) {
			data.handleTermination( SessionTerminationType.EXPIRED );
			throw new SessionIdentifierException( session );
		}

		try {
			data.handleTermination( SessionTerminationType.MANUAL );
			this.logger.log( LogLevel.INFO , "terminated '" + session + "'" ).flush( );
		} catch ( RuntimeException e ) {
			this.logger.log( LogLevel.ERROR , "error while terminating '" + session + "'" , e ).flush( );
			throw new SessionInternalException( false , e );
		}
	}
}
