package com.deepclone.lw.beans.sys;


import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Map;

import javax.sql.DataSource;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.SqlOutParameter;
import org.springframework.jdbc.core.simple.SimpleJdbcCall;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.interfaces.sys.MaintenanceData;
import com.deepclone.lw.interfaces.sys.MaintenanceStatusException;
import com.deepclone.lw.interfaces.sys.SystemStatus;
import com.deepclone.lw.interfaces.sys.TickStatusException;
import com.deepclone.lw.sqld.sys.Status;
import com.deepclone.lw.utils.StoredProc;



public class SystemStatusBean
		implements SystemStatus
{

	/** Database interface */
	private JdbcTemplate dTemplate;

	/** Transaction template */
	private TransactionTemplate tTemplate;

	/** System status record */
	private Status status = null;

	/** Current maintenance mode record */
	private MaintenanceData maintenance = null;

	private SimpleJdbcCall doStartTick;
	private SimpleJdbcCall doCheckTick;

	private StoredProc fEnterMaintenanceMode;
	private StoredProc fExtendMaintenanceMode;
	private StoredProc fExitMaintenanceMode;


	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );

		this.fEnterMaintenanceMode = new StoredProc( dataSource , "sys" , "enter_maintenance_mode" );
		this.fEnterMaintenanceMode.addParameter( "admin_id" , Types.INTEGER );
		this.fEnterMaintenanceMode.addParameter( "reason" , Types.VARCHAR );
		this.fEnterMaintenanceMode.addParameter( "duration" , Types.INTEGER );
		this.fEnterMaintenanceMode.addOutput( "success" , Types.BOOLEAN );

		this.fExtendMaintenanceMode = new StoredProc( dataSource , "sys" , "extend_maintenance_mode" );
		this.fExtendMaintenanceMode.addParameter( "admin_id" , Types.INTEGER );
		this.fExtendMaintenanceMode.addParameter( "duration" , Types.INTEGER );
		this.fExtendMaintenanceMode.addOutput( "success" , Types.BOOLEAN );

		this.fExitMaintenanceMode = new StoredProc( dataSource , "sys" , "exit_maintenance_mode" );
		this.fExitMaintenanceMode.addParameter( "admin_id" , Types.INTEGER );
		this.fExitMaintenanceMode.addOutput( "success" , Types.BOOLEAN );

		this.doStartTick = new SimpleJdbcCall( dataSource );
		this.doStartTick.withCatalogName( "sys" ).withFunctionName( "start_tick" );
		this.doStartTick.withoutProcedureColumnMetaDataAccess( );
		this.doStartTick.addDeclaredParameter( new SqlOutParameter( "tick_id" , Types.BIGINT ) );

		this.doCheckTick = new SimpleJdbcCall( dataSource );
		this.doCheckTick.withCatalogName( "sys" ).withFunctionName( "check_stuck_tick" );
		this.doCheckTick.withoutProcedureColumnMetaDataAccess( );
		this.doCheckTick.addDeclaredParameter( new SqlOutParameter( "tick_id" , Types.BIGINT ) );

	}


	/**
	 * Sets the transaction manager interface (DI)
	 * 
	 * @param transactionManager
	 *            the transaction manager
	 */
	@Autowired( required = true )
	public void setTransactionManager( PlatformTransactionManager transactionManager )
	{
		this.tTemplate = new TransactionTemplate( transactionManager );
	}


	private void loadStatus( )
	{
		String sql = "SELECT * FROM sys.status";
		RowMapper< Status > mapper = new RowMapper< Status >( ) {

			@Override
			public Status mapRow( ResultSet rs , int rowNum )
					throws SQLException
			{
				Status s = new Status( );
				s.setNextTickIdentifier( rs.getLong( "next_tick" ) );
				s.setCurrentTick( rs.getLong( "current_tick" ) );
				s.setLastMsgRecap( rs.getTimestamp( "last_msg_recap" ) );
				s.setLastAdminRecap( rs.getTimestamp( "last_admin_recap" ) );
				s.setLastErrorCheck( rs.getTimestamp( "last_error_recap" ) );
				s.setMaintenanceStart( rs.getTimestamp( "maintenance_start" ) );
				s.setMaintenanceEnd( rs.getTimestamp( "maintenance_end" ) );
				s.setMaintenanceReason( rs.getString( "maintenance_text" ) );
				return s;
			}

		};

		this.status = this.dTemplate.queryForObject( sql , mapper );

		// Update maintenance status
		if ( this.status.getMaintenanceReason( ) != null ) {
			this.maintenance = new MaintenanceData( this.status.getMaintenanceStart( ) , this.status
					.getMaintenanceEnd( ) , this.status.getMaintenanceReason( ) );
		} else {
			this.maintenance = null;
		}
	}


	/**
	 * Initialises the system's status.
	 * 
	 * <p>
	 * This method attempts to read the system's previous status from the database. If the entry
	 * doesn't exist, it is created. Some of the record's data is then stored locally and the bean
	 * is marked as initialised.
	 */
	private void initialise( )
	{
		if ( this.status != null ) {
			return;
		}

		// Load status
		this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {

			@Override
			protected void doInTransactionWithoutResult( TransactionStatus status )
			{
				loadStatus( );
			}

		} );
	}


	/* Documented in interface */
	@Override
	synchronized public MaintenanceData checkMaintenance( )
	{
		this.initialise( );
		return this.maintenance;
	}


	/* Documented in interface */
	@Override
	synchronized public void startMaintenance( final int adminId , final String reason , final int duration )
			throws MaintenanceStatusException
	{
		if ( duration <= 0 || reason == null ) {
			throw new IllegalArgumentException( );
		}

		boolean s = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				Map< String , Object > m = fEnterMaintenanceMode.execute( adminId , reason , duration );
				loadStatus( );
				return (Boolean) m.get( "success" );
			}

		} );

		if ( !s ) {
			throw new MaintenanceStatusException( this.maintenance );
		}
	}


	/* Documented in interface */
	@Override
	synchronized public void updateMaintenance( final int adminId , final int durationFromNow )
			throws MaintenanceStatusException
	{
		if ( durationFromNow <= 0 ) {
			throw new IllegalArgumentException( );
		}

		boolean s = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				Map< String , Object > m = fExtendMaintenanceMode.execute( adminId , durationFromNow );
				loadStatus( );
				return (Boolean) m.get( "success" );
			}

		} );

		if ( !s ) {
			throw new MaintenanceStatusException( );
		}
	}


	/* Documented in interface */
	@Override
	synchronized public void endMaintenance( final int adminId )
			throws MaintenanceStatusException
	{
		boolean s = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				Map< String , Object > m = fExitMaintenanceMode.execute( adminId );
				loadStatus( );
				return (Boolean) m.get( "success" );
			}

		} );

		if ( !s ) {
			throw new MaintenanceStatusException( );
		}
	}


	/* Documented in interface */
	@Override
	synchronized public long startTick( )
			throws TickStatusException , MaintenanceStatusException
	{
		Long tid = this.tTemplate.execute( new TransactionCallback< Long >( ) {

			@Override
			public Long doInTransaction( TransactionStatus status )
			{
				Map< String , Object > m = doStartTick.execute( );
				loadStatus( );
				return (Long) m.get( "tick_id" );
			}

		} );

		if ( tid == null ) {
			if ( this.maintenance != null ) {
				throw new MaintenanceStatusException( this.maintenance );
			} else {
				throw new TickStatusException( this.status.getCurrentTick( ) );
			}
		}

		return tid;
	}


	/* Documented in interface */
	@Override
	public Long checkStuckTick( )
			throws MaintenanceStatusException
	{
		Long tid = this.tTemplate.execute( new TransactionCallback< Long >( ) {

			@Override
			public Long doInTransaction( TransactionStatus status )
			{
				Map< String , Object > m = doCheckTick.execute( );
				loadStatus( );
				return (Long) m.get( "tick_id" );
			}

		} );

		if ( tid == null && this.maintenance != null ) {
			throw new MaintenanceStatusException( this.maintenance );
		}

		return tid;
	}

}
