package com.deepclone.lw.beans.sys;


import java.util.HashSet;
import java.util.Set;

import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;

import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.sys.ConstantsManager;
import com.deepclone.lw.interfaces.sys.Ticker;



/**
 * Task scheduler bean.
 * 
 * <p>
 * This class implements Legacy Worlds' task scheduler, used for most repetitive actions such as
 * updating the game's state, cleaning logs or de-activating accounts.
 * 
 * <ul>
 * <li>Tasks with {@link Frequency#HIGH} are executed once every time the period is over.</li>
 * <li>Tasks with {@link Frequency#MEDIUM} are executed once every 6 periods.</li>
 * <li>Tasks with {@link Frequency#MINUTE} are executed once every 12 periods.</li>
 * <li>Tasks with {@link Frequency#LOW} are executed once every 60 periods.</li>
 * </ul>
 * 
 * <p>
 * <em>ticker.interval</em> defaults to 5 seconds.
 * 
 * @author tseeker
 */
public class TickerBean
		implements Ticker , InitializingBean , DisposableBean
{

	/** System logger for the bean */
	private SystemLogger logger;

	/** Constants manager bean */
	private ConstantsManager constantsManager;

	/** Main thread for ticker control */
	private TickerThread mainThread;

	private TickerTaskStatusHandler tickerManager;


	/**
	 * Sets the system logger (DI)
	 * 
	 * @param logger
	 *            reference to the logger bean
	 */
	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.logger = logger.getSystemLogger( "Ticker" );
	}


	/**
	 * Sets the constants manager (DI)
	 * 
	 * @param manager
	 *            reference to the constants manager bean
	 */
	@Autowired( required = true )
	public void setConstantsManager( ConstantsManager manager )
	{
		this.constantsManager = manager;
	}


	@Autowired( required = true )
	public void setManager( TickerTaskStatusHandler manager )
	{
		this.tickerManager = manager;
	}


	/**
	 * Initialises the bean.
	 * 
	 * <p>
	 * When all dependencies have been set, the bean will register the <em>ticker.interval</em>
	 * system constant. The main control thread will then be created and registered as an user of
	 * the constant (which causes the control thread to schedule itself using the timer).
	 */
	@Override
	public void afterPropertiesSet( )
	{
		this.logger.log( LogLevel.INFO , "Initialisation" ).flush( );

		// Create thread
		this.mainThread = new TickerThread( this.logger , this.tickerManager );

		// Register thread as a constants user
		Set< String > use = new HashSet< String >( );
		use.add( "ticker.interval" );
		this.constantsManager.registerUser( this.mainThread , use );
	}


	/**
	 * Destroys the bean.
	 * 
	 * This method will abort the main control thread's execution, and unregister it from the
	 * constants manager.
	 */
	@Override
	public void destroy( )
	{
		this.logger.log( LogLevel.INFO , "Destruction" ).flush( );
		this.mainThread.terminate( );
		this.constantsManager.unregisterUser( this.mainThread );
		this.constantsManager = null;
	}


	/* Documented in Ticker interface */
	@Override
	public void registerTask( Ticker.Frequency frequency , String name , Runnable task )
	{
		this.logger.log( LogLevel.DEBUG , "Registering task " + name + " at frequency " + frequency ).flush( );
		int id = this.tickerManager.registerTask( name );
		this.mainThread.registerTask( id , frequency , name , task );
	}


	/* Documented in Ticker interface */
	@Override
	public void pause( )
			throws IllegalStateException
	{
		this.mainThread.pause( );
	}


	/* Documented in Ticker interface */
	@Override
	public void unpause( )
			throws IllegalStateException
	{
		this.mainThread.unpause( );
	}


	/* Documented in Ticker interface */
	@Override
	public boolean isActive( )
	{
		return this.mainThread.isActive( );
	}
}
