package com.deepclone.lw.beans.sys;


import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionCallbackWithoutResult;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cmd.admin.tick.TickerTaskInfo;
import com.deepclone.lw.cmd.admin.tick.TickerTaskStatus;
import com.deepclone.lw.interfaces.sys.TickerManager;
import com.deepclone.lw.sqld.sys.TickerTaskRecord;
import com.deepclone.lw.utils.StoredProc;



public class TickerManagerBean
		implements TickerManager , TickerTaskStatusHandler , InitializingBean
{

	private final Map< Integer , TickerTaskRecord > tasks = new HashMap< Integer , TickerTaskRecord >( );
	private final Map< String , Integer > taskIds = new HashMap< String , Integer >( );
	private final Set< Integer > registered = new HashSet< Integer >( );

	private TransactionTemplate tTemplate;
	private JdbcTemplate dTemplate;

	private final RowMapper< TickerTaskRecord > mTask;

	private StoredProc fRegisterTask;
	private StoredProc fSetTaskStarted;
	private StoredProc fSetTaskRunning;
	private StoredProc fScheduleTask;


	public TickerManagerBean( )
	{
		this.mTask = new RowMapper< TickerTaskRecord >( ) {
			@Override
			public TickerTaskRecord mapRow( ResultSet rs , int rowNum )
					throws SQLException
			{
				TickerTaskRecord task = new TickerTaskRecord( );
				task.setId( rs.getInt( "id" ) );
				task.setName( rs.getString( "task_name" ) );
				task.setStatus( rs.getString( "status" ) );
				task.setTimestamp( rs.getTimestamp( "auto_start" ) );
				return task;
			}
		};
	}


	@Autowired( required = true )
	public void setTransactionManager( PlatformTransactionManager manager )
	{
		this.tTemplate = new TransactionTemplate( manager );
	}


	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );

		this.fRegisterTask = new StoredProc( dataSource , "sys" , "register_ticker_task" );
		this.fRegisterTask.addParameter( "task_name" , Types.VARCHAR );
		this.fRegisterTask.addOutput( "id" , Types.INTEGER );

		this.fSetTaskStarted = new StoredProc( dataSource , "sys" , "set_task_started" );
		this.fSetTaskStarted.addParameter( "id" , Types.INTEGER );

		this.fSetTaskRunning = new StoredProc( dataSource , "sys" , "set_task_running" );
		this.fSetTaskRunning.addParameter( "admin_id" , Types.INTEGER );
		this.fSetTaskRunning.addParameter( "task_id" , Types.INTEGER );
		this.fSetTaskRunning.addParameter( "running" , Types.BOOLEAN );

		this.fScheduleTask = new StoredProc( dataSource , "sys" , "schedule_task" );
		this.fScheduleTask.addParameter( "admin_id" , Types.INTEGER );
		this.fScheduleTask.addParameter( "task_id" , Types.INTEGER );
		this.fScheduleTask.addParameter( "time_to_start" , Types.BIGINT );
		this.fScheduleTask.addOutput( "start_at" , Types.TIMESTAMP );
	}


	private List< TickerTaskRecord > getDBTasks( )
	{
		return this.tTemplate.execute( new TransactionCallback< List< TickerTaskRecord > >( ) {
			@Override
			public List< TickerTaskRecord > doInTransaction( TransactionStatus status )
			{
				return dTemplate.query( "SELECT * FROM sys.ticker" , mTask );
			}
		} );
	}


	private void checkAutoStart( final TickerTaskRecord record )
	{
		if ( record.getStatus( ).equals( "AUTO" ) ) {
			Timestamp start = record.getTimestamp( );
			if ( start.before( new Date( ) ) ) {
				this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {
					@Override
					protected void doInTransactionWithoutResult( TransactionStatus status )
					{
						fSetTaskStarted.execute( record.getId( ) );
					}
				} );
				record.setStatus( "RUNNING" );
				record.setTimestamp( null );
			}
		}
	}


	private void setTaskRunning( final int administrator , final TickerTaskRecord record , final boolean running )
	{
		this.tTemplate.execute( new TransactionCallbackWithoutResult( ) {
			@Override
			protected void doInTransactionWithoutResult( TransactionStatus status )
			{
				fSetTaskRunning.execute( administrator , record.getId( ) , running );
			}
		} );
		record.setStatus( running ? "RUNNING" : "STOPPED" );
		record.setTimestamp( null );
	}


	@Override
	public void afterPropertiesSet( )
	{
		List< TickerTaskRecord > rTasks = this.getDBTasks( );
		for ( TickerTaskRecord task : rTasks ) {
			this.tasks.put( task.getId( ) , task );
			this.taskIds.put( task.getName( ) , task.getId( ) );
		}
	}


	@Override
	synchronized public List< TickerTaskInfo > getTasks( )
	{
		List< TickerTaskInfo > result = new LinkedList< TickerTaskInfo >( );
		long now = new Date( ).getTime( );
		for ( Integer id : this.registered ) {
			TickerTaskRecord record = this.tasks.get( id );
			this.checkAutoStart( record );

			TickerTaskInfo info = new TickerTaskInfo( );
			info.setId( record.getId( ) );
			info.setName( record.getName( ) );
			info.setStatus( TickerTaskStatus.valueOf( record.getStatus( ) ) );
			if ( info.getStatus( ) == TickerTaskStatus.AUTO ) {
				info.setStart( record.getTimestamp( ) );
				info.setTimeToStart( ( record.getTimestamp( ).getTime( ) - now ) / 1000 );
			}
			result.add( info );
		}
		return result;
	}


	@Override
	synchronized public void startTask( int administrator , int id )
	{
		TickerTaskRecord record = this.tasks.get( id );
		if ( record == null || record.getStatus( ).equals( "RUNNING" ) ) {
			return;
		}
		this.setTaskRunning( administrator , record , true );
	}


	@Override
	synchronized public void stopTask( int administrator , int id )
	{
		TickerTaskRecord record = this.tasks.get( id );
		if ( record == null || record.getStatus( ).equals( "STOPPED" ) ) {
			return;
		}
		this.setTaskRunning( administrator , record , false );
	}


	@Override
	synchronized public void setTaskStart( final int administrator , final int id , final long time )
	{
		if ( time <= 0 ) {
			this.startTask( administrator , id );
			return;
		}

		TickerTaskRecord record = this.tasks.get( id );
		if ( record == null ) {
			return;
		}

		Timestamp startAt = this.tTemplate.execute( new TransactionCallback< Timestamp >( ) {
			@Override
			public Timestamp doInTransaction( TransactionStatus status )
			{
				return (Timestamp) fScheduleTask.execute( administrator , id , time ).get( "start_at" );
			}
		} );
		record.setStatus( "AUTO" );
		record.setTimestamp( startAt );
	}


	@Override
	synchronized public int registerTask( final String name )
	{
		Integer eId = this.taskIds.get( name );
		if ( eId != null ) {
			this.registered.add( eId );
			return eId;
		}

		// Register new task
		eId = this.tTemplate.execute( new TransactionCallback< Integer >( ) {
			@Override
			public Integer doInTransaction( TransactionStatus status )
			{
				return (Integer) fRegisterTask.execute( name ).get( "id" );
			}
		} );

		TickerTaskRecord record = new TickerTaskRecord( );
		record.setId( eId );
		record.setName( name );
		record.setStatus( "RUNNING" );
		record.setTimestamp( null );
		this.tasks.put( eId , record );
		this.taskIds.put( name , eId );

		return eId;
	}


	@Override
	synchronized public boolean isTaskRunning( int id )
	{
		TickerTaskRecord record = this.tasks.get( id );
		this.checkAutoStart( record );
		return ( record.getStatus( ).equals( "RUNNING" ) );
	}

}
