package com.deepclone.lw.beans.sys;


import java.lang.ref.WeakReference;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;



/**
 * Ticker task execution class.
 * 
 * <p>
 * This class is responsible for both the execution and the control of ticker tasks.
 * 
 * @author tseeker
 */
class TickerTask
		extends Thread
{
	/** Possible states of a task execution thread. */
	private static enum State {

		/** The task thread is being initialised. */
		STARTING ,

		/** The task thread is waiting for instructions */
		WAITING ,

		/** The task is currently being executed */
		RUNNING ,

		/** The task thread is exiting. */
		EXITING ,

		/** The task thread is finished. */
		FINISHED
	}

	/** Reference to the system logger */
	private final SystemLogger logger;

	/** Name of the task */
	private final String name;

	/** Weak reference to the task's code */
	private final WeakReference< Runnable > task;

	/** Lock used to protect the task thread's state */
	private final Lock lock = new ReentrantLock( );

	/** Condition triggered whenever the thread's state changes */
	private final Condition stateChanged = this.lock.newCondition( );

	/** The thread's current state */
	private State state = State.STARTING;


	/**
	 * @param name
	 *            the task's name
	 * @param task
	 *            reference to the task's code
	 * @param logger
	 *            system logger instance
	 */
	TickerTask( String name , Runnable task , SystemLogger logger )
	{
		super( "Ticker thread for '" + name + "'" );
		this.name = name;
		this.task = new WeakReference< Runnable >( task );
		this.logger = logger;
	}


	/**
	 * Starts the task execution thread.
	 * 
	 * <p>
	 * Starts the thread itself, then waits for the thread's state to change from its default
	 * {@link State#STARTING} value.
	 */
	@Override
	public void start( )
	{
		super.start( );
		this.lock.lock( );
		try {
			State state = this.getTaskState( );
			while ( state == State.STARTING ) {
				state = this.waitTaskState( );
			}
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * Retrieves the task execution thread's state.
	 * 
	 * <p>
	 * This method retrieves the task execution thread's state, locking {@link #lock} to prevent
	 * concurrent access.
	 * 
	 * @return the task execution thread's current state.
	 */
	private State getTaskState( )
	{
		this.lock.lock( );
		try {
			return this.state;
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * Changes the task execution thread's state.
	 * 
	 * <p>
	 * This method modifies the thread's state, locking {@link #lock} to prevent concurrent access
	 * and triggering {@link #stateChanged} once it's done.
	 * 
	 * @param newState
	 *            the thread's new state.
	 */
	private void setTaskState( State newState )
	{
		this.lock.lock( );
		try {
			this.state = newState;
			this.stateChanged.signal( );
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * Waits for a change of state.
	 * 
	 * <p>
	 * This method waits for the task's state to change using {@link #stateChanged}, then returns
	 * the new state.
	 * 
	 * @return the task execution thread's new state.
	 */
	private State waitTaskState( )
	{
		this.lock.lock( );
		try {
			this.stateChanged.await( );
			return this.state;
		} catch ( InterruptedException e ) {
			return this.state;
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * Task execution loop.
	 * 
	 * <p>
	 * This method makes sure that the task is executed every time the state is set to
	 * {@link State#RUNNING}, setting it to {@link State#WAITING} when it is inactive.
	 * 
	 * <p>
	 * If the thread's state is set to {@link State#EXITING}, or if the task code is no longer
	 * referenced, the method will exit, setting the state to {@link State#FINISHED}.
	 */
	@Override
	public void run( )
	{
		while ( true ) {
			State state;

			this.lock.lock( );
			try {
				if ( this.getTaskState( ) == State.EXITING ) {
					break;
				}
				this.setTaskState( State.WAITING );
				state = this.waitTaskState( );
			} finally {
				this.lock.unlock( );
			}

			if ( state == State.EXITING || !this.runTask( ) ) {
				break;
			}
		}
		this.setTaskState( State.FINISHED );
	}


	/**
	 * Runs the task's code.
	 * 
	 * <p>
	 * This method runs the task's code, if it is still referenced. All exceptions will be caught
	 * and logged, but they will not affect the main loop.
	 * 
	 * @return <code>true</code> if the task was still referenced, <code>false</code> if it wasn't.
	 */
	private boolean runTask( )
	{
		Runnable task = this.task.get( );
		if ( task == null ) {
			this.logger.log( LogLevel.INFO , "task '" + this.name + "' is no longer referenced, exiting" )
					.flush( );
			return false;
		}

		try {
			this.logger.log( LogLevel.TRACE , "task '" + this.name + "' started" ).flush( );
			task.run( );
			this.logger.log( LogLevel.TRACE , "task '" + this.name + "' ended" ).flush( );
		} catch ( Throwable t ) {
			this.logger.log( LogLevel.ERROR , "task '" + this.name + "' failed due to exception" , t ).flush( );
		}

		return true;
	}


	/**
	 * Triggers the task's execution.
	 * 
	 * <p>
	 * This method attempts to trigger the task's execution. It will not do anything if the task
	 * thread is exiting or finished, but if the task is still running, a warning will be added to
	 * the log.
	 */
	void startTask( )
	{
		this.lock.lock( );
		try {
			State state = this.getTaskState( );
			if ( state == State.EXITING || state == State.FINISHED ) {
				return;
			} else if ( state == State.RUNNING ) {
				this.logger.log( LogLevel.WARNING , "task '" + this.name + "' didn't manage to run in time" )
						.flush( );
				return;
			}
			this.setTaskState( State.RUNNING );
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * Terminates the task execution thread.
	 * 
	 * <p>
	 * This method shuts down the task execution thread, waiting until it has completed to return.
	 */
	void terminate( )
	{
		this.lock.lock( );
		try {
			State state = this.getTaskState( );
			while ( state != State.FINISHED ) {
				this.setTaskState( State.EXITING );
				state = this.waitTaskState( );
			}
		} finally {
			this.lock.unlock( );
		}
		this.logger.log( LogLevel.DEBUG , "task '" + this.name + "' terminated" ).flush( );
	}


	/**
	 * Waits until the task is done running.
	 * 
	 * <p>
	 * This method waits until the task is no longer running. If the task is not running when the
	 * method is called, it will return immediately.
	 */
	void waitForTask( )
	{
		this.lock.lock( );
		try {
			State state = this.getTaskState( );
			while ( state == State.RUNNING ) {
				state = this.waitTaskState( );
			}
		} finally {
			this.lock.unlock( );
		}
	}


	/**
	 * @return <code>true</code> if the task execution thread is no longer running,
	 *         <code>false</code> otherwise.
	 */
	boolean isFinished( )
	{
		return ( this.getTaskState( ) == State.FINISHED );
	}
}
