package com.deepclone.lw.beans.game.technologies;


import java.sql.Types;
import java.util.List;

import javax.sql.DataSource;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;

import com.deepclone.lw.cmd.player.gdata.empire.ResearchData;
import com.deepclone.lw.interfaces.game.technologies.PlayerTechnologiesDAO;
import com.deepclone.lw.utils.StoredProc;



/**
 * Data access component for player technologies
 * 
 * <p>
 * This class implements database queries and calls to stored procedures which allow players to
 * access and manage their empires' research.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class PlayerTechnologiesDAOBean
		implements PlayerTechnologiesDAO
{

	/** SQL query that fetches an empire's research state */
	private static final String Q_EMPIRE_RESEARCH = "SELECT * FROM emp.technologies_view WHERE empire_id = ?";

	/** Row mapper for research state entries */
	private final ResearchRowMapper mResearch;
	
	/** Stored procedure that initiates a research priority update */
	private StoredProc fResprioUpdateStart;

	/** Stored procedure that uploads a single research priority */
	private StoredProc fResprioUpdateSet;

	/** Stored procedure that applies a research priority update */
	private StoredProc fResprioUpdateApply;
	
	/** Stored procedure that implements a technology */
	private StoredProc fTechnologyImplement;

	/** Spring JDBC interface */
	private JdbcTemplate dTemplate;


	/** Initialise the necessary row mappers */
	public PlayerTechnologiesDAOBean( )
	{
		this.mResearch = new ResearchRowMapper( );
	}


	/**
	 * Dependency injector that sets the data source and initialises stored procedures 
	 * 
	 * @param dataSource
	 *            the data source
	 */
	@Autowired( required = true )
	public void setDataSource( DataSource dataSource )
	{
		this.dTemplate = new JdbcTemplate( dataSource );

		this.fResprioUpdateStart = new StoredProc( dataSource , "emp" , "resprio_update_start" )
			.addParameter( "_empire" , Types.INTEGER )
			.addOutput( "_result" , Types.BOOLEAN );

		this.fResprioUpdateSet = new StoredProc( dataSource , "emp" , "resprio_update_set" )
			.addParameter( "_technology" , Types.VARCHAR )
			.addParameter( "_priority" , Types.INTEGER )
			.addOutput( "_result" , Types.BOOLEAN );

		this.fResprioUpdateApply = new StoredProc( dataSource , "emp" , "resprio_update_apply" )
			.addOutput( "_result" , Types.BOOLEAN );

		this.fTechnologyImplement = new StoredProc( dataSource , "emp" , "technology_implement" )
			.addParameter( "_empire" , Types.INTEGER )
			.addParameter( "_technology" , Types.VARCHAR )
			.addOutput( "_result" , Types.BOOLEAN );
	}


	/**
	 * Query and map research entries for an empire
	 * 
	 * <p>
	 * Run the query against <code>emp.technologies_view</code> then map all rows using a
	 * {@link ResearchRowMapper}.
	 */
	@Override
	public List< ResearchData > getResearchData( int empire )
	{
		return this.dTemplate.query( Q_EMPIRE_RESEARCH , this.mResearch , empire );
	}


	/**
	 * Call <code>emp.resprio_update_start</code>
	 */
	@Override
	public boolean startPriorityUpdate( int empire )
	{
		return (Boolean) this.fResprioUpdateStart.execute( empire ).get( "_result" );
	}


	/**
	 * Call <code>emp.resprio_update_set</code>
	 */
	@Override
	public boolean uploadResearchPriority( String identifier , int priority )
	{
		return (Boolean) this.fResprioUpdateSet.execute( identifier , priority ).get( "_result" );
	}


	/**
	 * Call <code>emp.resprio_update_apply</code>
	 */
	@Override
	public boolean applyPriorityUpdate( )
	{
		return (Boolean) this.fResprioUpdateApply.execute( ).get( "_result" );
	}


	/**
	 * Call <code>emp.technology_implement</code>
	 */
	@Override
	public boolean implement( int empire , String technology )
	{
		return (Boolean) this.fTechnologyImplement.execute( empire , technology ).get( "_result" );
	}

}
