package com.deepclone.lw.beans.game.technologies;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import com.deepclone.lw.cmd.player.gdata.GamePageData;
import com.deepclone.lw.cmd.player.gdata.empire.ResearchData;
import com.deepclone.lw.cmd.player.tech.GetResearchResponse;
import com.deepclone.lw.interfaces.game.EmpireManagement;
import com.deepclone.lw.interfaces.game.technologies.PlayerTechnologiesDAO;
import com.deepclone.lw.interfaces.game.technologies.ResearchController;
import com.deepclone.lw.interfaces.i18n.TranslationException;
import com.deepclone.lw.interfaces.i18n.Translator;



/**
 * Research control component
 * 
 * <p>
 * This component implements all actions allowing a player to view and update its empire's research
 * state.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@Transactional
class ResearchControllerBean
		implements ResearchController
{
	/** Main empire management component */
	private EmpireManagement empireManagement;

	/** String translator */
	private Translator translator;

	/** Data access object for technologies */
	private PlayerTechnologiesDAO playerTechnologiesDAO;


	/**
	 * Dependency injector that sets the empire management component
	 * 
	 * @param empireManagement
	 *            the empire management component
	 */
	@Autowired( required = true )
	public void setEmpireManagement( EmpireManagement empireManagement )
	{
		this.empireManagement = empireManagement;
	}


	/**
	 * Dependency injector that sets the translation component
	 * 
	 * @param translator
	 *            the translation component
	 */
	@Autowired( required = true )
	public void setTranslator( Translator translator )
	{
		this.translator = translator;
	}


	/**
	 * Dependency injector that sets the data access object for technologies
	 * 
	 * @param playerTechnologiesDAO
	 *            the data access object for technologies
	 */
	@Autowired( required = true )
	public void setPlayerTechnologiesDAO( PlayerTechnologiesDAO playerTechnologiesDAO )
	{
		this.playerTechnologiesDAO = playerTechnologiesDAO;
	}


	/**
	 * Get the raw research data from the database, then handle translation and dependencies
	 * 
	 * <p>
	 * In order to generate the response, this method fetches the raw research state information
	 * from the database. It then proceeds with translating all categories, names and descriptions,
	 * then adds reverse dependency information to all loaded technologies.
	 */
	@Override
	public GetResearchResponse getResearchData( int empire )
	{
		GamePageData pageData = this.empireManagement.getGeneralInformation( empire );
		List< ResearchData > technologies = this.playerTechnologiesDAO.getResearchData( empire );

		// Translate categories, names and descriptions
		Map< String , String > translations = this.getTranslationsFor( technologies , pageData.getLanguage( ) );
		Map< String , ResearchData > byId = new HashMap< String , ResearchData >( );
		for ( ResearchData tech : technologies ) {
			byId.put( tech.getIdentifier( ) , tech );
			tech.setCategory( translations.get( tech.getCategory( ) ) );
			if ( tech.getName( ) != null ) {
				tech.setName( translations.get( tech.getName( ) ) );
				tech.setDescription( translations.get( tech.getDescription( ) ) );
			}

			tech.setBuildings( this.translateList( tech.getBuildings( ) , translations ) );
		}

		// Add reverse dependency identifiers
		for ( ResearchData tech : technologies ) {
			for ( String dependency : tech.getDependencies( ) ) {
				byId.get( dependency ).addReverseDependency( tech.getIdentifier( ) );
			}
		}

		return new GetResearchResponse( pageData , technologies );
	}


	/**
	 * Get translations used by a list of raw research entries
	 * 
	 * <p>
	 * This method is fed the raw research state data from the player technologies data access
	 * object. It generates a set of string identifiers which contains all categories, names and
	 * descriptions, then fetches these strings' translations.
	 * 
	 * @param technologies
	 *            the research state data from the DAO
	 * @param language
	 *            the player's selected language
	 * 
	 * @return the map of string identifiers to string contents
	 */
	private Map< String , String > getTranslationsFor( List< ResearchData > technologies , String language )
	{
		Set< String > identifiers = new HashSet< String >( );
		for ( ResearchData tech : technologies ) {
			identifiers.add( tech.getCategory( ) );
			if ( tech.getName( ) != null ) {
				identifiers.add( tech.getName( ) );
				identifiers.add( tech.getDescription( ) );
			}
			identifiers.addAll( tech.getBuildings( ) );
		}

		try {
			return this.translator.translate( language , identifiers );
		} catch ( TranslationException e ) {
			throw new RuntimeException( "error while translating technology-related data" , e );
		}
	}


	/**
	 * Translate a list of strings
	 * 
	 * <p>
	 * This method fetches translations for a list of string identifiers.
	 * 
	 * @param input
	 *            the list of strings to translate
	 * @param translations
	 *            the map of translations returned by {@link #getTranslationsFor(List, String)}
	 * 
	 * @return the translated list of strings
	 */
	private List< String > translateList( List< String > input , Map< String , String > translations )
	{
		ArrayList< String > result = new ArrayList< String >( input.size( ) );
		for ( String str : input ) {
			result.add( translations.get( str ) );
		}
		return result;
	}


	/**
	 * Update research priorities
	 * 
	 * <p>
	 * This method will start a research priority update for the specified empire, then upload each
	 * value in the map to the database. Finally, it will try to apply the update.
	 * 
	 * <p>
	 * If any of the above steps fail, the method will abort.
	 */
	@Override
	public boolean updatePriorities( int empire , Map< String , Integer > priorities )
	{
		if ( !this.playerTechnologiesDAO.startPriorityUpdate( empire ) ) {
			return false;
		}

		for ( Map.Entry< String , Integer > entry : priorities.entrySet( ) ) {
			if ( !this.playerTechnologiesDAO.uploadResearchPriority( entry.getKey( ) , entry.getValue( ) ) ) {
				return false;
			}
		}

		return this.playerTechnologiesDAO.applyPriorityUpdate( );
	}


	/**
	 * Implement a technology
	 * 
	 * <p>
	 * Call the DAO's technology implementation method in order to access the corresponding stored
	 * procedure.
	 */
	@Override
	public boolean implement( int empire , String technology )
	{
		return this.playerTechnologiesDAO.implement( empire , technology );
	}

}
