package com.deepclone.lw.beans.game.technologies;


import java.sql.ResultSet;
import java.sql.SQLException;

import org.postgresql.util.PGobject;

import org.springframework.jdbc.core.RowMapper;

import com.deepclone.lw.cmd.player.gdata.empire.ResearchData;



/**
 * Row mapper for empire research state entries
 * 
 * <p>
 * This class is responsible for converting empire research state information rows into instances of
 * the corresponding class, {@link ResearchData}. The instances it produces are "raw": they contain
 * string identifiers instead of translations, and their reverse dependencies are not set.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
class ResearchRowMapper
		implements RowMapper< ResearchData >
{

	/**
	 * Map a row from <code>emp.technologies_view</code>
	 * 
	 * <p>
	 * This method maps a single row from the <code>emp.technologies_view</code> view into a
	 * {@link ResearchData} instance.
	 */
	@Override
	public ResearchData mapRow( ResultSet rs , int rowNum )
			throws SQLException
	{
		ResearchData output = new ResearchData( );

		output.setIdentifier( rs.getString( "emptech_id" ) );
		output.setCategory( rs.getString( "technology_category" ) );

		if ( rs.getBoolean( "emptech_visible" ) ) {
			output.setName( rs.getString( "technology_name" ) );
			output.setDescription( rs.getString( "technology_description" ) );
			output.setPrice( rs.getLong( "technology_price" ) );
		}

		int ratio = rs.getInt( "emptech_ratio" );
		if ( rs.wasNull( ) ) {
			String state = ( (PGobject) rs.getObject( "emptech_state" ) ).getValue( );
			output.setImplemented( "KNOWN".equals( state ) );
		} else {
			output.setCompletion( ratio );
			output.setPriority( rs.getInt( "emptech_priority" ) );
		}

		output.setDependencies( this.splitField( rs , "technology_dependencies" ) );
		output.setBuildings( this.splitField( rs , "technology_buildings" ) );

		return output;
	}


	/**
	 * Extract a comma-separated field
	 * 
	 * <p>
	 * This method accesses then extracts the contents of a comma-separated field (for example
	 * dependencies or unlocked buildings).
	 * 
	 * @param rs
	 *            the SQL result set
	 * @param field
	 *            the field's name
	 * 
	 * @return an array of strings containing the field's values
	 * 
	 * @throws SQLException
	 *             if something goes wrong while accessing the field
	 */
	private String[] splitField( ResultSet rs , String field )
			throws SQLException
	{
		String fValue = rs.getString( field );
		if ( fValue == null || "".equals( fValue ) ) {
			return new String[ 0 ];
		}
		return fValue.split( "," );
	}

}
