package com.deepclone.lw.beans.updates;


import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cmd.admin.logs.LogLevel;
import com.deepclone.lw.interfaces.eventlog.Logger;
import com.deepclone.lw.interfaces.eventlog.SystemLogger;
import com.deepclone.lw.interfaces.game.updates.GameUpdateProcessor;
import com.deepclone.lw.interfaces.game.updates.UpdatesDAO;
import com.deepclone.lw.interfaces.sys.MaintenanceStatusException;
import com.deepclone.lw.interfaces.sys.SystemStatus;
import com.deepclone.lw.interfaces.sys.TickStatusException;



/**
 * Game update processor component
 * 
 * <p>
 * This component implements game updates processing. It will update the system status as necessary,
 * checking for maintenance mode, and of course implement the necessary operations for game updates.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class GameUpdateProcessorBean
		implements GameUpdateProcessor
{
	/** System state management component */
	private SystemStatus systemStatus;

	/** Logger used by this component */
	private SystemLogger logger;

	/** Transaction template used to run update batches */
	private TransactionTemplate tTemplate;

	/** Update processing database interface */
	private UpdatesDAO updatesDao;

	/** Server-level update processors registry */
	private ServerProcessorRegistry registry;

	/** Game update processor lock */
	private boolean executing;


	/**
	 * Dependency injector that sets the sytstem state management component
	 * 
	 * @param systemStatus
	 *            the sytstem state management component
	 */
	@Autowired( required = true )
	public void setSystemStatus( SystemStatus systemStatus )
	{
		this.systemStatus = systemStatus;
	}


	/**
	 * Dependency injector that initialises the logger from the logging component
	 * 
	 * @param logger
	 *            the logging component
	 */
	@Autowired( required = true )
	public void setLogger( Logger logger )
	{
		this.logger = logger.getSystemLogger( "GameUpdate" );
	}


	/**
	 * Dependency injector that initialises the transaction template from the transaction manager
	 * 
	 * @param transactionManager
	 *            the transaction manager
	 */
	@Autowired( required = true )
	public void setTransactionManager( PlatformTransactionManager transactionManager )
	{
		this.tTemplate = new TransactionTemplate( transactionManager );
	}


	/**
	 * Dependency injector that sets the database access component for game updates
	 * 
	 * @param updatesDao
	 *            the database access component for game updates
	 */
	@Autowired( required = true )
	public void setUpdatesDAO( UpdatesDAO updatesDao )
	{
		this.updatesDao = updatesDao;
	}


	/**
	 * Dependency injector that sets the processor registry
	 * 
	 * @param registry
	 *            the processor registry
	 */
	@Autowired( required = true )
	public void setRegistry( ServerProcessorRegistry registry )
	{
		this.registry = registry;
	}


	/**
	 * Try and lock the processor
	 * 
	 * <p>
	 * Test the {@link #executing} field, setting it to <code>true</code> and returning
	 * <code>true</code> if it wasn't or returning <code>false</code> if it was.
	 */
	@Override
	public synchronized boolean tryLock( )
	{
		if ( this.executing ) {
			return false;
		}
		this.executing = true;
		return true;
	}


	/**
	 * Unlock the processor
	 * 
	 * <p>
	 * Set {@link #executing} back to <code>false</code>. If it was already set to
	 * <code>false</code>, throw an {@link IllegalStateException}.
	 */
	@Override
	public synchronized void unlock( )
			throws IllegalStateException
	{
		if ( !this.executing ) {
			throw new IllegalStateException( "Trying to unlock game update processor, but it isn't locked" );
		}
		this.executing = false;
	}


	/**
	 * Check for an unfinished update cycle and terminate it
	 * 
	 * <p>
	 * Check for a "stuck" update cycle. If there is one, issue a warning and process it through the
	 * {@link #executeTick(long)} method.
	 */
	@Override
	public boolean endPreviousCycle( )
	{
		Long currentTick;
		try {
			currentTick = this.systemStatus.checkStuckTick( );
		} catch ( MaintenanceStatusException e ) {
			return true;
		}

		if ( currentTick == null ) {
			return false;
		}

		this.logger.log( LogLevel.WARNING , "Tick " + currentTick + " restarted" ).flush( );
		this.executeTick( currentTick );
		return true;
	}


	/**
	 * Execute an update cycle
	 * 
	 * <p>
	 * Start by checking for, and executing if necessary, a "stuck" update cycle. If one is found,
	 * return. Otherwise, start a new tick through the system state manager, and call
	 * {@link #executeTick(long)} to process it.
	 * 
	 * <p>
	 * If maintenance mode is active, the new tick will not be started.
	 */
	@Override
	public void executeUpdateCycle( )
	{
		// Attempt to end the previous tick, if e.g. maintenance mode was initiated while it was
		// being processed
		if ( this.endPreviousCycle( ) ) {
			return;
		}

		// Initiate next tick
		long tickId;
		try {
			tickId = this.systemStatus.startTick( );
		} catch ( TickStatusException e ) {
			throw new RuntimeException( "tick initiated while previous tick still being processed" , e );
		} catch ( MaintenanceStatusException e ) {
			return;
		}

		// Execute tick
		this.logger.log( LogLevel.DEBUG , "Tick " + tickId + " started" ).flush( );
		this.executeTick( tickId );
	}


	/**
	 * Process all remaining update batches for an update cycle
	 * 
	 * <p>
	 * This method will try and process all update batches using a {@link GameUpdateTransaction}
	 * instance. It loops until the transaction indicates that no further updates are to be
	 * processed. However, the loop will exit if maintenance mode becomes active between update
	 * batches.
	 * 
	 * @param tickId
	 *            the game update cycle's identifier
	 */
	private void executeTick( final long tickId )
	{
		GameUpdateTransaction transaction = new GameUpdateTransaction( this.updatesDao , this.registry , tickId );
		while ( !this.tTemplate.execute( transaction ) ) {
			if ( this.systemStatus.checkMaintenance( ) != null ) {
				this.logger.log( LogLevel.INFO , "Tick " + tickId + " interrupted by system maintenance" ).flush( );
				return;
			}
		}
		this.logger.log( LogLevel.TRACE , "Tick " + tickId + " completed" ).flush( );
	}

}
