package com.deepclone.lw.beans.updates;


import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;

import com.deepclone.lw.interfaces.game.updates.GameUpdateProcessor;
import com.deepclone.lw.interfaces.sys.Ticker;
import com.deepclone.lw.interfaces.sys.Ticker.Frequency;



/**
 * Game update ticker task
 * 
 * <p>
 * This component will register itself as a ticker task to be executed once per minute. When it
 * runs, it will try and execute an update cycle.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class GameUpdateTaskBean
		implements InitializingBean , Runnable
{
	/** The ticker component */
	private Ticker ticker;

	/** The game updates processor */
	private GameUpdateProcessor gameUpdateProcessor;


	/**
	 * Dependency injector that sets the ticker component
	 * 
	 * @param ticker
	 *            the ticker component
	 */
	@Autowired( required = true )
	public void setTicker( Ticker ticker )
	{
		this.ticker = ticker;
	}


	/**
	 * Dependency injector that sets the game update processor
	 * 
	 * @param gameUpdateProcessor
	 *            the game update processor
	 */
	@Autowired( required = true )
	public void setGameUpdateProcessor( GameUpdateProcessor gameUpdateProcessor )
	{
		this.gameUpdateProcessor = gameUpdateProcessor;
	}


	/**
	 * Finish a previous update cycle if there is one, then register the component to the ticker
	 */
	@Override
	public void afterPropertiesSet( )
	{
		if ( this.gameUpdateProcessor.tryLock( ) ) {

			try {
				this.gameUpdateProcessor.endPreviousCycle( );
			} finally {
				this.gameUpdateProcessor.unlock( );
			}

		}
		this.ticker.registerTask( Frequency.MINUTE , "Game update" , this );
	}


	/**
	 * Run a game update cycle
	 * 
	 * <p>
	 * Try locking the game update processor. On success, execute an update cycle, then unlock the
	 * processor.
	 */
	@Override
	public void run( )
	{
		if ( !this.gameUpdateProcessor.tryLock( ) ) {
			return;
		}
		try {
			this.gameUpdateProcessor.executeUpdateCycle( );
		} finally {
			this.gameUpdateProcessor.unlock( );
		}
	}
}
