package com.deepclone.lw.beans.updates;


import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;

import com.deepclone.lw.interfaces.game.updates.UpdateBatchProcessor;
import com.deepclone.lw.interfaces.game.updates.UpdatesDAO;
import com.deepclone.lw.sqld.sys.GameUpdateResult;



/**
 * Game update transaction
 * 
 * <p>
 * This class is responsible for executing the necessary operations during a game update
 * transaction. It is used by the {@link GameUpdateProcessorBean} component to implement its main
 * processing loop.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class GameUpdateTransaction
		implements TransactionCallback< Boolean >
{
	/** The game updates database access interface */
	private final UpdatesDAO updatesDao;

	/** The registry of server-side update batch processors */
	private final ServerProcessorRegistry registry;

	/** The identifier of the current update cycle */
	private final long tickId;


	/**
	 * Initialise the various dependencies and set the identifier of the current cycle
	 * 
	 * @param updatesDao
	 *            the game updates database access interface
	 * @param registry
	 *            the registry of server-side update batch processors
	 * @param tickId
	 *            the identifier of the current update cycle
	 */
	GameUpdateTransaction( UpdatesDAO updatesDao , ServerProcessorRegistry registry , long tickId )
	{
		this.updatesDao = updatesDao;
		this.registry = registry;
		this.tickId = tickId;
	}


	/**
	 * Update batch processing transaction
	 * 
	 * <p>
	 * Call the database's game update processing function, returning immediately if the function
	 * indicates that the cycle has been completed. Otherwise, execute a server-side processor if
	 * necessary.
	 * 
	 * <p>
	 * If the database requests server-side processing but no processor is found, an
	 * {@link UnsupportedUpdateException} will be thrown.
	 * 
	 * @return <code>true</code> if the current update cycle has been completed, <code>false</code>
	 *         if more updates are needed.
	 */
	@Override
	public Boolean doInTransaction( TransactionStatus status )
	{
		GameUpdateResult result = this.updatesDao.processUpdates( this.tickId );
		String local = result.getLocalExecution( );
		if ( result.isFinished( ) ) {
			return true;
		} else if ( local == null ) {
			return false;
		}

		UpdateBatchProcessor processor = this.registry.getProcessorFor( local );
		if ( processor == null ) {
			throw new UnsupportedUpdateException( local );
		}

		processor.processBatch( this.tickId );
		return false;
	}

}
