package com.deepclone.lw.beans.updates;


import java.util.HashMap;
import java.util.Map;

import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanInitializationException;
import org.springframework.beans.factory.config.BeanPostProcessor;

import com.deepclone.lw.interfaces.game.updates.UpdateBatchProcessor;



/**
 * Server-side update processor registry component
 * 
 * <p>
 * This component will look through registered Spring components for update processors, and use the
 * results as the registry's contents.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
class ServerProcessorRegistryBean
		implements ServerProcessorRegistry , BeanPostProcessor
{

	/** Known server-side update batch processors */
	private final Map< String , UpdateBatchProcessor > processors = new HashMap< String , UpdateBatchProcessor >( );


	/** Do nothing - we're not interested in uninitialised components */
	@Override
	public Object postProcessBeforeInitialization( Object bean , String beanName )
	{
		return bean;
	}


	/**
	 * If an initialised component implements the {@link UpdateBatchProcessor} interface, add it to
	 * the {@link #processors} map. A {@link BeanInitializationException} will be thrown if two
	 * components seem to process the same update type.
	 */
	@Override
	public Object postProcessAfterInitialization( Object bean , String beanName )
			throws BeansException
	{
		if ( bean instanceof UpdateBatchProcessor ) {
			UpdateBatchProcessor processor = (UpdateBatchProcessor) bean;
			String updType = processor.getUpdateType( );
			if ( this.processors.containsKey( updType ) ) {
				throw new BeanInitializationException( "Duplicate update processor for update type " + updType );
			}
			this.processors.put( updType , processor );
		}
		return bean;
	}


	/** Obtain the processor from the {@link #processors} map */
	@Override
	public UpdateBatchProcessor getProcessorFor( String type )
	{
		return this.processors.get( type );
	}

}
