/*
 * Tests of the update type definitions
 */
BEGIN;
	-- Make sure we get rid of all existing definitions first
	DELETE FROM sys.update_types;
	DELETE FROM sys.update_targets;

	-- Create a test schema & tables which will be used as targets
	CREATE SCHEMA test
		CREATE TABLE test(
			test_id		SERIAL NOT NULL PRIMARY KEY
		);

	-- Create target definitions
	INSERT INTO sys.update_targets(
			updtgt_id , updtgt_name , updtgt_schema , updtgt_table
		) VALUES (
			1 , 'Test 1' , 'test' , 'test'
		);

	-- Create a dummy stored procedure in the sys schema
	CREATE FUNCTION sys.dummy_update_func( c_tick BIGINT )
			RETURNS INT LANGUAGE SQL AS 'SELECT 1;';

	-- ***** TESTS BEGIN HERE *****
	SELECT plan( 14 );
	
	SELECT diag_test_name( 'Update types - Definition with invalid target' );
	SELECT throws_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering
		) VALUES (
			1 , 5 , 'Test' , '' , 1
		);
	$$ , 23503 );
	
	SELECT diag_test_name( 'Update types - Definition with invalid stored procedure name' );
	SELECT throws_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering , updtype_proc_name
		) VALUES (
			1 , 5 , 'Test' , '' , 1 , 'does_not_exist'
		);
	$$ , 'P0001' );
	
	SELECT diag_test_name( 'Update types - Successful insertion with no procedure name' );
	SELECT lives_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering
		) VALUES (
			1 , 1 , 'Test' , '' , 1
		);
	$$ );
	DELETE FROM sys.update_types;
	
	SELECT diag_test_name( 'Update types - Successful insertion with a procedure name' );
	SELECT lives_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering , updtype_proc_name
		) VALUES (
			1 , 1 , 'Test' , '' , 1 , 'dummy_update_func'
		);
	$$ );
	
	SELECT diag_test_name( 'Update types - Setting an existing row''s stored procedure to something invalid fails' );
	SELECT throws_ok( $$
		UPDATE sys.update_types
			SET updtype_proc_name = 'does_not_exist'
			WHERE updtype_id = 1
	$$ , 'P0001' );

	SELECT diag_test_name( 'Update types - Failure on duplicate name' );
	SELECT throws_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering
		) VALUES (
			2 , 1 , 'Test' , '' , 1
		);
	$$ , 23505 );

	SELECT diag_test_name( 'Update types - Failure on duplicate stored procedure' );
	SELECT throws_ok( $$
		INSERT INTO sys.update_types(
			updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering , updtype_proc_name
		) VALUES (
			3 , 1 , 'Test 2' , '' , 1 , 'dummy_update_func'
		);
	$$ , 23505 );

	SELECT diag_test_name( 'Update types - Target deletion fails if type definitions exist' );
	SELECT throws_ok( $$ DELETE FROM sys.update_targets WHERE updtgt_id = 1 $$ , 23503 );

	DELETE FROM sys.update_types;

	INSERT INTO sys.update_types(
		updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering
	) VALUES (
			1 , 1 , 'Test 1' , '' , 1
		) , (
			2 , 1 , 'Test 3' , '' , 75
		) , (
			3 , 1 , 'Test 2' , '' , 2
		);
	SELECT diag_test_name( 'Update types - Ordering column is updated' );
	SELECT set_eq(
		$$ SELECT updtype_id , updtype_ordering FROM sys.update_types $$ ,
		$$ VALUES ( 1 , 2 ) , ( 3 , 4 ) , ( 2 , 6 ) $$
	);
	DELETE FROM sys.update_types;
	
	
	INSERT INTO test.test VALUES ( 1 ) , ( 2 );
	INSERT INTO sys.update_types(
		updtype_id , updtgt_id , updtype_name , updtype_description , updtype_ordering
	) VALUES (
			1 , 1 , 'Test 1' , '' , 1
	);
	SELECT diag_test_name( 'Update types - Type creation adds specific rows for missing items' );
	SELECT set_eq(
		$$ SELECT updtgt_id , updtype_id , test_id FROM test.test_updates $$ ,
		$$ VALUES ( 1 , 1 , 1 ) , ( 1 , 1 , 2 ) $$
	);
	SELECT diag_test_name( 'Update types - Type creation adds genertic rows for missing items' );
	SELECT set_eq(
		$$ SELECT test_id , update_state , update_last
				FROM test.test_updates
					INNER JOIN sys.updates USING ( updtgt_id , updtype_id , update_id ) $$ ,
		$$ VALUES ( 1 , 'FUTURE'::sys.update_state_type , -1 ) ,
				( 2 , 'FUTURE'::sys.update_state_type , '-1' ) $$
	);
	
	INSERT INTO test.test VALUES ( 3 );
	SELECT diag_test_name( 'Update types - New items have automatically inserted update rows' );
	SELECT set_eq(
		$$ SELECT test_id , updtype_id , update_state , update_last
				FROM test.test_updates
					INNER JOIN sys.updates USING ( updtgt_id , updtype_id , update_id )
				WHERE test_id = 3 $$ ,
		$$ VALUES ( 3 , 1 , 'FUTURE'::sys.update_state_type , -1 ) $$
	);

	DELETE FROM sys.update_types WHERE updtype_id = 1;
	SELECT diag_test_name( 'Update types - Type deletion cascades to specific rows' );
	SELECT is_empty( $$ SELECT * FROM test.test_updates $$ );
	SELECT diag_test_name( 'Update types - Type deletion cascades to generic rows' );
	SELECT is_empty( $$ SELECT * FROM sys.updates $$ );

	SELECT * FROM finish( );
ROLLBACK;