package com.deepclone.lw.interfaces.i18n;


import java.util.Collection;
import java.util.Map;
import java.util.Set;



/**
 * Translator service interface
 * 
 * <p>
 * This interface defines the methods available on the Translator service. One such service should
 * be present on all nodes. All Translator service instances are managed by the I18NManager service,
 * which is shared between nodes and notifies Translator instances of database updates.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public interface Translator
{

	/**
	 * @return the list of supported languages
	 */
	public Set< String > getSupportedLanguages( );


	/**
	 * Checks if a language is supported.
	 * 
	 * @param language
	 *            the identifier of the language
	 * @return true if the language is present and fully supported, false otherwise.
	 */
	public boolean isLanguageSupported( String language );


	/**
	 * Returns the plain text name of a language from its identifier.
	 * 
	 * @param language
	 *            the identifier of the language
	 * @return the name of the language
	 * @throws UnknownLanguageException
	 *             if the specified language does not exist
	 */
	public String getLanguageName( String language )
			throws UnknownLanguageException;;


	/**
	 * Translates a string specified using its identifier.
	 * 
	 * @param language
	 *            the identifier of the language to translate to
	 * @param string
	 *            the identifier of the string to translate
	 * @return the string's translation
	 * @throws UnknownStringException
	 *             if the string is unknown
	 * @throws UnknownLanguageException
	 *             if the language does not exist or is not supported
	 */
	public String translate( String language , String string )
			throws UnknownStringException , UnknownLanguageException;


	/**
	 * Translate multiple strings based on their identifiers
	 * 
	 * <p>
	 * This method must be implemented to allow "en masse" string translations. It will fetch the
	 * translations in a given language for an arbitrary quantity of string identifiers, returning
	 * the results as a map of identifiers to translations.
	 * 
	 * @param language
	 *            the identifier of the language to translate to
	 * @param strings
	 *            a collection of string identifiers
	 * 
	 * @return a map associating string identifiers to translations
	 * 
	 * @throws UnknownStringException
	 *             if one of the string identifiers does not match a string
	 * @throws UnknownLanguageException
	 *             if the language does not exist or is not supported
	 */
	public Map< String , String > translate( String language , Collection< String > strings )
			throws UnknownStringException , UnknownLanguageException;
}
