package com.deepclone.lw.interfaces.sys;


/**
 * This class is used by beans which define constants in order to notify the constants manager of
 * their existence and specifics.
 * 
 * @author tseeker
 */
public final class ConstantDefinition
{

	/** The constant's name, as it is used in the code */
	public final String name;

	/** The human-readable category name used in the administration interface */
	public final String category;

	/** The constant's description used in the administrative interface */
	public final String description;

	/** The optional minimal value */
	public final Double minValue;

	/** The optional maximal value */
	public final Double maxValue;

	/** The constant's default value */
	public final Double defaultValue;


	/**
	 * Creates a constant definition with no constraints.
	 * 
	 * @param name
	 *            the constant's name
	 * @param category
	 *            the constant's category
	 * @param description
	 *            the constant's description
	 * @param value
	 *            the constant's default value
	 */
	public ConstantDefinition( String name , String category , String description , Double value )
	{
		this.name = name;
		this.category = category;
		this.description = description;
		this.defaultValue = value;
		this.maxValue = this.minValue = null;
	}


	/**
	 * Creates a constant definition with a single constraint.
	 * 
	 * @param name
	 *            the constant's name
	 * @param category
	 *            the constant's category
	 * @param description
	 *            the constant's description
	 * @param value
	 *            the constant's default value
	 * @param constraint
	 *            a constraint to apply on the constant's definition
	 * @param isMinimal
	 *            whether the constraint is the minimal or maximal value
	 * @throws IllegalArgumentException
	 *             if the default value is out of bounds
	 */
	public ConstantDefinition( String name , String category , String description , Double value , Double constraint ,
			boolean isMinimal )
			throws IllegalArgumentException
	{
		if ( isMinimal && value < constraint || !isMinimal && value > constraint ) {
			throw new IllegalArgumentException( );
		}
		this.name = name;
		this.category = category;
		this.description = description;
		this.defaultValue = value;
		this.maxValue = isMinimal ? null : constraint;
		this.minValue = isMinimal ? constraint : null;
	}


	/**
	 * Creates a constant definition with two constraints.
	 * 
	 * @param name
	 *            the constant's name
	 * @param category
	 *            the constant's category
	 * @param description
	 *            the constant's description
	 * @param value
	 *            the constant's default value
	 * @param min
	 *            the constant's minimal value
	 * @param max
	 *            the constant's maximal value
	 * @throws IllegalArgumentException
	 *             if the minimal is greater than the maximal or if the default value is out of
	 *             bounds
	 */
	public ConstantDefinition( String name , String category , String description , Double value , Double min ,
			Double max )
			throws IllegalArgumentException
	{
		if ( min > max || value < min || value > max ) {
			throw new IllegalArgumentException( );
		}
		this.name = name;
		this.category = category;
		this.description = description;
		this.defaultValue = value;
		this.maxValue = max;
		this.minValue = min;

	}
}
