package com.deepclone.lw.interfaces.sys;


/**
 * System task scheduler
 * 
 * This interface provides access to the system task scheduler, which is used for everything that is
 * supposed to happen at a regular interval (from log cleaning to ticks).
 * 
 * @author tseeker
 */
public interface Ticker
{

	/** Task execution frequencies */
	public enum Frequency {
		/** Tasks that are meant to be executed every five seconds */
		HIGH ,

		/** Tasks that are meant to be executed every thirty seconds */
		MEDIUM ,

		/** Tasks that are meant to be executed every minute */
		MINUTE ,

		/** Tasks that are meant to be executed every five minutes */
		LOW
	}


	/**
	 * Registers a new task.
	 * 
	 * This method may be called to register a new task in the system ticker. The ticker will use a
	 * weak reference to store the task's execution instance.
	 * 
	 * @param frequency
	 *            task execution frequency
	 * @param name
	 *            task name
	 * @param task
	 *            runnable instance to be executed at the specified frequency
	 */
	public void registerTask( Ticker.Frequency frequency , String name , Runnable task );


	/**
	 * Pauses the ticker.
	 * 
	 * This method pauses the ticker. Once the method is called, no new tasks will be started;
	 * however, the method will wait for any currently running tasks before returning.
	 * 
	 * @throws IllegalStateException
	 *             if the ticker was already paused
	 */
	public void pause( )
			throws IllegalStateException;


	/**
	 * Restarts the ticker.
	 * 
	 * This method attempts to restart the ticker after it has been paused.
	 * 
	 * @throws IllegalStateException
	 *             if the ticker was already running.
	 */
	public void unpause( )
			throws IllegalStateException;


	/**
	 * @return <code>true</code> the ticker is currently running or <code>false</code> if it has
	 *         been paused.
	 */
	public boolean isActive( );

}
