package com.deepclone.lw.cli;


import java.io.*;
import java.sql.Types;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.sqld.game.BuildingOutputType;
import com.deepclone.lw.utils.StoredProc;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;
import com.thoughtworks.xstream.annotations.XStreamImplicit;



public class ImportBuildables
		extends CLITool
{
	private final Logger logger = Logger.getLogger( ImportBuildables.class );

	@SuppressWarnings( "serial" )
	@XStreamAlias( "buildables" )
	public static class BuildablesData
			implements Serializable
	{
		@XStreamImplicit
		public List< BuildableData > buildables;
	}

	@SuppressWarnings( "serial" )
	public static abstract class BuildableData
			implements Serializable
	{
		@XStreamAsAttribute
		public String name;

		@XStreamAsAttribute
		public String description;

		public CostData cost;

		@XStreamAlias( "require-technology" )
		public String tech;
	}

	@SuppressWarnings( "serial" )
	@XStreamAlias( "cost" )
	public static class CostData
			implements Serializable
	{
		@XStreamAsAttribute
		public int build;

		@XStreamAsAttribute
		public int upkeep;

		@XStreamAsAttribute
		public int work;
	}

	@SuppressWarnings( "serial" )
	@XStreamAlias( "building" )
	public static class BuildingData
			extends BuildableData
	{
		@XStreamAsAttribute
		public BuildingOutputType type;

		@XStreamAsAttribute
		public int output;

		@XStreamAsAttribute
		public int workers;
	}

	@SuppressWarnings( "serial" )
	@XStreamAlias( "ship" )
	public static class ShipData
			extends BuildableData
	{
		@XStreamAsAttribute
		public int time;

		@XStreamAsAttribute
		public int power;
	}

	private File file;
	private TransactionTemplate tTemplate;
	private StoredProc uocBuildingNoDep;
	private StoredProc uocBuildingDep;
	private StoredProc uocShipNoDep;
	private StoredProc uocShipDep;


	private XStream initXStream( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( BuildablesData.class );
		xstream.processAnnotations( BuildingData.class );
		xstream.processAnnotations( ShipData.class );
		return xstream;
	}


	private BuildablesData loadData( )
	{
		FileInputStream fis;
		try {
			fis = new FileInputStream( this.file );
		} catch ( FileNotFoundException e ) {
			return null;
		}

		try {
			XStream xstream = this.initXStream( );
			return (BuildablesData) xstream.fromXML( fis );
		} catch ( Exception e ) {
			e.printStackTrace( );
			return null;
		} finally {
			try {
				fis.close( );
			} catch ( IOException e ) {
				// EMPTY
			}
		}
	}


	private ClassPathXmlApplicationContext createContext( )
	{
		// Load data source
		String[] dataConfig = {
			this.getDataSource( )
		};
		FileSystemXmlApplicationContext ctx = new FileSystemXmlApplicationContext( dataConfig );
		ctx.refresh( );

		// Load Hibernate bean
		String[] cfg = {
			"configuration/transactions.xml"
		};
		return new ClassPathXmlApplicationContext( cfg , true , ctx );
	}


	private void createTemplates( ApplicationContext ctx )
	{
		PlatformTransactionManager tManager = ctx.getBean( PlatformTransactionManager.class );
		this.tTemplate = new TransactionTemplate( tManager );

		DataSource dataSource = ctx.getBean( DataSource.class );

		this.uocBuildingNoDep = new StoredProc( dataSource , "tech" , "uoc_building" );
		this.uocBuildingNoDep.addParameter( "name" , Types.VARCHAR );
		this.uocBuildingNoDep.addParameter( "description" , Types.VARCHAR );
		this.uocBuildingNoDep.addParameter( "cost" , Types.INTEGER );
		this.uocBuildingNoDep.addParameter( "work" , Types.INTEGER );
		this.uocBuildingNoDep.addParameter( "upkeep" , Types.INTEGER );
		this.uocBuildingNoDep.addParameter( "workers" , Types.INTEGER );
		this.uocBuildingNoDep.addParameter( "output_type" , "building_output_type" );
		this.uocBuildingNoDep.addParameter( "output" , Types.INTEGER );

		this.uocBuildingDep = new StoredProc( dataSource , "tech" , "uoc_building" );
		this.uocBuildingDep.addParameter( "name" , Types.VARCHAR );
		this.uocBuildingDep.addParameter( "description" , Types.VARCHAR );
		this.uocBuildingDep.addParameter( "cost" , Types.INTEGER );
		this.uocBuildingDep.addParameter( "work" , Types.INTEGER );
		this.uocBuildingDep.addParameter( "upkeep" , Types.INTEGER );
		this.uocBuildingDep.addParameter( "workers" , Types.INTEGER );
		this.uocBuildingDep.addParameter( "output_type" , "building_output_type" );
		this.uocBuildingDep.addParameter( "output" , Types.INTEGER );
		this.uocBuildingDep.addParameter( "_technology" , Types.VARCHAR );

		this.uocShipNoDep = new StoredProc( dataSource , "tech" , "uoc_ship" );
		this.uocShipNoDep.addParameter( "name" , Types.VARCHAR );
		this.uocShipNoDep.addParameter( "description" , Types.VARCHAR );
		this.uocShipNoDep.addParameter( "cost" , Types.INTEGER );
		this.uocShipNoDep.addParameter( "work" , Types.INTEGER );
		this.uocShipNoDep.addParameter( "upkeep" , Types.INTEGER );
		this.uocShipNoDep.addParameter( "power" , Types.INTEGER );
		this.uocShipNoDep.addParameter( "flight_time" , Types.INTEGER );

		this.uocShipDep = new StoredProc( dataSource , "tech" , "uoc_ship" );
		this.uocShipDep.addParameter( "name" , Types.VARCHAR );
		this.uocShipDep.addParameter( "description" , Types.VARCHAR );
		this.uocShipDep.addParameter( "cost" , Types.INTEGER );
		this.uocShipDep.addParameter( "work" , Types.INTEGER );
		this.uocShipDep.addParameter( "upkeep" , Types.INTEGER );
		this.uocShipDep.addParameter( "power" , Types.INTEGER );
		this.uocShipDep.addParameter( "flight_time" , Types.INTEGER );
		this.uocShipDep.addParameter( "_technology" , Types.VARCHAR );
	}


	private void importBuildables( BuildablesData buildables )
	{
		for ( BuildableData buildable : buildables.buildables ) {
			if ( buildable instanceof BuildingData ) {
				this.importBuilding( (BuildingData) buildable );
			} else {
				this.importShip( (ShipData) buildable );
			}
		}
	}


	private void importShip( ShipData ship )
	{
		this.logger.info( "Importing ship " + ship.name );

		if ( ship.tech == null ) {
			this.uocShipNoDep.execute( ship.name , ship.description , ship.cost.build , ship.cost.work ,
					ship.cost.upkeep , ship.power , ship.time );
		} else {
			this.uocShipDep.execute( ship.name , ship.description , ship.cost.build , ship.cost.work ,
					ship.cost.upkeep , ship.power , ship.time , ship.tech );
		}
	}


	private void importBuilding( BuildingData building )
	{
		this.logger.info( "Importing building " + building.name );
		if ( building.tech == null ) {
			this.uocBuildingNoDep.execute( building.name , building.description , building.cost.build ,
					building.cost.work , building.cost.upkeep , building.workers , building.type.toString( ) ,
					building.output );
		} else {
			this.uocBuildingDep.execute( building.name , building.description , building.cost.build ,
					building.cost.work , building.cost.upkeep , building.workers , building.type.toString( ) ,
					building.output , building.tech );
		}
	}


	@Override
	public void run( )
	{
		final BuildablesData buildables = this.loadData( );
		AbstractApplicationContext ctx = this.createContext( );
		this.createTemplates( ctx );

		boolean rv = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				boolean rv;
				try {
					importBuildables( buildables );
					rv = true;
				} catch ( RuntimeException e ) {
					if ( e.getCause( ) != null ) {
						logger.error( "Exception encoutered" , e );
					}
					logger.error( e.getMessage( ) );
					rv = false;
				}
				if ( !rv ) {
					status.setRollbackOnly( );
				}
				return rv;
			}

		} );

		if ( rv ) {
			this.logger.info( "Import successful" );
		}

		ToolBase.destroyContext( ctx );
	}


	@Override
	public boolean setOptions( String... options )
	{
		if ( options.length != 1 ) {
			return false;
		}
		this.file = new File( options[ 0 ] );
		if ( ! ( this.file.isFile( ) && this.file.canRead( ) ) ) {
			return false;
		}
		return true;
	}

}
