package com.deepclone.lw.cli;


import java.io.File;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.postgresql.util.PGobject;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cli.xmlimport.EventDefinitionLoader;
import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.evdef.EntityEventField;
import com.deepclone.lw.cli.xmlimport.data.evdef.EventDefinition;
import com.deepclone.lw.cli.xmlimport.data.evdef.EventDefinitions;
import com.deepclone.lw.cli.xmlimport.data.evdef.EventFieldBase;
import com.deepclone.lw.cli.xmlimport.data.evdef.IntegerEventField;
import com.deepclone.lw.cli.xmlimport.data.evdef.RealEventField;
import com.deepclone.lw.cli.xmlimport.data.evdef.TextEventField;
import com.deepclone.lw.utils.StoredProc;



public class ImportEvents
		extends CLITool
{

	/**
	 * Error codes that may be returned by the various import stored procedures; see the
	 * corresponding SQL code for details about each value.
	 */
	private static enum SPErrorCode {
		OK ,
		BAD_ID ,
		BAD_STRINGS ,
		DUPLICATE ,
		BAD_SPEC ,
		NO_DATA
	};

	/** Logging system */
	private final Logger logger = Logger.getLogger( ImportResources.class );

	/** File to read the definitions from */
	private File file;

	/** Spring transaction template */
	private TransactionTemplate tTemplate;

	/** Error counter */
	private int errors = 0;

	/** Skipped definitions counter */
	private int skipped = 0;

	/** Successful imports counter */
	private int successful = 0;

	/** Stored procedure that starts recording an event definition */
	private StoredProc fEvdefStart;

	/** Stored procedure that finalises an event definition's recording */
	private StoredProc fEvdefFinalise;

	/** Stored procedure that defines a boolean field */
	private StoredProc fEvdefAddfldBool;

	/** Stored procedure that defines an I18N reference field */
	private StoredProc fEvdefAddfldI18N;

	/** Stored procedure that defines an entity reference field */
	private StoredProc fEvdefAddfldEntity;

	/** Unbounded integer field definition procedure */
	private StoredProc fEvdefAddfldInt;

	/** Stored procedure that defines an integer field with a lower bound */
	private StoredProc fEvdefAddfldIntMin;

	/** Stored procedure that defines an integer field with a higher bound */
	private StoredProc fEvdefAddfldIntMax;

	/** Stored procedure that defines a bounded integer field */
	private StoredProc fEvdefAddfldIntRange;

	/** Unbounded double precision field definition procedure */
	private StoredProc fEvdefAddfldReal;

	/** Stored procedure that defines a double precision field with a lower bound */
	private StoredProc fEvdefAddfldRealMin;

	/** Stored procedure that defines a double precision field with a higher bound */
	private StoredProc fEvdefAddfldRealMax;

	/** Stored procedure that defines a bounded double precision field */
	private StoredProc fEvdefAddfldRealRange;

	/** Unbounded text field definition procedure */
	private StoredProc fEvdefAddfldText;

	/** Stored procedure that defines a text field with a minimum length */
	private StoredProc fEvdefAddfldTextMin;

	/** Stored procedure that defines a text field with a maximal length */
	private StoredProc fEvdefAddfldTextMax;

	/** Stored procedure that defines a bounded text field */
	private StoredProc fEvdefAddfldTextRange;


	/**
	 * Obtain the name of the definitions file
	 * 
	 * <p>
	 * Check the command line options, setting the definitions file accordingly.
	 */
	@Override
	public boolean setOptions( String... options )
	{
		if ( options.length != 1 ) {
			return false;
		}
		this.file = new File( options[ 0 ] );
		if ( ! ( this.file.isFile( ) && this.file.canRead( ) ) ) {
			return false;
		}
		return true;
	}


	/**
	 * Run the event definitions import tool
	 * 
	 * <p>
	 * Loads the data file, then connects to the database and creates all missing definitions.
	 */
	@Override
	public void run( )
	{
		EventDefinitions data;
		try {
			data = new EventDefinitionLoader( this.file ).load( );
		} catch ( DataImportException e ) {
			this.logger.error( "Error while loading event definitions" , e );
			return;
		}

		AbstractApplicationContext ctx = this.createContext( );
		this.createTemplates( ctx );
		this.executeImportTransaction( data );
		ToolBase.destroyContext( ctx );
	}


	/**
	 * Create the Spring context
	 * 
	 * <p>
	 * Load the definition of the data source as a Spring context, then adds the transaction
	 * management component.
	 * 
	 * @return the Spring application context
	 */
	private ClassPathXmlApplicationContext createContext( )
	{
		FileSystemXmlApplicationContext ctx = new FileSystemXmlApplicationContext( new String[] {
			this.getDataSource( )
		} );
		ctx.refresh( );

		return new ClassPathXmlApplicationContext( new String[] {
			"configuration/transactions.xml"
		} , true , ctx );
	}


	/**
	 * Create database access templates
	 * 
	 * <p>
	 * Initialise the transaction template and the four stored procedure definitions.
	 * 
	 * @param ctx
	 *            the Spring application context
	 */
	private void createTemplates( ApplicationContext ctx )
	{
		DataSource dSource = ctx.getBean( DataSource.class );
		PlatformTransactionManager tManager = ctx.getBean( PlatformTransactionManager.class );

		this.fEvdefStart = new StoredProc( dSource , "events" , "evdef_start" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_prio" , java.sql.Types.INTEGER )
				.addParameter( "_adj" , java.sql.Types.BOOLEAN ).addParameter( "_i18n" , java.sql.Types.VARCHAR )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefFinalise = new StoredProc( dSource , "events" , "evdef_finalise" ).addOutput( "_error" ,
				java.sql.Types.OTHER ).addOutput( "_fld" , java.sql.Types.VARCHAR );

		this.fEvdefAddfldBool = new StoredProc( dSource , "events" , "evdef_addfld_bool" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefAddfldI18N = new StoredProc( dSource , "events" , "evdef_addfld_i18n" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefAddfldEntity = new StoredProc( dSource , "events" , "evdef_addfld_entity" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_etype" , "events.entity_field_type" ).addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefAddfldInt = new StoredProc( dSource , "events" , "evdef_addfld_int" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldIntMin = new StoredProc( dSource , "events" , "evdef_addfld_int_min" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.INTEGER ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldIntMax = new StoredProc( dSource , "events" , "evdef_addfld_int_max" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_max" , java.sql.Types.INTEGER ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldIntRange = new StoredProc( dSource , "events" , "evdef_addfld_int_range" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.INTEGER ).addParameter( "_max" , java.sql.Types.INTEGER )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefAddfldReal = new StoredProc( dSource , "events" , "evdef_addfld_real" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldRealMin = new StoredProc( dSource , "events" , "evdef_addfld_real_min" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.DOUBLE ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldRealMax = new StoredProc( dSource , "events" , "evdef_addfld_real_max" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_max" , java.sql.Types.DOUBLE ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldRealRange = new StoredProc( dSource , "events" , "evdef_addfld_real_range" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.DOUBLE ).addParameter( "_max" , java.sql.Types.DOUBLE )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.fEvdefAddfldText = new StoredProc( dSource , "events" , "evdef_addfld_text" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldTextMin = new StoredProc( dSource , "events" , "evdef_addfld_text_min" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.INTEGER ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldTextMax = new StoredProc( dSource , "events" , "evdef_addfld_text_max" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_max" , java.sql.Types.INTEGER ).addOutput( "_error" , java.sql.Types.OTHER );
		this.fEvdefAddfldTextRange = new StoredProc( dSource , "events" , "evdef_addfld_text_range" )
				.addParameter( "_id" , java.sql.Types.VARCHAR ).addParameter( "_optional" , java.sql.Types.BOOLEAN )
				.addParameter( "_min" , java.sql.Types.INTEGER ).addParameter( "_max" , java.sql.Types.INTEGER )
				.addOutput( "_error" , java.sql.Types.OTHER );

		this.tTemplate = new TransactionTemplate( tManager );
	}


	/**
	 * Execute the event definitions importation transaction
	 * 
	 * <p>
	 * Run a transaction and execute the importation code inside it. Roll back if anything goes
	 * wrong.
	 * 
	 * @param data
	 *            the event definitions top-level instance
	 */
	private void executeImportTransaction( final EventDefinitions data )
	{
		boolean rv = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {
			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				boolean rv = ImportEvents.this.doTransaction( data );
				if ( !rv ) {
					status.setRollbackOnly( );
				}
				return rv;
			}
		} );

		if ( rv ) {
			this.logger.info( "Import complete; " + this.successful + " new type(s), " + this.skipped + " skipped" );
		} else {
			this.logger.error( this.errors + " error(s) occurred, no changes were made" );
		}
	}


	/**
	 * Import transaction body
	 * 
	 * <p>
	 * Import all definitions and handle exceptions.
	 * 
	 * @param data
	 *            the event definitions top-level instance
	 * @return <code>true</code> on success, <code>false</code> otherwise
	 */
	private boolean doTransaction( EventDefinitions data )
	{
		try {
			this.importDefinitions( data );
			return this.errors == 0;
		} catch ( RuntimeException e ) {
			this.logger.error( "Caught runtime exception" , e );
			this.errors ++;
		}
		return false;
	}


	/**
	 * Extract the error code from a stored procedure's return value
	 * 
	 * <p>
	 * This method is used to extract the error code from the maps returned by a definition stored
	 * procedure's execution.
	 * 
	 * @param result
	 *            the map returned by the stored procedure's execution
	 * 
	 * @return the error code
	 */
	private SPErrorCode getSPResult( Map< String , Object > result )
	{
		return SPErrorCode.valueOf( ( (PGobject) result.get( "_error" ) ).getValue( ) );
	}


	/**
	 * Import all event definitions from the top-level importable instance
	 * 
	 * <p>
	 * This method iterates through all of the definitions, trying to import each of them.
	 * 
	 * @param data
	 *            the top-level importable instance
	 */
	private void importDefinitions( EventDefinitions data )
	{
		for ( EventDefinition evdef : data ) {
			this.importDefinition( evdef );
		}
	}


	/**
	 * Try to import an event definition
	 * 
	 * <p>
	 * This method will attempt to import an event definition. It will increase the state counters (
	 * {@link #errors}, {@link #skipped} and {@link #successful}) depending on the results.
	 * 
	 * @param evdef
	 *            the event definition to import
	 */
	private void importDefinition( EventDefinition evdef )
	{
		boolean ok;

		// Start the import
		try {
			this.startImport( evdef );
			ok = true;
		} catch ( DataImportException e ) {
			this.logger.error( e.getMessage( ) );
			ok = false;
		}

		// Import all fields
		if ( ok ) {
			for ( EventFieldBase field : evdef ) {
				ok = ok && this.importField( evdef , field );
				if ( !ok ) {
					break;
				}
			}
		}

		// Finalise the operation
		boolean wasNew = false;
		try {
			wasNew = this.endImport( evdef );
		} catch ( DataImportException e ) {
			if ( ok ) {
				this.logger.error( e.getMessage( ) );
				ok = false;
			}
		}

		// Update counters and print to log
		if ( ok ) {
			if ( wasNew ) {
				this.logger.info( evdef.getIdentifier( ) + ": new event definition imported" );
				this.successful++;
			} else {
				this.logger.warn( evdef.getIdentifier( ) + ": event definition exists (skipped)" );
				this.skipped++;
			}
		} else {
			this.errors++;
		}
	}


	/**
	 * Start importing an event definition
	 * 
	 * <p>
	 * This method calls the <code>events.evdef_start()</code> stored procedure to initialise the
	 * event definition registration, and handles its return value.
	 * 
	 * @param evdef
	 *            the event definition instance from the XML definition
	 * 
	 * @throws DataImportException
	 *             if the definition's identifier is invalid or if one of the internationalised
	 *             strings is missing
	 */
	private void startImport( EventDefinition evdef )
			throws DataImportException
	{
		SPErrorCode ec;
		ec = this.getSPResult( this.fEvdefStart.execute( evdef.getIdentifier( ) , evdef.getPriority( ) ,
				evdef.isAdjustable( ) , evdef.getI18NStrings( ) ) );

		switch ( ec ) {
			case OK:
				break;

			case BAD_ID:
				throw new DataImportException( evdef.getIdentifier( ) + ": bad event type identifier" );

			case BAD_STRINGS:
				throw new DataImportException( evdef.getIdentifier( ) + ": I18N strings not found" );

			default:
				throw new RuntimeException( "Unsupported return value for events.evdef_start(): " + ec.toString( ) );
		}
	}


	/**
	 * Import a field definition
	 * 
	 * <p>
	 * Call the appropriate field definition importation method depending on the field's type.
	 * 
	 * @param evdef
	 *            the event definition
	 * @param field
	 *            the field definition
	 * 
	 * @return <code>true</code> on success, <code>false</code> if something went wrong
	 */
	private boolean importField( EventDefinition evdef , EventFieldBase field )
	{
		SPErrorCode ec;
		switch ( field.getType( ) ) {
			case BOOLEAN:
				ec = this.importBooleanField( field );
				break;

			case I18N:
				ec = this.importI18NField( field );
				break;

			case ENTITY:
				ec = this.importEntityField( (EntityEventField) field );
				break;

			case INTEGER:
				ec = this.importIntegerField( (IntegerEventField) field );
				break;

			case REAL:
				ec = this.importRealField( (RealEventField) field );
				break;

			case TEXT:
				ec = this.importTextField( (TextEventField) field );
				break;

			default:
				throw new RuntimeException( "Unsupported field type " + field.getType( ).toString( ) );
		}

		switch ( ec ) {
			case OK:
				return true;

			case BAD_ID:
				this.logger.error( evdef.getIdentifier( ) + ": bad field identifier '" + field.getIdentifier( ) + "'" );
				break;

			default:
				throw new RuntimeException( "Unsupported return value for field definition procedure: " + ec.toString( ) );
		}
		return false;
	}


	/**
	 * Import a boolean field
	 * 
	 * <p>
	 * This method calls <code>events.evdef_addfld_bool</code> to register a new boolean field.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importBooleanField( EventFieldBase field )
	{
		return this.getSPResult( this.fEvdefAddfldBool.execute( field.getIdentifier( ) , !field.isRequired( ) ) );
	}


	/**
	 * Import an I18N reference field
	 * 
	 * <p>
	 * This method calls <code>events.evdef_addfld_i18n</code> to register a new I18N reference
	 * field.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importI18NField( EventFieldBase field )
	{
		return this.getSPResult( this.fEvdefAddfldI18N.execute( field.getIdentifier( ) , !field.isRequired( ) ) );
	}


	/**
	 * Import an entity reference field
	 * 
	 * <p>
	 * This method calls <code>events.evdef_addfld_entity</code> to register a new entity reference
	 * field.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importEntityField( EntityEventField field )
	{
		return this.getSPResult( this.fEvdefAddfldEntity.execute( field.getIdentifier( ) , !field.isRequired( ) ,
				field.getEntityType( ) ) );
	}


	/**
	 * Import an integer field definition
	 * 
	 * <p>
	 * This method calls one of the integer field definitions stored procedures, depending on the
	 * definition's parameters.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importIntegerField( IntegerEventField field )
	{
		String id = field.getIdentifier( );
		boolean opt = !field.isRequired( );
		Integer min = field.getMin( );
		Integer max = field.getMax( );

		Map< String , Object > result;
		if ( min == null && max == null ) {
			result = this.fEvdefAddfldInt.execute( id , opt );
		} else if ( min == null ) {
			result = this.fEvdefAddfldIntMax.execute( id , opt , max );
		} else if ( max == null ) {
			result = this.fEvdefAddfldIntMin.execute( id , opt , min );
		} else {
			result = this.fEvdefAddfldIntRange.execute( id , opt , min , max );
		}

		return this.getSPResult( result );
	}


	/**
	 * Import a double precision field definition
	 * 
	 * <p>
	 * This method calls one of the double precision field definitions stored procedures, depending
	 * on the definition's parameters.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importRealField( RealEventField field )
	{
		String id = field.getIdentifier( );
		boolean opt = !field.isRequired( );
		Double min = field.getMin( );
		Double max = field.getMax( );

		Map< String , Object > result;
		if ( min == null && max == null ) {
			result = this.fEvdefAddfldReal.execute( id , opt );
		} else if ( min == null ) {
			result = this.fEvdefAddfldRealMax.execute( id , opt , max );
		} else if ( max == null ) {
			result = this.fEvdefAddfldRealMin.execute( id , opt , min );
		} else {
			result = this.fEvdefAddfldRealRange.execute( id , opt , min , max );
		}

		return this.getSPResult( result );
	}


	/**
	 * Import a text field definition
	 * 
	 * <p>
	 * This method calls one of the text field definitions stored procedures, depending on the
	 * definition's parameters.
	 * 
	 * @param field
	 *            the field's definition
	 * 
	 * @return the stored procedure's error code
	 */
	private SPErrorCode importTextField( TextEventField field )
	{
		String id = field.getIdentifier( );
		boolean opt = !field.isRequired( );
		Integer min = field.getMinLength( );
		Integer max = field.getMaxLength( );

		Map< String , Object > result;
		if ( min == null && max == null ) {
			result = this.fEvdefAddfldText.execute( id , opt );
		} else if ( min == null ) {
			result = this.fEvdefAddfldTextMax.execute( id , opt , max );
		} else if ( max == null ) {
			result = this.fEvdefAddfldTextMin.execute( id , opt , min );
		} else {
			result = this.fEvdefAddfldTextRange.execute( id , opt , min , max );
		}

		return this.getSPResult( result );
	}


	/**
	 * Finalise the current event definition
	 * 
	 * <p>
	 * This method calls the <code>events.evdef_finalise()</code> stored procedure, and handles its
	 * return values.
	 * 
	 * @param evdef
	 *            the event definition from the XML file
	 * 
	 * @return <code>true</code> if the definition was finalised or if it had no effect (except for
	 *         cleanup) due to a previous error, <code>false</code> if the definition already
	 *         existed and was skipped.
	 * 
	 * @throws DataImportException
	 *             if the event's specification is invalid, if a field name is duplicated, or if a
	 *             field's specification is invalid.
	 */
	private boolean endImport( EventDefinition evdef )
			throws DataImportException
	{
		Map< String , Object > result = this.fEvdefFinalise.execute( );
		SPErrorCode ec = this.getSPResult( result );
		String field = (String) result.get( "_fld" );

		switch ( ec ) {
			case OK:
			case NO_DATA:
				return true;

			case BAD_SPEC:
				if ( field == null ) {
					throw new DataImportException( evdef.getIdentifier( ) + ": bad event type specification" );
				}
				throw new DataImportException( evdef.getIdentifier( ) + ": bad field specification (" + field + ")" );

			case DUPLICATE:
				if ( field == null ) {
					return false;
				}
				throw new DataImportException( evdef.getIdentifier( ) + ": duplicate field name " + field );

			default:
				throw new RuntimeException( "Unsupported return value for events.evdef_finalise(): " + ec.toString( ) );
		}
	}

}
