package com.deepclone.lw.cli;


import java.io.File;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.postgresql.util.PGobject;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cli.xmlimport.ResourceLoader;
import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.resources.BasicResource;
import com.deepclone.lw.cli.xmlimport.data.resources.NaturalResource;
import com.deepclone.lw.cli.xmlimport.data.resources.Resources;
import com.deepclone.lw.cli.xmlimport.data.resources.UOCResourceErrorCode;
import com.deepclone.lw.utils.StoredProc;



/**
 * Resources import tool
 * 
 * <p>
 * This class defines the body of the resource import tool. It loads the data file specified on the
 * command line, validates it and then loads the resource definitions into the database.
 * 
 * @author <a href="mailto:tim@mitheren.com">T. Rosser</a>
 */
public class ImportResources
		extends CLITool
{
	/** Logging system */
	private final Logger logger = Logger.getLogger( ImportResources.class );

	/** File to read the definitions from */
	private File file;

	/** Spring transaction template */
	private TransactionTemplate tTemplate;

	/** Basic resource with category update or create stored procedure */
	private StoredProc uocBasicResourceWithCategory;

	/** Basic resource without category update or create stored procedure */
	private StoredProc uocBasicResourceWithoutCategory;

	/** Natural resource with category update or create stored procedure */
	private StoredProc uocNaturalResourceWithCategory;

	/** Natural resource without category update or create stored procedure */
	private StoredProc uocNaturalResourceWithoutCategory;


	/**
	 * Create the Spring context
	 * 
	 * <p>
	 * Load the definition of the data source as a Spring context, then adds the transaction
	 * management component.
	 * 
	 * @return the Spring application context
	 */
	private ClassPathXmlApplicationContext createContext( )
	{
		FileSystemXmlApplicationContext ctx = new FileSystemXmlApplicationContext( new String[] {
			this.getDataSource( )
		} );
		ctx.refresh( );

		return new ClassPathXmlApplicationContext( new String[] {
			"configuration/transactions.xml"
		} , true , ctx );
	}


	/**
	 * Create database access templates
	 * 
	 * <p>
	 * Initialise the transaction template and the four stored procedure definitions.
	 * 
	 * @param ctx
	 *            the Spring application context
	 */
	private void createTemplates( ApplicationContext ctx )
	{
		DataSource dSource = ctx.getBean( DataSource.class );
		PlatformTransactionManager tManager = ctx.getBean( PlatformTransactionManager.class );

		this.uocBasicResourceWithoutCategory = new StoredProc( dSource , "defs" , "uoc_resource" )
				.addParameter( "_name" , java.sql.Types.VARCHAR )
				.addParameter( "_description" , java.sql.Types.VARCHAR )
				.addParameter( "_weight" , java.sql.Types.INTEGER ).addOutput( "_return" , java.sql.Types.OTHER );

		this.uocBasicResourceWithCategory = new StoredProc( dSource , "defs" , "uoc_resource" )
				.addParameter( "_name" , java.sql.Types.VARCHAR )
				.addParameter( "_description" , java.sql.Types.VARCHAR )
				.addParameter( "_category" , java.sql.Types.VARCHAR ).addParameter( "_weight" , java.sql.Types.INTEGER )
				.addOutput( "_return" , java.sql.Types.OTHER );

		this.uocNaturalResourceWithoutCategory = new StoredProc( dSource , "defs" , "uoc_natural_resource" )
				.addParameter( "_name" , java.sql.Types.VARCHAR )
				.addParameter( "_description" , java.sql.Types.VARCHAR )
				.addParameter( "_weight" , java.sql.Types.INTEGER ).addParameter( "_presence" , java.sql.Types.DOUBLE )
				.addParameter( "_quantity_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_quantity_dev" , java.sql.Types.DOUBLE )
				.addParameter( "_difficulty_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_difficulty_dev" , java.sql.Types.DOUBLE )
				.addParameter( "_recovery_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_recovery_dev" , java.sql.Types.DOUBLE ).addOutput( "_return" , java.sql.Types.OTHER );

		this.uocNaturalResourceWithCategory = new StoredProc( dSource , "defs" , "uoc_natural_resource" )
				.addParameter( "_name" , java.sql.Types.VARCHAR )
				.addParameter( "_description" , java.sql.Types.VARCHAR )
				.addParameter( "_category" , java.sql.Types.VARCHAR ).addParameter( "_weight" , java.sql.Types.INTEGER )
				.addParameter( "_presence" , java.sql.Types.DOUBLE )
				.addParameter( "_quantity_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_quantity_dev" , java.sql.Types.DOUBLE )
				.addParameter( "_difficulty_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_difficulty_dev" , java.sql.Types.DOUBLE )
				.addParameter( "_recovery_avg" , java.sql.Types.DOUBLE )
				.addParameter( "_recovery_dev" , java.sql.Types.DOUBLE ).addOutput( "_return" , java.sql.Types.OTHER );

		this.tTemplate = new TransactionTemplate( tManager );
	}


	/**
	 * Import all resource definitions
	 * 
	 * <p>
	 * Import all resource definitions from the top-level Resources data instance based on the type
	 * of resource, and whether or not it has a category.
	 * 
	 * @param data
	 *            the top level Resources data instance
	 * 
	 * @throws DataImportException
	 *             when some external resource definition fails to load
	 */
	private void importResources( Resources data )
			throws DataImportException
	{
		for ( BasicResource br : data ) {
			if ( br instanceof NaturalResource ) {
				if ( br.getCategory( ) == null ) {
					this.importNaturalResourceWithoutCategory( (NaturalResource) br );
				} else {
					this.importNaturalResourceWithCategory( (NaturalResource) br );
				}
			} else {
				if ( br.getCategory( ) == null ) {
					this.importBasicResourceWithoutCategory( br );
				} else {
					this.importBasicResourceWithCategory( br );
				}
			}
		}
	}


	/**
	 * Import a Natural Resource
	 * 
	 * <p>
	 * Import a natural resource that does not have a defined category.
	 * 
	 * @param nr
	 *            the resources definition data
	 * 
	 * @throws DataImportException
	 *             when some external resource definition fails to load
	 */
	private void importNaturalResourceWithoutCategory( NaturalResource nr )
			throws DataImportException
	{
		UOCResourceErrorCode err = UOCResourceErrorCode.valueOf( ( (PGobject) this.uocNaturalResourceWithoutCategory
				.execute( nr.getName( ) , nr.getDescription( ) , nr.getWeight( ) , nr.getPresenceProbability( ) ,
						nr.getQuantity( ).getAverage( ) , nr.getQuantity( ).getDeviation( ) ,
						nr.getDifficulty( ).getAverage( ) , nr.getDifficulty( ).getDeviation( ) ,
						nr.getRecovery( ).getAverage( ) , nr.getRecovery( ).getDeviation( ) ).get( "_return" ) )
				.getValue( ) );

		if ( !err.equals( UOCResourceErrorCode.CREATED ) && !err.equals( UOCResourceErrorCode.UPDATED ) ) {
			throw new DataImportException( "uocNaturalResourceWithoutCategory returned error code " + err.toString( ) );
		}
	}


	/**
	 * Import a Natural Resource
	 * 
	 * <p>
	 * Import a natural resource that does have a defined category.
	 * 
	 * @param nr
	 *            the resources definition data
	 * 
	 * @throws DataImportException
	 *             when some external resource definition fails to load
	 */
	private void importNaturalResourceWithCategory( NaturalResource nr )
			throws DataImportException
	{
		UOCResourceErrorCode err = UOCResourceErrorCode.valueOf( ( (PGobject) this.uocNaturalResourceWithCategory
				.execute( nr.getName( ) , nr.getDescription( ) , nr.getCategory( ) , nr.getWeight( ) ,
						nr.getPresenceProbability( ) , nr.getQuantity( ).getAverage( ) ,
						nr.getQuantity( ).getDeviation( ) , nr.getDifficulty( ).getAverage( ) ,
						nr.getDifficulty( ).getDeviation( ) , nr.getRecovery( ).getAverage( ) ,
						nr.getRecovery( ).getDeviation( ) ).get( "_return" ) ).getValue( ) );

		if ( !err.equals( UOCResourceErrorCode.CREATED ) && !err.equals( UOCResourceErrorCode.UPDATED ) ) {
			throw new DataImportException( "uocNaturalResourceWithCategory returned error code " + err.toString( ) );
		}
	}


	/**
	 * Import a Basic Resource
	 * 
	 * <p>
	 * Import a basic resource that does not have a defined category.
	 * 
	 * @param br
	 *            the resources definition data
	 * 
	 * @throws DataImportException
	 *             when some external resource definition fails to load
	 */
	private void importBasicResourceWithoutCategory( BasicResource br )
			throws DataImportException
	{
		UOCResourceErrorCode err = UOCResourceErrorCode.valueOf( ( (PGobject) this.uocBasicResourceWithoutCategory
				.execute( br.getName( ) , br.getDescription( ) , br.getWeight( ) ).get( "_return" ) ).getValue( ) );

		if ( !err.equals( UOCResourceErrorCode.CREATED ) && !err.equals( UOCResourceErrorCode.UPDATED ) ) {
			throw new DataImportException( "uocBasicResourceWithoutCategory returned error code " + err.toString( ) );
		}
	}


	/**
	 * Import a Basic Resource
	 * 
	 * <p>
	 * Import a basic resource that does not have a defined category.
	 * 
	 * @param br
	 *            the resources definition data
	 * 
	 * @throws DataImportException
	 *             when some external resource definition fails to load
	 */
	private void importBasicResourceWithCategory( BasicResource br )
			throws DataImportException
	{
		UOCResourceErrorCode err = UOCResourceErrorCode
				.valueOf( ( (PGobject) this.uocBasicResourceWithCategory.execute( br.getName( ) , br.getDescription( ) ,
						br.getCategory( ) , br.getWeight( ) ).get( "_return" ) ).getValue( ) );

		if ( !err.equals( UOCResourceErrorCode.CREATED ) && !err.equals( UOCResourceErrorCode.UPDATED ) ) {
			throw new DataImportException( "uocBasicResourceWithCategory returned error code " + err.toString( ) );
		}
	}


	/**
	 * Run the resource definitions import tool
	 * 
	 * <p>
	 * Loads the data file, the connects to the database and creates or updates all definitions.
	 */
	@Override
	public void run( )
	{
		Resources data;
		try {
			data = new ResourceLoader( this.file ).load( );
		} catch ( DataImportException e ) {
			System.err.println( "Error while loading '" + this.file + "': " + e.getMessage( ) );
			this.logger.error( "Error while loading resources data" , e );
			return;
		}

		AbstractApplicationContext ctx = this.createContext( );
		this.createTemplates( ctx );
		this.executeImportTransaction( data );
		ToolBase.destroyContext( ctx );
	}


	/**
	 * Execute the resource definitions importation transaction
	 * 
	 * <p>
	 * Run a transaction and execute the importation code inside it. Roll back if anything goes
	 * wrong.
	 * 
	 * @param data
	 *            the Resources definitions instance
	 */
	private void executeImportTransaction( final Resources data )
	{
		boolean rv = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {
			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				boolean rv = ImportResources.this.doTransaction( data );
				if ( !rv ) {
					status.setRollbackOnly( );
				}
				return rv;
			}
		} );
		if ( rv ) {
			this.logger.info( "Resource import successful" );
		}
	}


	/**
	 * Import transaction body
	 * 
	 * <p>
	 * Import all definitions and handle exceptions.
	 * 
	 * @param data
	 *            the Resources definitions instance
	 * @return
	 */
	private boolean doTransaction( Resources data )
	{
		try {
			this.importResources( data );
			return true;
		} catch ( RuntimeException e ) {
			this.logger.error( "Caught runtime exception" , e );
		} catch ( DataImportException e ) {
			this.logger.error( "Error while importing resources" , e );
		}
		return false;
	}


	/**
	 * Obtain the name of the definitions file
	 * 
	 * <p>
	 * Check the command line options, setting the definitions file accordingly.
	 */
	@Override
	public boolean setOptions( String... options )
	{
		if ( options.length != 1 ) {
			return false;
		}
		this.file = new File( options[ 0 ] );
		if ( ! ( this.file.isFile( ) && this.file.canRead( ) ) ) {
			return false;
		}
		return true;
	}
}
