package com.deepclone.lw.cli;


import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.postgresql.util.PGobject;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cli.xmlimport.TechnologyLoader;
import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.techs.Technologies;
import com.deepclone.lw.cli.xmlimport.data.techs.Technology;
import com.deepclone.lw.cli.xmlimport.data.techs.TechnologyDefinitionResult;
import com.deepclone.lw.utils.StoredProc;



/**
 * Technology graph import tool
 * 
 * <p>
 * This class defines the body of the technology graph import tool. It loads the data file specified
 * on the command line, validates it and then loads the technology definitions and dependencies into
 * the database.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class ImportTechGraph
		extends CLITool
{
	/** Logging system */
	private final Logger logger = Logger.getLogger( ImportTechGraph.class );

	/** File to read the definitions from */
	private File file;

	/** Spring transaction template */
	private TransactionTemplate tTemplate;

	/** Technology creation or update stored procedure */
	private StoredProc fUocTechnology;

	/** Stored procedure that removes all dependencies from a technology */
	private StoredProc fTechdepClear;

	/** Stored procedure that adds a dependency to a technology */
	private StoredProc fTechdepAdd;


	/**
	 * Create the Spring context
	 * 
	 * <p>
	 * Load the definition of the data source as a Spring context, then adds the transaction
	 * management component.
	 * 
	 * @return the Spring application context
	 */
	private ClassPathXmlApplicationContext createContext( )
	{
		FileSystemXmlApplicationContext ctx = new FileSystemXmlApplicationContext( new String[] {
			this.getDataSource( )
		} );
		ctx.refresh( );

		return new ClassPathXmlApplicationContext( new String[] {
			"configuration/transactions.xml"
		} , true , ctx );
	}


	/**
	 * Create database access templates
	 * 
	 * <p>
	 * Initialise the transaction template and the stored procedure wrappers.
	 * 
	 * @param ctx
	 *            the Spring application context
	 */
	private void createTemplates( ApplicationContext ctx )
	{
		DataSource dSource = ctx.getBean( DataSource.class );
		PlatformTransactionManager tManager = ctx.getBean( PlatformTransactionManager.class );

		this.fUocTechnology = new StoredProc( dSource , "defs" , "uoc_technology" )
				.addParameter( "_name" , java.sql.Types.VARCHAR ).addParameter( "_category" , java.sql.Types.VARCHAR )
				.addParameter( "_discovery" , java.sql.Types.VARCHAR )
				.addParameter( "_description" , java.sql.Types.VARCHAR )
				.addParameter( "_price" , java.sql.Types.BIGINT ).addParameter( "_points" , java.sql.Types.BIGINT )
				.addOutput( "_result" , java.sql.Types.OTHER );

		this.fTechdepClear = new StoredProc( dSource , "defs" , "techdep_clear" ).addParameter( "_tech" ,
				java.sql.Types.VARCHAR );

		this.fTechdepAdd = new StoredProc( dSource , "defs" , "techdep_add" )
				.addParameter( "_dependent" , java.sql.Types.VARCHAR )
				.addParameter( "_dependency" , java.sql.Types.VARCHAR ).addOutput( "_result" , java.sql.Types.OTHER );

		this.tTemplate = new TransactionTemplate( tManager );
	}


	/**
	 * Import all technology definitions
	 * 
	 * <p>
	 * Import all technology definitions from the top-level {@link Technologies}, creating a map of
	 * dependencies. Once all definitions have been imported, add all dependencies.
	 * 
	 * @param data
	 *            the top level {@link Technologies} data instance
	 * 
	 * @throws DataImportException
	 *             when some technology definition or dependency cannot be imported
	 */
	private void importTechnologies( Technologies data )
			throws DataImportException
	{
		Map< String , List< String > > dependencies = new HashMap< String , List< String > >( );
		for ( Technology technology : data ) {
			this.importTechnologyDefinition( technology );
			dependencies.put( technology.getName( ) , technology.getDependencies( ) );
		}

		for ( Map.Entry< String , List< String > > techDeps : dependencies.entrySet( ) ) {
			for ( String dependency : techDeps.getValue( ) ) {
				this.importDependency( techDeps.getKey( ) , dependency );
			}
		}
	}


	/**
	 * Import a technology definition
	 * 
	 * <p>
	 * Import the definition, and if there was already a definition by that name, clear its
	 * dependencies.
	 * 
	 * @param technology
	 *            the definition to import
	 * @throws DataImportException
	 *             when the update or creation stored procedure returns anything other than
	 *             <code>CREATED</code> or <code>UPDATED</code>
	 */
	private void importTechnologyDefinition( Technology technology )
			throws DataImportException
	{
		this.logger.info( "Importing technology '" + technology.getName( ) + "'" );

		TechnologyDefinitionResult result = this.getResultOf( this.fUocTechnology.execute( technology.getName( ) ,
				technology.getCategory( ) , technology.getDiscovery( ) , technology.getDescription( ) ,
				technology.getCost( ) , technology.getPoints( ) ) );
		if ( result == TechnologyDefinitionResult.UPDATED ) {
			this.fTechdepClear.execute( technology.getName( ) );
		} else if ( result != TechnologyDefinitionResult.CREATED ) {
			throw new DataImportException( "Error while importing technology '" + technology.getName( ) + "': "
					+ result );
		}
	}


	/**
	 * Import a dependency between two technologies
	 * 
	 * <p>
	 * Attempt to add a dependency from a technology to another.
	 * 
	 * @param dependent
	 *            the dependent technology
	 * @param dependency
	 *            the technology to depend on
	 * 
	 * @throws DataImportException
	 *             if dependency creation fails
	 */
	private void importDependency( String dependent , String dependency )
			throws DataImportException
	{
		this.logger.info( "Importing dependency from " + dependent + " to " + dependency );

		TechnologyDefinitionResult result = this.getResultOf( this.fTechdepAdd.execute( dependent , dependency ) );
		switch ( result ) {
			case CREATED:
				return;
			case REDUNDANT:
				throw new DataImportException( "( " + dependent + " -> " + dependency + " ) is redundant" );
			case BAD_STRINGS:
				throw new DataImportException( "( " + dependent + " -> " + dependency + " ): undefined technology" );
			case CYCLE:
				throw new DataImportException( "( " + dependent + " -> " + dependency + " ) would create a cycle" );
			default:
				throw new DataImportException( "( " + dependent + " -> " + dependency + " ): unexpected error code "
						+ result );
		}
	}


	/**
	 * Helper method for stored procedure results
	 * 
	 * <p>
	 * This method converts the map returned by one of the definition stored procedures into a
	 * {@link TechnologyDefinitionResult} value.
	 * 
	 * @param spResult
	 *            the return value of {@link StoredProc#execute(Object...)}
	 * 
	 * @return the converted value of the _result field
	 */
	private TechnologyDefinitionResult getResultOf( Map< String , Object > spResult )
	{
		return TechnologyDefinitionResult.valueOf( ( (PGobject) spResult.get( "_result" ) ).getValue( ) );
	}


	/**
	 * Run the technology definitions import tool
	 * 
	 * <p>
	 * Loads the data file, the connects to the database and creates or updates all definitions.
	 */
	@Override
	public void run( )
	{
		Technologies data;
		try {
			data = new TechnologyLoader( this.file ).load( );
		} catch ( DataImportException e ) {
			this.logger.error( "Error while loading technology definitions" , e );
			return;
		}

		AbstractApplicationContext ctx = this.createContext( );
		this.createTemplates( ctx );
		this.executeImportTransaction( data );
		ToolBase.destroyContext( ctx );
	}


	/**
	 * Execute the technology definitions importation transaction
	 * 
	 * <p>
	 * Run a transaction and execute the importation code inside it. Roll back if anything goes
	 * wrong.
	 * 
	 * @param data
	 *            the {@link Technologies} definitions instance
	 */
	private void executeImportTransaction( final Technologies data )
	{
		boolean rv = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {
			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				boolean rv = ImportTechGraph.this.doTransaction( data );
				if ( !rv ) {
					status.setRollbackOnly( );
				}
				return rv;
			}
		} );
		if ( rv ) {
			this.logger.info( "Technology import successful" );
		}
	}


	/**
	 * Import transaction body
	 * 
	 * <p>
	 * Import all definitions and handle exceptions.
	 * 
	 * @param data
	 *            the {@link Technologies} definitions instance
	 * @return
	 */
	private boolean doTransaction( Technologies data )
	{
		try {
			this.importTechnologies( data );
			return true;
		} catch ( RuntimeException e ) {
			this.logger.error( "Caught runtime exception" , e );
		} catch ( DataImportException e ) {
			this.logger.error( e.getMessage( ) );
		}
		return false;
	}


	/**
	 * Obtain the name of the definitions file
	 * 
	 * <p>
	 * Check the command line options, setting the definitions file accordingly.
	 */
	@Override
	public boolean setOptions( String... options )
	{
		if ( options.length != 1 ) {
			return false;
		}
		this.file = new File( options[ 0 ] );
		if ( ! ( this.file.isFile( ) && this.file.canRead( ) ) ) {
			return false;
		}
		return true;
	}
}
