package com.deepclone.lw.cli;


import java.io.File;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.TransactionCallback;
import org.springframework.transaction.support.TransactionTemplate;

import com.deepclone.lw.cli.xmlimport.I18NLoader;
import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.i18n.I18NText;
import com.deepclone.lw.cli.xmlimport.data.i18n.LanguageDefinition;
import com.deepclone.lw.cli.xmlimport.data.i18n.StringDefinition;
import com.deepclone.lw.utils.StoredProc;



/**
 * I18N text import tool
 * 
 * <p>
 * This class defines the body of the I18N text import tool. It loads the data file specified on the
 * command line, validates it, then proceeds to loading the languages and strings to the database.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
public class ImportText
		extends CLITool
{

	/** Logging system */
	private final Logger logger = Logger.getLogger( ImportText.class );

	/** File to read the definitions from */
	private File file;

	/** Spring transaction template */
	private TransactionTemplate tTemplate;

	/** Language creation or update stored procedure */
	private StoredProc uocLanguage;
	/** Translation creation or update stored procedure */
	private StoredProc uocTranslation;


	/**
	 * Create the Spring context
	 * 
	 * <p>
	 * Load the definition of the data source as a Spring context, then adds the transaction
	 * management component.
	 * 
	 * @return the Spring application context
	 */
	private ClassPathXmlApplicationContext createContext( )
	{
		FileSystemXmlApplicationContext ctx = new FileSystemXmlApplicationContext( new String[] {
			this.getDataSource( )
		} );
		ctx.refresh( );

		return new ClassPathXmlApplicationContext( new String[] {
			"configuration/transactions.xml"
		} , true , ctx );
	}


	/**
	 * Create database access templates
	 * 
	 * <p>
	 * Initialise the transaction template as well as both stored procedure definitions.
	 * 
	 * @param ctx
	 *            the Spring application context
	 */
	private void createTemplates( ApplicationContext ctx )
	{
		DataSource dSource = ctx.getBean( DataSource.class );
		PlatformTransactionManager tManager = ctx.getBean( PlatformTransactionManager.class );

		this.uocLanguage = new StoredProc( dSource , "defs" , "uoc_language" ).addParameter( "lid" ,
				java.sql.Types.VARCHAR ).addParameter( "lname" , java.sql.Types.VARCHAR );

		this.uocTranslation = new StoredProc( dSource , "defs" , "uoc_translation" )
				.addParameter( "lid" , java.sql.Types.VARCHAR ).addParameter( "sid" , java.sql.Types.VARCHAR )
				.addParameter( "trans" , java.sql.Types.VARCHAR );

		this.tTemplate = new TransactionTemplate( tManager );
	}


	/**
	 * Import all languages and strings
	 * 
	 * <p>
	 * Import all language and string definitions from the top-level I18N data instance.
	 * 
	 * @param data
	 *            the top-level I18N data instance
	 * 
	 * @throws DataImportException
	 *             when some external string definition fails to load
	 */
	private void importText( I18NText data )
			throws DataImportException
	{
		for ( LanguageDefinition ld : data ) {
			this.importLanguage( ld );
		}
	}


	/**
	 * Import a language definition and the translations it contains
	 * 
	 * <p>
	 * Import the language's definition, then iterate over all string definitions and import them as
	 * well.
	 * 
	 * @param ld
	 *            the language's definition data
	 * 
	 * @throws DataImportException
	 *             when some external string definition fails to load
	 */
	private void importLanguage( LanguageDefinition ld )
			throws DataImportException
	{
		// Try creating or updating the language
		this.uocLanguage.execute( ld.getId( ) , ld.getName( ) );

		// Import translations
		for ( StringDefinition sd : ld ) {
			this.uocTranslation.execute( ld.getId( ) , sd.getId( ) , sd.getString( ) );
		}
	}


	/**
	 * Run the I18N definitions import tool
	 * 
	 * <p>
	 * Load the data file, then connects to the database and create or update all definitions.
	 */
	@Override
	public void run( )
	{
		I18NText data;
		try {
			data = new I18NLoader( this.file ).load( );
		} catch ( DataImportException e ) {
			System.err.println( "Error while loading '" + this.file + "': " + e.getMessage( ) );
			this.logger.error( "Error while loading I18N data" , e );
			return;
		}

		AbstractApplicationContext ctx = this.createContext( );
		this.createTemplates( ctx );
		this.executeImportTransaction( data );
		ToolBase.destroyContext( ctx );
	}


	/**
	 * Execute the I18N definitions importation transaction
	 * 
	 * <p>
	 * Run a transaction and execute the importation code inside it. Roll back if anything goes
	 * awry.
	 * 
	 * @param data
	 *            the I18N definitions instance
	 */
	private void executeImportTransaction( final I18NText data )
	{
		boolean rv = this.tTemplate.execute( new TransactionCallback< Boolean >( ) {

			@Override
			public Boolean doInTransaction( TransactionStatus status )
			{
				boolean rv = ImportText.this.doTransaction( data );
				if ( !rv ) {
					status.setRollbackOnly( );
				}
				return rv;
			}

		} );
		if ( rv ) {
			this.logger.info( "Text import successful" );
		}
	}


	/**
	 * Import transaction body
	 * 
	 * <p>
	 * Import all definitions and handle exceptions.
	 * 
	 * @param data
	 *            the I18N definitions instance
	 * 
	 * @return <code>true</code> on success, <code>false</code> otherwise
	 */
	private boolean doTransaction( I18NText data )
	{
		try {
			this.importText( data );
			return true;
		} catch ( RuntimeException e ) {
			this.logger.error( "Caught runtime exception" , e );
		} catch ( DataImportException e ) {
			this.logger.error( "Error while importing external strings" , e );
		}
		return false;
	}


	/**
	 * Obtain the name of the definitions file
	 * 
	 * <p>
	 * Check the command line options, setting the definitions file accordingly.
	 */
	@Override
	public boolean setOptions( String... options )
	{
		if ( options.length != 1 ) {
			return false;
		}
		this.file = new File( options[ 0 ] );
		if ( ! ( this.file.isFile( ) && this.file.canRead( ) ) ) {
			return false;
		}
		return true;
	}
}
