package com.deepclone.lw.cli.xmlimport;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.i18n.FileString;
import com.deepclone.lw.cli.xmlimport.data.i18n.I18NText;
import com.deepclone.lw.cli.xmlimport.data.i18n.InlineString;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.XStreamException;



/**
 * I18N text definitions loader
 * 
 * <p>
 * This class can be used to load all I18N definitions. It will extract them from the XML file,
 * verify them and set their origin.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class I18NLoader
{
	/** The file to read the XML from */
	private final File file;


	/**
	 * Initialise the loader
	 * 
	 * @param file
	 *            the XML file that contains the definitions
	 */
	public I18NLoader( File file )
	{
		this.file = file.getAbsoluteFile( );
	}


	/**
	 * Initialise the necessary XStream instance
	 * 
	 * <p>
	 * Initialise the XStream instance by processing annotations in all I18N importable data
	 * classes.
	 * 
	 * @return the XStream instance to use when loading the data
	 */
	private XStream initXStream( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( I18NText.class );
		xstream.processAnnotations( InlineString.class );
		xstream.processAnnotations( FileString.class );
		return xstream;
	}


	/**
	 * Load the I18N definitions
	 * 
	 * <p>
	 * Load the XML file and process the definitions file using XStream.
	 * 
	 * @return the top-level importable data instance
	 * 
	 * @throws DataImportException
	 *             if reading from the file or parsing its contents fail
	 */
	private I18NText loadXMLFile( )
			throws DataImportException
	{
		FileInputStream fis;
		try {
			fis = new FileInputStream( this.file );
		} catch ( FileNotFoundException e ) {
			throw new DataImportException( "Unable to load I18N definitions" , e );
		}

		try {
			try {
				XStream xstream = this.initXStream( );
				return (I18NText) xstream.fromXML( fis );
			} finally {
				fis.close( );
			}
		} catch ( IOException e ) {
			throw new DataImportException( "Input error while loading I18N definitions" , e );
		} catch ( XStreamException e ) {
			throw new DataImportException( "XML error while loading I18N definitions" , e );
		}
	}


	/**
	 * Load and process I18N definition
	 * 
	 * <p>
	 * Attempt to load all I18N definitions, make sure they are valid, then set the original file's
	 * path.
	 * 
	 * @return the top-level importable data instance
	 * 
	 * @throws DataImportException
	 *             if loading or verifying the data fails
	 */
	public I18NText load( )
			throws DataImportException
	{
		I18NText text = this.loadXMLFile( );
		text.verifyData( );
		text.setReadFrom( this.file );
		return text;
	}

}
