package com.deepclone.lw.cli.xmlimport.data.i18n;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;



/**
 * An external string definition
 * 
 * <p>
 * This class corresponds to I18N string definitions which use external files to store the actual
 * string.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@SuppressWarnings( "serial" )
@XStreamAlias( "from-file" )
public class FileString
		extends StringDefinition
{
	/** The name of the file which contains the string's text */
	@XStreamAsAttribute
	private String source;


	/**
	 * Verify an external string definition
	 * 
	 * <p>
	 * Make sure that the definition actually includes the name of the file to load the string from.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		super.verifyData( );
		if ( this.source == null || "".equals( this.source.trim( ) ) ) {
			throw new DataImportException( "Missing file name for string definition '" + this.getId( ) + "'" );
		}
	}


	/**
	 * Load the string
	 * 
	 * <p>
	 * This implementation opens the file defined as the source of the string's text, reads it then
	 * returns it.
	 * 
	 * @throws DataImportException
	 *             if the file cannot be opened or read from
	 */
	@Override
	public String getString( )
			throws DataImportException
	{
		File source = getSourceFile( );

		StringBuilder sBuilder = new StringBuilder( );
		try {
			BufferedReader in = new BufferedReader( new FileReader( source ) );
			try {
				String str;
				while ( ( str = in.readLine( ) ) != null ) {
					sBuilder.append( str );
					sBuilder.append( "\n" );
				}
			} finally {
				in.close( );
			}
		} catch ( IOException e ) {
			throw new DataImportException( "Could not read from " + source.getPath( ) , e );
		}

		return sBuilder.toString( );
	}


	/**
	 * Access the source file
	 * 
	 * <p>
	 * Create the source file instance using the path to the data file the current instance was read
	 * from as the base directory.
	 * 
	 * @return the source file
	 */
	private File getSourceFile( )
	{
		File parentDirectory = this.getReadFrom( ).getParentFile( );
		if ( parentDirectory == null ) {
			parentDirectory = new File( "." );
		}
		return new File( parentDirectory , this.source ).getAbsoluteFile( );
	}
}
