package com.deepclone.lw.cli.xmlimport.data.i18n;


import java.io.File;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;
import com.thoughtworks.xstream.annotations.XStreamImplicit;



/**
 * A language definition
 * 
 * <p>
 * Language definitions for the I18N text importer possess an identifier and a name. In addition,
 * they may contain any amount of string definitions.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@SuppressWarnings( "serial" )
public class LanguageDefinition
		extends ImportableData
		implements Serializable , Iterable< StringDefinition >
{
	/** The language's identifier */
	@XStreamAsAttribute
	private String id;

	/** The language's name */
	@XStreamAsAttribute
	private String name;

	/** All strings defined for this language */
	@XStreamImplicit
	private final List< StringDefinition > strings = new LinkedList< StringDefinition >( );


	/**
	 * Set the source file
	 * 
	 * <p>
	 * Update the definition's own source file, then update all string definitions.
	 */
	@Override
	public void setReadFrom( File readFrom )
	{
		super.setReadFrom( readFrom );
		for ( StringDefinition definition : this.strings ) {
			definition.setReadFrom( readFrom );
		}
	}


	/**
	 * Validate the language definition
	 * 
	 * <p>
	 * Make sure that all definition characteristics are properly defined, then check all strings
	 * and make sure they are unique.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		if ( this.id == null || "".equals( this.id.trim( ) ) ) {
			throw new DataImportException( "Missing language identifier" );
		}
		if ( this.name == null || "".equals( this.name.trim( ) ) ) {
			throw new DataImportException( "Missing language name for identifier '" + this.id + "'" );
		}
		if ( this.strings == null || this.strings.isEmpty( ) ) {
			throw new DataImportException( "No strings defined for language '" + this.id + "'" );
		}

		this.checkStringDefinitions( );
	}


	/**
	 * Check all string definitions
	 * 
	 * <p>
	 * Iterate over all included string definitions, validating them and making sure they all have
	 * unique identifiers.
	 * 
	 * @throws DataImportException
	 *             if a string definition is invalid or if string identifiers contain duplicates.
	 */
	private void checkStringDefinitions( )
			throws DataImportException
	{
		HashSet< String > strings = new HashSet< String >( );
		for ( StringDefinition definition : this.strings ) {
			definition.verifyData( );

			String stringId = definition.getId( );
			if ( strings.contains( stringId ) ) {
				throw new DataImportException( "In language '" + this.id + "': duplicate string '" + stringId + "'" );
			}
			strings.add( stringId );
		}
	}


	/** @return the language's identifier */
	public String getId( )
	{
		return this.id;
	}


	/** @return the language's name */
	public String getName( )
	{
		return this.name;
	}


	/**
	 * Iterate over all string definitions
	 * 
	 * <p>
	 * This method grants access to the list of string definitions in read-only mode.
	 * 
	 * <p>
	 * <strong>Warning:</strong> this method should not be called if {@link #containsStrings()}
	 * returns <code>false</code>.
	 * 
	 * @return the read-only string definition iterator
	 */
	@Override
	public Iterator< StringDefinition > iterator( )
	{
		return Collections.unmodifiableList( this.strings ).iterator( );
	}

}
