package com.deepclone.lw.cli.xmlimport.data.i18n;


import java.io.Serializable;
import java.util.regex.Pattern;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;



/**
 * A string definition
 * 
 * <p>
 * This class is the base for both types of string definitions used by the initial I18N text
 * importer. String definitions always contain an identifier, and it is always possible to obtain
 * the actual text they contain.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@SuppressWarnings( "serial" )
public abstract class StringDefinition
		extends ImportableData
		implements Serializable
{
	/** Pattern followed by string identifiers */
	private static final Pattern identifierCheck = Pattern.compile( "^[a-z][a-z0-9]*([A-Z][a-z0-9]*)*$" );

	/** The string's identifier */
	@XStreamAsAttribute
	private String id;


	/**
	 * Check a string definition's identifier
	 * 
	 * <p>
	 * Make sure that a string definition's identifier is both present and valid.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		if ( this.id == null ) {
			throw new DataImportException( "Missing string identifier" );
		} else if ( !identifierCheck.matcher( this.id ).find( ) ) {
			throw new DataImportException( "Invalid string identifier '" + this.id + "'" );
		}
	}


	/** @return the string's identifier */
	public final String getId( )
	{
		return this.id;
	}


	/**
	 * This method must be overridden to provide a way of reading the actual text of the string that
	 * is being defined.
	 * 
	 * @return the string's text
	 * 
	 * @throws DataImportException
	 *             if accessing the string's actual text fails
	 */
	public abstract String getString( )
			throws DataImportException;

}
