package com.deepclone.lw.cli.xmlimport.data.resources;


import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;



/**
 * A Basic Resource
 * 
 * <p>
 * This class is the base for all resource classes providing a name, description, category and
 * weight. In certain cases the category can be null. It provides a method to verify that data has
 * been imported successfully.
 * 
 * @author <a href="mailto:tim@mitheren.com">T. Rosser</a>
 */
@SuppressWarnings( "serial" )
@XStreamAlias( "basic-resource" )
public class BasicResource
		extends ImportableData
{
	/** The resource's name */
	@XStreamAsAttribute
	private String name;

	/** The resource's description */
	@XStreamAsAttribute
	private String description;

	/**
	 * The resource's category
	 * 
	 * <p>
	 * Can be null for resources that do not belong to a category
	 */
	@XStreamAsAttribute
	private String category;

	/**
	 * The resource's weight
	 * 
	 * <p>
	 * Used when sorting resources and resource categories.
	 */
	@XStreamAsAttribute
	private Integer weight; // Is int enough?


	/**
	 * Check a resource's data
	 * 
	 * <p>
	 * Make sure that a resource's properties are both present and valid.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		if ( this.name == null || "".equals( this.name.trim( ) ) ) {
			throw new DataImportException( "Missing name string" );
		}
		if ( this.description == null || "".equals( this.description.trim( ) ) ) {
			throw new DataImportException( "Missing description string" );
		}
		if ( this.category != null && "".equals( this.category.trim( ) ) ) {
			throw new DataImportException( "Category invalid" );
		}
		if ( this.weight == null ) {
			throw new DataImportException( "Missing weight" );
		}
	}


	/** @return the resource's name */
	public String getName( )
	{
		return this.name;
	}


	/** @return the resource's description */
	public String getDescription( )
	{
		return this.description;
	}


	/** @return the resource's category */
	public String getCategory( )
	{
		return this.category;
	}


	/** @return the resource's weight */
	public Integer getWeight( )
	{
		return this.weight;
	}

}
