package com.deepclone.lw.cli.xmlimport.data.resources;


import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;



/**
 * A Natural Resource
 * 
 * <p>
 * This class represents naturally occurring resources. As well as extending {@link BasicResource}
 * it contains the presence-probability, quantity, difficulty and recovery of a resource and a
 * method to verify the data of these properties.
 * 
 * @author <a href="mailto:tim@mitheren.com">T. Rosser</a>
 */
@SuppressWarnings( "serial" )
@XStreamAlias( "natural-resource" )
public class NaturalResource
		extends BasicResource
{
	/**
	 * The resource's presence-probability
	 * 
	 * <p>
	 * Used by the universe generator to decide whether to add this type of resource to a particular
	 * planet or not.
	 */
	@XStreamAlias( "presence-probability" )
	@XStreamAsAttribute
	private Double presenceProbability;

	/** The resource's quantity */
	@XStreamAlias( "quantity" )
	private ResourceParameter quantity;

	/**
	 * The resource's difficulty
	 * 
	 * <p>
	 * The difficulty of extracting the resource.
	 */
	@XStreamAlias( "difficulty" )
	private ResourceParameter difficulty;

	/**
	 * The resource's recovery
	 * 
	 * <p>
	 * The recovery rate of the resource.
	 */
	@XStreamAlias( "recovery" )
	private ResourceParameter recovery;


	/**
	 * Check a resource's data
	 * 
	 * <p>
	 * Make sure that a resource's properties are both present and valid. Calls the base class'
	 * verifyData() method to check all of the data.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		super.verifyData( );

		if ( this.presenceProbability == null ) {
			throw new DataImportException( "Missing presence-probability" );
		}
		if ( this.quantity == null ) {
			throw new DataImportException( "Missing quantity" );
		}
		this.quantity.verifyData( "quantity" );

		if ( this.difficulty == null ) {
			throw new DataImportException( "Missing difficulty" );
		}
		this.difficulty.verifyData( "difficulty" );

		if ( this.recovery == null ) {
			throw new DataImportException( "Missing recovery" );
		}
		this.recovery.verifyData( "recovery" );

	}


	/** @return the resource's presence-probability */
	public Double getPresenceProbability( )
	{
		return this.presenceProbability;
	}


	/** @return the resource's quantity */
	public ResourceParameter getQuantity( )
	{
		return this.quantity;
	}


	/** @return the resource's difficulty */
	public ResourceParameter getDifficulty( )
	{
		return this.difficulty;
	}


	/** @return the resource's recovery */
	public ResourceParameter getRecovery( )
	{
		return this.recovery;
	}
}
