package com.deepclone.lw.cli.xmlimport.data.resources;


import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamImplicit;



/**
 * Resource Data
 * 
 * <p>
 * This class represents the contents of the resource text file. It contains a list of resource
 * definitions.
 * 
 * @author <a href="mailto:tim@mitheren.com">T. Rosser</a>
 */
@SuppressWarnings( "serial" )
@XStreamAlias( "lw-resources" )
public class Resources
		extends ImportableData
		implements Iterable< BasicResource >
{
	/** All present resource definitions */
	@XStreamImplicit
	private final List< BasicResource > resources = new LinkedList< BasicResource >( );


	/**
	 * Checks the resource data
	 * 
	 * <p>
	 * Checks each definition and ensures they are all unique.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		if ( this.resources == null ) {
			throw new DataImportException( "No resource definitions" );
		}

		HashSet< String > names = new HashSet< String >( );
		for ( BasicResource resource : this.resources ) {
			resource.verifyData( );
			this.checkUniqueItem( names , resource.getName( ) );
		}
	}


	/**
	 * Checks that the name of the resource is unique
	 * 
	 * <p>
	 * This helper method is used by {@link #verifyData()} to make sure resource names are unique.
	 * 
	 * @param existing
	 *            the existing set of items
	 * @param value
	 *            the item's value
	 * 
	 * @throws DataImportException
	 *             if the item's value is already present in the set of existing items
	 */
	public void checkUniqueItem( HashSet< String > existing , String value )
			throws DataImportException
	{
		if ( existing.contains( value ) ) {
			throw new DataImportException( "Duplicate resource name '" + value + "'" );
		}
		existing.add( value );
	}


	/**
	 * Resource definition iterator
	 * 
	 * <p>
	 * Grant access to the list of resources in read-only mode.
	 * 
	 * @return a read-only iterator on the list of resources.
	 */
	@Override
	public Iterator< BasicResource > iterator( )
	{
		return Collections.unmodifiableList( this.resources ).iterator( );
	}
}
