package com.deepclone.lw.cli.xmlimport.data.techs;


import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;
import com.thoughtworks.xstream.annotations.XStreamImplicit;



/**
 * A technology definition
 * 
 * <p>
 * This class represents a technology definition that can be imported from the data file. It
 * contains all required technology attributes, as well as a list of strings for the dependencies.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
@SuppressWarnings( "serial" )
public class Technology
		extends ImportableData
{

	/** Identifier of the name string */
	@XStreamAsAttribute
	private String name;

	/** Identifier of the category string */
	@XStreamAsAttribute
	private String category;

	/** Identifier of the description string */
	@XStreamAsAttribute
	private String description;

	/** Identifier of the string to display on technology discovery */
	@XStreamAsAttribute
	private String discovery;

	/** Monetary cost of the technology */
	@XStreamAsAttribute
	private Long cost;

	/** Required amount of research points */
	@XStreamAsAttribute
	private Long points;

	/** List of technology name string identifiers describing the dependencies */
	@XStreamImplicit( itemFieldName = "depends-on" )
	private final List< String > dependencies = new LinkedList< String >( );


	/**
	 * Technology definition verifier
	 * 
	 * <p>
	 * Make sure that all technology definition fields are both present and valid. If dependencies
	 * exist, also verify the names' validity.
	 */
	@Override
	public void verifyData( )
			throws DataImportException
	{
		if ( this.name == null || "".equals( this.name.trim( ) ) ) {
			throw new DataImportException( "Missing name string" );
		}
		if ( this.category == null || "".equals( this.category.trim( ) ) ) {
			throw new DataImportException( "Missing category string in " + this.name );
		}
		if ( this.description == null || "".equals( this.description.trim( ) ) ) {
			throw new DataImportException( "Missing description string in " + this.name );
		}
		if ( this.discovery == null || "".equals( this.discovery.trim( ) ) ) {
			throw new DataImportException( "Missing discovery string in " + this.name );
		}

		if ( this.points == null ) {
			throw new DataImportException( "Missing research points in " + this.name );
		} else if ( this.points <= 0 ) {
			throw new DataImportException( "Invalid research points in " + this.name );
		}

		if ( this.cost == null ) {
			throw new DataImportException( "Missing cost in " + this.name );
		} else if ( this.cost <= 0 ) {
			throw new DataImportException( "Invalid cost in " + this.name );
		}

		if ( this.dependencies != null ) {
			for ( String dep : this.dependencies ) {
				if ( dep == null || "".equals( dep.trim( ) ) ) {
					throw new DataImportException( "Empty dependency name in " + this.name );
				}
			}
		}
	}


	/**
	 * Gets the identifier of the name string.
	 * 
	 * @return the identifier of the name string
	 */
	public String getName( )
	{
		return this.name;
	}


	/**
	 * Gets the identifier of the category string.
	 * 
	 * @return the identifier of the category string
	 */
	public String getCategory( )
	{
		return this.category;
	}


	/**
	 * Gets the identifier of the description string.
	 * 
	 * @return the identifier of the description string
	 */
	public String getDescription( )
	{
		return this.description;
	}


	/**
	 * Gets the identifier of the string to display on technology discovery.
	 * 
	 * @return the identifier of the string to display on technology discovery
	 */
	public String getDiscovery( )
	{
		return this.discovery;
	}


	/**
	 * Gets the monetary cost of the technology.
	 * 
	 * @return the monetary cost of the technology
	 */
	public Long getCost( )
	{
		return this.cost;
	}


	/**
	 * Gets the required amount of research points.
	 * 
	 * @return the required amount of research points
	 */
	public Long getPoints( )
	{
		return this.points;
	}


	/**
	 * Gets the list of technology name string identifiers describing the dependencies.
	 * 
	 * @return the list of technology name string identifiers describing the dependencies
	 */
	public List< String > getDependencies( )
	{
		if ( this.dependencies == null ) {
			return Collections.emptyList( );
		}
		return Collections.unmodifiableList( this.dependencies );
	}

}
