package com.deepclone.lw.beans.bt.es.data;


import static org.junit.Assert.*;

import org.junit.Before;
import org.junit.Test;

import com.thoughtworks.xstream.XStream;



/**
 * Tests for the {@link EmpireResourceInformation} class.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class TestEmpireResourceInformation
{

	/** Resource identifier used in tests */
	private static final String TEST_ID = "Test";

	/** "Owed" value used in tests */
	private static final Double TEST_OWED = 1.0;

	/** "Possessed" value used in tests */
	private static final Double TEST_POSSESSED = 2.0;

	/** Mining priority value used in tests */
	private static final Integer TEST_PRIORITY = 3;

	/** Empire resource information instance used in tests */
	private EmpireResourceInformation empRes;


	/** Create the instance to be used in actual tests */
	@Before
	public void setUp( )
	{
		this.empRes = new EmpireResourceInformation( );
	}


	/** Default values are <code>null</code> */
	@Test
	public void testDefaultValues( )
	{
		assertNull( this.empRes.getResource( ) );
		assertNull( this.empRes.getOwed( ) );
		assertNull( this.empRes.getPossessed( ) );
		assertNull( this.empRes.getMiningPriority( ) );
	}


	/** Setting and reading the resource type */
	@Test
	public void testResource( )
	{
		this.empRes.setResource( TEST_ID );
		assertEquals( TEST_ID , this.empRes.getResource( ) );
	}


	/** Setting the resource type to <code>null</code> */
	@Test
	public void testNullResource( )
	{
		try {
			this.empRes.setResource( null );
			fail( "InvalidDumpContentsException expected" );
		} catch ( InvalidDumpContentsException exception ) {
			assertEquals( EmpireResourceInformation.class , exception.getRecordType( ) );
			assertEquals( "resource" , exception.getField( ) );
		}
	}


	/** Setting and reading the amount of resources possessed */
	@Test
	public void testPossessed( )
	{
		this.empRes.setPossessed( TEST_POSSESSED );
		assertEquals( TEST_POSSESSED , this.empRes.getPossessed( ) );
	}


	/** Setting and reading the amount of resources owed */
	@Test
	public void testOwed( )
	{
		this.empRes.setOwed( TEST_OWED );
		assertEquals( TEST_OWED , this.empRes.getOwed( ) );
	}


	/** Setting and reading the mining priority */
	@Test
	public void testMiningPriority( )
	{
		this.empRes.setMiningPriority( TEST_PRIORITY );
		assertEquals( TEST_PRIORITY , this.empRes.getMiningPriority( ) );
	}


	/** Serialising the instance to XML - no mining priority */
	@Test
	public void testXMLSerialisation( )
	{
		this.empRes.setResource( TEST_ID );
		this.empRes.setOwed( TEST_OWED );
		this.empRes.setPossessed( TEST_POSSESSED );

		String serialised = this.createXStreamInstance( ).toXML( this.empRes );
		assertNotNull( serialised );
		assertTrue( serialised.startsWith( "<resource " ) );
		assertTrue( serialised.endsWith( "/>" ) );
		assertTrue( serialised.contains( " id=\"" + TEST_ID + "\"" ) );
		assertTrue( serialised.contains( " owed=\"" + TEST_OWED + "\"" ) );
		assertTrue( serialised.contains( " possessed=\"" + TEST_POSSESSED + "\"" ) );
		assertFalse( serialised.contains( "mining-priority=\"" ) );
	}


	/** Serialising the instance to XML - including mining priority */
	@Test
	public void testXMLSerialisationWithPriority( )
	{
		this.empRes.setResource( TEST_ID );
		this.empRes.setOwed( TEST_OWED );
		this.empRes.setPossessed( TEST_POSSESSED );
		this.empRes.setMiningPriority( TEST_PRIORITY );

		String serialised = this.createXStreamInstance( ).toXML( this.empRes );
		assertNotNull( serialised );
		assertTrue( serialised.startsWith( "<resource " ) );
		assertTrue( serialised.endsWith( "/>" ) );
		assertTrue( serialised.contains( " id=\"" + TEST_ID + "\"" ) );
		assertTrue( serialised.contains( " owed=\"" + TEST_OWED + "\"" ) );
		assertTrue( serialised.contains( " possessed=\"" + TEST_POSSESSED + "\"" ) );
		assertTrue( serialised.contains( " mining-priority=\"" + TEST_PRIORITY + "\"" ) );
	}


	/** Deserialising an instance from XML - no mining priority */
	@Test
	public void testXMLDeserialisation( )
	{
		String xml = "<resource id=\"" + TEST_ID + "\" owed=\"" + TEST_OWED.toString( ) + "\" possessed=\""
				+ TEST_POSSESSED.toString( ) + "\" />";
		Object deserialised = this.createXStreamInstance( ).fromXML( xml );

		assertNotNull( deserialised );
		assertEquals( EmpireResourceInformation.class , deserialised.getClass( ) );
		this.empRes = (EmpireResourceInformation) deserialised;

		assertEquals( TEST_ID , this.empRes.getResource( ) );
		assertEquals( TEST_POSSESSED , this.empRes.getPossessed( ) );
		assertEquals( TEST_OWED , this.empRes.getOwed( ) );
		assertNull( this.empRes.getMiningPriority( ) );
	}


	/** Deserialising an instance from XML - including mining priority */
	@Test
	public void testXMLDeserialisationWithPriority( )
	{
		String xml = "<resource id=\"" + TEST_ID + "\" owed=\"" + TEST_OWED.toString( ) + "\" possessed=\""
				+ TEST_POSSESSED.toString( ) + "\" mining-priority=\"" + TEST_PRIORITY.toString( ) + "\" />";
		Object deserialised = this.createXStreamInstance( ).fromXML( xml );

		assertNotNull( deserialised );
		assertEquals( EmpireResourceInformation.class , deserialised.getClass( ) );
		this.empRes = (EmpireResourceInformation) deserialised;

		assertEquals( TEST_ID , this.empRes.getResource( ) );
		assertEquals( TEST_POSSESSED , this.empRes.getPossessed( ) );
		assertEquals( TEST_OWED , this.empRes.getOwed( ) );
		assertEquals( TEST_PRIORITY , this.empRes.getMiningPriority( ) );
	}


	/**
	 * Create and set up the {@link XStream} instance used in the serialisation tests
	 * 
	 * @return the {@link XStream} instance to use
	 */
	private XStream createXStreamInstance( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( EmpireResourceInformation.class );
		return xstream;
	}
}
