package com.deepclone.lw.beans.game.resources;


import com.deepclone.lw.interfaces.game.resources.MiningSettingsDAO;



/**
 * Mock mining settings DAO which can be used to simulate failures and trace which methods were
 * called.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
public class MockMiningSettingsDAO
		implements MiningSettingsDAO
{
	/**
	 * The empire identifier with which {@link #startUpdate(int)} or {@link #togglePlanet(int, int)}
	 * was called
	 */
	private Integer empire = null;

	/** The planet identifier with which {@link #togglePlanet(int, int)} was called */
	private Integer planet = null;

	/** The amount of calls to {@link #setNewPriority(String, int)} */
	private int callsToSet = 0;

	/** Whether {@link #startUpdate(int)} was called */
	private boolean startUpdateCalled = false;

	/** Whether {@link #applyUpdate()} was called */
	private boolean applyCalled = false;

	/** Whether {@link #togglePlanet(int, int)} was called */
	private boolean togglePlanetCalled = false;

	/** Whether {@link #startUpdate(int)} will succeed or fail */
	private boolean startUpdateSucceeds = true;

	/** Whether applyUpdate will succeed */
	private boolean applyUpdateSucceeds = true;

	/**
	 * Amount of calls to {@link #setNewPriority(String, int)} that will succeed, or
	 * <code>null</code> if they will all succeed.
	 */
	private Integer maxSetCalls = null;


	/** @return the empire identifier */
	public Integer getEmpire( )
	{
		return this.empire;
	}


	/** @return the planet identifier */
	public Integer getPlanet( )
	{
		return this.planet;
	}


	/** @return <code>true</code> if {@link #startUpdate(int)()} was called */
	public boolean wasStartUpdateCalled( )
	{
		return this.startUpdateCalled;
	}


	/** @return the amount of calls to {@link #setNewPriority(String, int)} */
	public int getCallsToSet( )
	{
		return this.callsToSet;
	}


	/** @return <code>true</code> if {@link #applyUpdate()} was called */
	public boolean wasApplyCalled( )
	{
		return this.applyCalled;
	}


	/** @return <code>true</code> if {@link #togglePlanet(int, int)} was called */
	public boolean wasTogllePlanetCalled( )
	{
		return this.togglePlanetCalled;
	}


	/**
	 * Determine whether calls to {@link #startUpdate(int)} will succeed
	 * 
	 * @param startUpdateSucceeds
	 *            <code>true</code> if the call is to succeed, <code>false</code> if it is to fail
	 */
	public void setStartUpdateSucceeds( boolean startUpdateSucceeds )
	{
		this.startUpdateSucceeds = startUpdateSucceeds;
	}


	/**
	 * Determine whether calls to {@link #applyUpdate()} will succeed
	 * 
	 * @param applyUpdateSucceeds
	 *            <code>true</code> if the call is to succeed, <code>false</code> if it is to fail
	 */
	public void setApplyUpdateSucceeds( boolean applyUpdateSucceeds )
	{
		this.applyUpdateSucceeds = applyUpdateSucceeds;
	}


	/**
	 * Set the amount of calls to {@link #setNewPriority(String, int)} that will succeed
	 * 
	 * @param maxSetCalls
	 *            the amount of calls that will succeed, or <code>null</code> if the method must
	 *            always succeed
	 */
	public void setMaxSetCalls( Integer maxSetCalls )
	{
		this.maxSetCalls = maxSetCalls;
	}


	@Override
	public boolean startUpdate( int empireId )
	{
		this.startUpdateCalled = true;
		this.empire = empireId;
		return this.startUpdateSucceeds;
	}


	@Override
	public boolean startUpdate( int empire , int planet )
	{
		this.startUpdateCalled = true;
		this.empire = empire;
		this.planet = planet;
		return this.startUpdateSucceeds;
	}


	@Override
	public boolean setNewPriority( String resource , int priority )
	{
		this.callsToSet++;
		return ( this.maxSetCalls == null || this.maxSetCalls > this.callsToSet - 1 );
	}


	@Override
	public boolean applyUpdate( )
	{
		this.applyCalled = true;
		return this.applyUpdateSucceeds;
	}


	@Override
	public void togglePlanet( int empire , int planet )
	{
		this.togglePlanetCalled = true;
		this.empire = empire;
		this.planet = planet;
	}

}
