package com.deepclone.lw.beans.game.resources;


import static org.junit.Assert.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.junit.Before;
import org.junit.Test;

import com.deepclone.lw.interfaces.game.resources.MiningSettingsDAO;



/**
 * Tests for {@link ResourcesControllerBean}
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
public class TestResourcesControllerBean
{

	/** Empire identifier used in the tests */
	private static final Integer EMPIRE_ID = 42;

	/** Planet identifier used in the tests */
	private static final Integer PLANET_ID = 43;

	/** Mining settings used in the tests */
	private static final Map< String , Integer > MINING_SETTINGS;

	/**
	 * Initialise the mining settings map
	 */
	static {
		HashMap< String , Integer > tempMap = new HashMap< String , Integer >( );
		tempMap.put( "test 1" , 1 );
		tempMap.put( "test 2" , 2 );
		tempMap.put( "test 3" , 3 );
		tempMap.put( "test 4" , 4 );
		MINING_SETTINGS = Collections.unmodifiableMap( tempMap );
	}

	/** The mock database access object */
	private MockMiningSettingsDAO miningSettingsDAO;

	/** The instance to test on */
	private ResourcesControllerBean ctrl;


	/**
	 * Create the resource controller that will be used in the tests and its various (usually fake)
	 * dependencies
	 */
	@Before
	public void setUp( )
	{
		this.miningSettingsDAO = new MockMiningSettingsDAO( );
		this.ctrl = new ResourcesControllerBean( );
		this.ctrl.setMiningSettingsDAO( this.miningSettingsDAO );
	}


	/**
	 * When calling {@link MiningSettingsDAO#startUpdate(int)} fails, the empire settings update is
	 * interrupted.
	 */
	@Test
	public void testEmpireStartUpdateFails( )
	{
		this.miningSettingsDAO.setStartUpdateSucceeds( false );
		this.ctrl.updateEmpireSettings( EMPIRE_ID , MINING_SETTINGS );
		assertTrue( this.miningSettingsDAO.wasStartUpdateCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( 0 , this.miningSettingsDAO.getCallsToSet( ) );
		assertFalse( this.miningSettingsDAO.wasApplyCalled( ) );
	}


	/**
	 * When calling {@link MiningSettingsDAO#startUpdate(int,int)} fails, the planet settings update
	 * is interrupted.
	 */
	@Test
	public void testPlanetStartUpdateFails( )
	{
		this.miningSettingsDAO.setStartUpdateSucceeds( false );
		this.ctrl.updatePlanetSettings( EMPIRE_ID , PLANET_ID , MINING_SETTINGS );
		assertTrue( this.miningSettingsDAO.wasStartUpdateCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( PLANET_ID , this.miningSettingsDAO.getPlanet( ) );
		assertEquals( 0 , this.miningSettingsDAO.getCallsToSet( ) );
		assertFalse( this.miningSettingsDAO.wasApplyCalled( ) );
	}


	/**
	 * When calling {@link MiningSettingsDAO#startUpdate(int)} succeeds but one of the calls to
	 * {@link MiningSettingsDAO#setNewPriority(String, int)} fails, the update is interrupted.
	 */
	@Test
	public void testSetMiningPriorityFails( )
	{
		this.miningSettingsDAO.setMaxSetCalls( 2 );
		this.ctrl.updateEmpireSettings( EMPIRE_ID , MINING_SETTINGS );
		assertTrue( this.miningSettingsDAO.wasStartUpdateCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( 3 , this.miningSettingsDAO.getCallsToSet( ) );
		assertFalse( this.miningSettingsDAO.wasApplyCalled( ) );
	}


	/**
	 * If both {@link MiningSettingsDAO#startUpdate(int)} and
	 * {@link MiningSettingsDAO#setNewPriority(String, int)} succeed,
	 * {@link MiningSettingsDAO#applyUpdate()} is called.
	 */
	@Test
	public void testSettingsSuccess( )
	{
		this.ctrl.updateEmpireSettings( EMPIRE_ID , MINING_SETTINGS );
		assertTrue( this.miningSettingsDAO.wasStartUpdateCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( 4 , this.miningSettingsDAO.getCallsToSet( ) );
		assertTrue( this.miningSettingsDAO.wasApplyCalled( ) );
	}


	/**
	 * A failure of {@link MiningSettingsDAO#applyUpdate()} has no influence on the call.
	 */
	@Test
	public void testSettingsApplyFail( )
	{
		this.miningSettingsDAO.setApplyUpdateSucceeds( false );
		this.ctrl.updateEmpireSettings( EMPIRE_ID , MINING_SETTINGS );
		assertTrue( this.miningSettingsDAO.wasStartUpdateCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( 4 , this.miningSettingsDAO.getCallsToSet( ) );
		assertTrue( this.miningSettingsDAO.wasApplyCalled( ) );
	}

	/**
	 * "Toggle planet source" calls {@link MiningSettingsDAO#togglePlanet(int, int)}
	 */
	@Test
	public void tesTogglePlanet()
	{
		this.ctrl.togglePlanet( EMPIRE_ID , PLANET_ID );
		assertTrue( this.miningSettingsDAO.wasTogllePlanetCalled( ) );
		assertEquals( EMPIRE_ID , this.miningSettingsDAO.getEmpire( ) );
		assertEquals( PLANET_ID , this.miningSettingsDAO.getPlanet( ) );
	}
}
