package com.deepclone.lw.cli.xmlimport;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;

import org.junit.Test;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.i18n.I18NText;
import com.deepclone.lw.cli.xmlimport.data.i18n.LanguageDefinition;
import com.thoughtworks.xstream.converters.ConversionException;
import com.thoughtworks.xstream.io.StreamException;



/**
 * Unit tests for the {@link I18NLoader} class.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
public class TestI18NLoader
{

	/**
	 * Try initialising the loader with a <code>null</code> file instance.
	 * 
	 * @throws NullPointerException
	 *             when the constructor is executed
	 */
	@Test( expected = NullPointerException.class )
	public void testNullFile( )
			throws NullPointerException
	{
		new I18NLoader( null );
	}


	/**
	 * Try loading a file that does not exist
	 */
	@Test
	public void testMissingFile( )
	{
		I18NLoader loader = new I18NLoader( new File( "does-not-exist" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a " + e.getCause( ).getClass( ).getName( ) , e.getCause( ) instanceof IOException );
			return;
		}
		fail( "no exception after trying to load a missing file" );
	}


	/**
	 * Try loading a file that contains something that definitely isn't XML.
	 */
	@Test
	public void testBadXML( )
	{
		I18NLoader loader = new I18NLoader( new File( "TestFiles/i18n-loader/bad-xml.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a " + e.getCause( ).getClass( ).getName( ) , e.getCause( ) instanceof StreamException );
			return;
		}
		fail( "no exception after loading stuff that isn't XML" );
	}


	/**
	 * Try loading a file that contains XML but which cannot be desertialised to an {@link I18NText}
	 * instance.
	 */
	@Test
	public void testBadContents( )
	{
		I18NLoader loader = new I18NLoader( new File( "TestFiles/i18n-loader/bad-contents.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a " + e.getCause( ).getClass( ).getName( ) ,
					e.getCause( ) instanceof ConversionException );
			return;
		}
		fail( "no exception after loading bad XML" );
	}


	/**
	 * Try loading a file that contains valid XML for an {@link I18NText} instance with semantic
	 * errors.
	 */
	@Test
	public void testBadData( )
	{
		I18NLoader loader = new I18NLoader( new File( "TestFiles/i18n-loader/bad-data.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertNull( e.getCause( ) );
			return;
		}
		fail( "no exception after loading bad data" );
	}


	/**
	 * Try loading valid data, make sure that it contains exactly what is expected, and that it
	 * loads external strings from the appropriate directory.
	 */
	@Test
	public void testGoodData( )
	{
		I18NLoader loader = new I18NLoader( new File( "TestFiles/i18n-loader/good.xml" ) );
		I18NText text;
		try {
			text = loader.load( );
		} catch ( DataImportException e ) {
			fail( "could not load valid file" );
			return;
		}
		assertNotNull( text );

		int lCount = 0;
		for ( @SuppressWarnings( "unused" )
		LanguageDefinition ld : text ) {
			lCount++;
		}
		assertEquals( 1 , lCount );
	}
}
