package com.deepclone.lw.cli.xmlimport;


import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;

import org.junit.Test;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;
import com.deepclone.lw.cli.xmlimport.data.techs.Technologies;
import com.deepclone.lw.cli.xmlimport.data.techs.Technology;
import com.thoughtworks.xstream.converters.ConversionException;
import com.thoughtworks.xstream.io.StreamException;



/**
 * Unit tests for the {@link TechnologyLoader} class.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class TestTechnologyLoader
{
	/**
	 * Try initialising the loader with a <code>null</code> file instance.
	 * 
	 * @throws NullPointerException
	 *             when the constructor is executed
	 */
	@Test( expected = NullPointerException.class )
	public void testNullFile( )
			throws NullPointerException
	{
		new TechnologyLoader( null );
	}


	/** Try loading a file that does not exist */
	@Test
	public void testMissingFile( )
	{
		TechnologyLoader loader = new TechnologyLoader( new File( "does-not-exist" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a" + e.getCause( ).getClass( ).getName( ) , e.getCause( ) instanceof IOException );
			return;
		}
		fail( "no exception after trying to load a missing file" );
	}


	/** Try loading a file that contains something that is not XML. */
	@Test
	public void testBadXML( )
	{
		TechnologyLoader loader = new TechnologyLoader( new File( "TestFiles/technology-loader/bad-xml.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a " + e.getCause( ).getClass( ).getName( ) , e.getCause( ) instanceof StreamException );
			return;
		}
		fail( "no exception after loading stuff that isn't XML" );
	}


	/**
	 * Test loading a file that contains XML but which cannot be deserialised to a
	 * {@link Technologies} instance.
	 */
	@Test
	public void testBadContents( )
	{
		TechnologyLoader loader = new TechnologyLoader( new File( "TestFiles/technology-loader/bad-contents.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertTrue( "cause is a " + e.getCause( ).getClass( ).getName( ) ,
					e.getCause( ) instanceof ConversionException );
			return;
		}
		fail( "no exception after loading bad XML" );
	}


	/**
	 * Try loading a file that contains valid XML for a {@link Technologies} instance with semantic
	 * errors.
	 */
	@Test
	public void testBadData( )
	{
		TechnologyLoader loader = new TechnologyLoader( new File( "TestFiles/technology-loader/bad-data.xml" ) );
		try {
			loader.load( );
		} catch ( DataImportException e ) {
			assertNull( e.getCause( ) );
			return;
		}
		fail( "no exception after loading bad data" );
	}


	/** Try loading valid data, make sure that it contains one record */
	@Test
	public void testGoodData( )
	{
		TechnologyLoader loader = new TechnologyLoader( new File( "TestFiles/technology-loader/good-data.xml" ) );
		Technologies technologies;
		try {
			technologies = loader.load( );
		} catch ( DataImportException e ) {
			fail( "could not load valid file" );
			return;
		}
		assertNotNull( technologies );

		int rCount = 0;
		for ( @SuppressWarnings( "unused" )
		Technology tech : technologies ) {
			rCount++;
		}
		assertEquals( 1 , rCount );
	}
}
