package com.deepclone.lw.cli.xmlimport.data.i18n;


import com.deepclone.lw.cli.xmlimport.I18NLoader;
import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.XStreamException;



/**
 * Base class for I18N import structures
 * 
 * <p>
 * This class is used as a parent by tests of the I18N import structures. It includes the code
 * required to actually create such resources, as this can normally only be done through XStream.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */
abstract class BaseTest
{

	/**
	 * Escape &amp;, &lt; and &gt; in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String xmlString( String string )
	{
		return string.replace( "&" , "&amp;" ).replace( "<" , "&lt;" ).replace( ">" , "&gt;" );
	}


	/**
	 * Escape &amp;, &lt; and &gt;, ' and " in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String quoteString( String string )
	{
		return "\"" + this.xmlString( string ).replace( "\"" , "&quot;" ).replace( "'" , "&apos;" ) + "\"";
	}


	/**
	 * Create an in-line string definition
	 * 
	 * <p>
	 * Generate the XML code that corresponds to an in-line string definition.
	 * 
	 * @param identifier
	 *            the string's identifier, or <code>null</code> to omit the identifier
	 * @param contents
	 *            the string's contents, or <code>null</code> to omit the contents.
	 * 
	 * @return the XML code corresponding to the string definition
	 */
	protected String createInlineStringDefinition( String identifier , String contents )
	{
		StringBuilder str = new StringBuilder( );
		str.append( "<inline-string" );
		if ( identifier != null ) {
			str.append( " id=" ).append( this.quoteString( identifier ) );
		}
		str.append( '>' );
		if ( contents != null ) {
			str.append( "<value>" ).append( this.xmlString( contents ) ).append( "</value>" );
		}
		str.append( "</inline-string>" );
		return str.toString( );
	}


	/**
	 * Create an external string definition
	 * 
	 * <p>
	 * Generate the XML code that corresponds to an external string definition.
	 * 
	 * @param identifier
	 *            the string's identifier, or <code>null</code> to omit the identifier
	 * @param source
	 *            the string's source file, or <code>null</code> to omit the source.
	 * 
	 * @return the XML code corresponding to the string definition
	 */
	protected String createExternalStringDefinition( String identifier , String source )
	{
		StringBuilder str = new StringBuilder( );
		str.append( "<from-file" );
		if ( identifier != null ) {
			str.append( " id=" ).append( this.quoteString( identifier ) );
		}
		if ( source != null ) {
			str.append( " source=" ).append( this.quoteString( source ) );
		}
		str.append( " />" );
		return str.toString( );
	}


	/**
	 * Create a language definition
	 * 
	 * @param identifier
	 *            the language's identifier, or <code>null</code> if the identifier is to be omitted
	 * @param name
	 *            the language's name, or <code>null</code> if the name is to be omitted
	 * @param stringDefinitions
	 *            XML definitions returned by {@link #createInlineStringDefinition(String, String)}
	 *            or {@link #createExternalStringDefinition(String, String)}
	 * 
	 * @return the XML code corresponding to the language definition
	 */
	protected String createLanguageDefinition( String identifier , String name , String... stringDefinitions )
	{
		StringBuilder str = new StringBuilder( );

		str.append( "<language" );
		if ( identifier != null ) {
			str.append( " id=" ).append( this.quoteString( identifier ) );
		}
		if ( name != null ) {
			str.append( " name=" ).append( this.quoteString( name ) );
		}
		str.append( ">" );
		for ( String definition : stringDefinitions ) {
			str.append( definition );
		}
		str.append( "</language>" );

		return str.toString( );
	}


	/**
	 * Create the XML code for a top-level I18N definition element
	 * 
	 * @param languages
	 *            XML definitions of languages, as returned by
	 *            {@link #createLanguageDefinition(String, String, String...)}
	 * 
	 * @return the top-level element's XML code
	 */
	protected String createTopLevel( String... languages )
	{
		StringBuilder str = new StringBuilder( );
		str.append( "<lw-text-data>" );
		for ( String language : languages ) {
			str.append( language );
		}
		str.append( "</lw-text-data>" );
		return str.toString( );
	}


	/**
	 * Create the necessary XStream instance
	 * 
	 * <p>
	 * Initialise an XStream instance and set it up so it can process I18N data definitions. Unlike
	 * the XStream instance generated by {@link I18NLoader}, this version also registers the alias
	 * for language definitions.
	 * 
	 * @return the initialised XStream instance.
	 */
	private XStream createXStreamInstance( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( I18NText.class );
		xstream.processAnnotations( InlineString.class );
		xstream.processAnnotations( FileString.class );
		xstream.alias( "language" , LanguageDefinition.class );
		return xstream;
	}


	/**
	 * Create an I18N definition object from its XML code
	 * 
	 * @param xml
	 *            the XML code
	 * @param cls
	 *            the class of the object
	 * 
	 * @return the object that was created from the code
	 * 
	 * @throws ClassCastException
	 *             if the code corresponds to some other type of object
	 * @throws XStreamException
	 *             if some error occurred while unserialising the object
	 */
	protected < T extends ImportableData > T createObject( String xml , Class< T > cls )
			throws ClassCastException , XStreamException
	{
		return cls.cast( this.createXStreamInstance( ).fromXML( xml ) );
	}
}
