package com.deepclone.lw.cli.xmlimport.data.i18n;


import static org.junit.Assert.*;

import java.io.File;

import org.junit.Test;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;



/**
 * Unit tests for the {@link FileString} class
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class TestExternalString
		extends BaseTest
{

	/**
	 * Test loading a valid external string record and the corresponding contents
	 * 
	 * @throws DataImportException
	 *             if a failure occurs while reading the source file.
	 */
	@Test
	public void testLoadStringOK( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "TestFiles/i18n-external-file/file" ) );
		assertEquals( "test" , instance.getId( ) );
		assertEquals( "This is a test." , instance.getString( ).trim( ) );
	}


	/**
	 * Test loading an external string record with no identifier
	 * 
	 * @throws DataImportException
	 *             if a failure occurs while reading the source file.
	 */
	@Test
	public void testLoadStringNullIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( null , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "TestFiles/i18n-external-file/file" ) );
		assertNull( instance.getId( ) );
		assertEquals( "This is a test." , instance.getString( ).trim( ) );
	}


	/**
	 * Test loading an external string record with no source file
	 * 
	 * @throws DataImportException
	 *             if a failure occurs while reading the source file
	 * @throws NullPointerException
	 *             when {@link FileString#getString()} is called while its source is
	 *             <code>null</code>
	 */
	@Test( expected = NullPointerException.class )
	public void testLoadStringNullSource( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , null );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "TestFiles/i18n-external-file/file" ) );
		assertEquals( "test" , instance.getId( ) );
		instance.getString( );
	}


	/**
	 * Test loading a valid external string record and obtaining the contents without setting the
	 * path to the data file first.
	 * 
	 * @throws DataImportException
	 *             if a failure occurs while reading the source file.
	 * @throws NullPointerException
	 *             when {@link FileString#getString()} is called while its original path is
	 *             <code>null</code>
	 */
	@Test( expected = NullPointerException.class )
	public void testLoadStringNoPath( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.getString( );
	}


	/**
	 * Test loading a valid external string record and obtaining the contents when the source file
	 * does not exist.
	 * 
	 * @throws DataImportException
	 *             when a failure occurs while reading the source file.
	 */
	@Test( expected = DataImportException.class )
	public void testLoadStringBadSource( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , "does-not-exist.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "TestFiles/i18n-external-file/file" ) );
		instance.getString( );
	}


	/**
	 * Test loading a valid external string record and obtaining the contents when the source file
	 * is a directory.
	 * 
	 * @throws DataImportException
	 *             when a failure occurs while reading the source file.
	 */
	@Test( expected = DataImportException.class )
	public void testLoadStringSourceIsDirectory( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , "." );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "TestFiles/i18n-external-file/file" ) );
		instance.getString( );
	}


	/**
	 * Test loading a valid external string record and obtaining the contents when the original
	 * file's path is incorrect.
	 * 
	 * @throws DataImportException
	 *             when a failure occurs while reading the source file.
	 */
	@Test( expected = DataImportException.class )
	public void testLoadStringBadParent( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "does-not-exist/does-not-exist" ) );
		instance.getString( );
	}


	/**
	 * Test loading a valid external string record and obtaining the contents when the original
	 * file's path is a single file name.
	 * 
	 * @throws DataImportException
	 *             if a failure occurs while reading the source file.
	 */
	@Test
	public void testLoadStringParentIsCurrentDirectory( )
			throws DataImportException , NullPointerException
	{
		String definition = this.createExternalStringDefinition( "test" , "TestFiles/i18n-external-file/test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.setReadFrom( new File( "does-not-exist" ) );
		instance.getString( );
	}



	/** Test validating a file string record with a short identifier and a source */
	@Test
	public void testValidateSimple( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/** Test validating an external string record with a camel case identifier and a source */
	@Test
	public void testValidateCamelCaseIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "testCamelCase" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/**
	 * Test validating an external string record with an identifier that includes a sequence of
	 * upper-case letters and a source
	 */
	@Test
	public void testValidateCapsIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "testCAPSIdentifier" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/**
	 * Test validating an external string record with an identifier that includes numbers and some
	 * contents
	 */
	@Test
	public void testValidateNumbersIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test123Numbers123" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/**
	 * Test rejecting an external string record with a <code>null</code> identifier and a source
	 */
	@Test( expected = DataImportException.class )
	public void testValidateNullIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( null , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/** Test rejecting an external string record with an identifier that starts with a bad character */
	@Test( expected = DataImportException.class )
	public void testValidateBadFirstCharacterIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( " test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/**
	 * Test rejecting an external string record with an identifier that starts with an upper-case
	 * letter
	 */
	@Test( expected = DataImportException.class )
	public void testValidateFirstCharacterWrongCaseIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "Test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/** Test rejecting an external string record with an identifier that includes "bad" characters */
	@Test( expected = DataImportException.class )
	public void testValidateBadMiddleCharacterIdentifier( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test Test" , "test.txt" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/** Test rejecting an external record with <code>null</code> contents */
	@Test( expected = DataImportException.class )
	public void testValidateNullSource( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test" , null );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/** Test rejecting an external record with empty contents */
	@Test( expected = DataImportException.class )
	public void testValidateEmptySource( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test" , "" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}


	/**
	 * Test rejecting an external record with contents the include only white space (spaces, tabs,
	 * new line, line feed)
	 */
	@Test( expected = DataImportException.class )
	public void testValidateSpacesOnlySource( )
			throws DataImportException
	{
		String definition = this.createExternalStringDefinition( "test" , " \t\r\n" );
		FileString instance = this.createObject( definition , FileString.class );
		instance.verifyData( );
	}

}
