package com.deepclone.lw.cli.xmlimport.data.i18n;


import static org.junit.Assert.*;

import org.junit.Before;
import org.junit.Test;

import com.deepclone.lw.cli.xmlimport.data.DataImportException;



/**
 * Unit tests for the {@link I18NText} class
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 */
public class TestI18NText
		extends BaseTest
{
	/** The language definition that will be used in most tests */
	private String languageDefinition;


	/** Initialise a language definition to use while testing */
	@Before
	public void setUp( )
	{
		String xml = this.createInlineStringDefinition( "test" , "test" );
		this.languageDefinition = this.createLanguageDefinition( "test" , "test" , xml );
	}


	/**
	 * Try loading a valid top-level I18N definition and make sure its contents are correct
	 * 
	 * @throws DataImportException
	 *             if the language definition check fails for some unknown reason
	 */
	@Test
	public void testLoadValidDefinitions( )
			throws DataImportException
	{
		String xml = this.createTopLevel( this.languageDefinition );
		I18NText text = this.createObject( xml , I18NText.class );

		int count = 0;
		for ( LanguageDefinition definition : text ) {
			definition.verifyData( );
			count++;
		}
		assertEquals( 1 , count );
	}


	/**
	 * Try loading an empty top-level element
	 * 
	 * @throws NullPointerException
	 *             when the iterator is accessed
	 */
	@Test( expected = NullPointerException.class )
	public void testLoadEmpty( )
			throws NullPointerException
	{
		String xml = this.createTopLevel( );
		I18NText text = this.createObject( xml , I18NText.class );
		text.iterator( );
	}


	/**
	 * Try validating a correct top-level element
	 * 
	 * @throws DataImportException
	 *             if the check fails
	 */
	@Test
	public void testValidateOK( )
			throws DataImportException
	{
		String xml = this.createTopLevel( this.languageDefinition );
		I18NText text = this.createObject( xml , I18NText.class );
		text.verifyData( );
	}


	/**
	 * Try rejecting a top-level element that does not contain any language definition
	 * 
	 * @throws DataImportException
	 *             when the check fails
	 */
	@Test( expected = DataImportException.class )
	public void testValidateEmpty( )
			throws DataImportException
	{
		String xml = this.createTopLevel( );
		I18NText text = this.createObject( xml , I18NText.class );
		text.verifyData( );
	}


	/**
	 * Try rejecting a top-level element that contains an invalid language definition
	 * 
	 * @throws DataImportException
	 *             when the check fails
	 */
	@Test( expected = DataImportException.class )
	public void testValidateBadDefinition( )
			throws DataImportException
	{
		String xml = this.createTopLevel( this.createLanguageDefinition( null , null ) );
		I18NText text = this.createObject( xml , I18NText.class );
		text.verifyData( );
	}


	/**
	 * Try rejecting a top-level element that contains two valid but duplicate language definitions
	 * 
	 * @throws DataImportException
	 *             when the check fails
	 */
	@Test( expected = DataImportException.class )
	public void testValidateDuplicateDefinition( )
			throws DataImportException
	{
		String xml = this.createTopLevel( this.languageDefinition , this.languageDefinition );
		I18NText text = this.createObject( xml , I18NText.class );
		text.verifyData( );
	}
}
