package com.deepclone.lw.cli.xmlimport.data.resources;


import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.deepclone.lw.cli.xmlimport.data.resources.BasicResource;
import com.deepclone.lw.cli.xmlimport.data.resources.NaturalResource;
import com.deepclone.lw.cli.xmlimport.data.resources.ResourceParameter;
import com.deepclone.lw.cli.xmlimport.data.resources.Resources;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.XStreamException;



/**
 * Base class for testing resource importation
 * 
 * <p>
 * This class is used as a parent for tests of the resource import structures. It includes the code
 * used to actually create these resources, as this can normally only be done through XStream.
 * 
 * @author <a href="mailto:tim@mitheren.com">T. Rosser</a>
 * 
 */

abstract class BaseTest
{
	/**
	 * Escape &amp;, &lt; and &gt; in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String xmlString( String string )
	{
		return string.replace( "&" , "&amp;" ).replace( "<" , "&lt;" ).replace( ">" , "&gt;" );
	}


	/**
	 * Escape &amp;, &lt; and &gt;, ' and " in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String quoteString( String string )
	{
		return "\"" + this.xmlString( string ).replace( "\"" , "&quot;" ).replace( "'" , "&apos;" ) + "\"";
	}


	/**
	 * Create a {@link BasicResource} definition.
	 * 
	 * <p>
	 * Generate the XML code that corresponds to a basic resource definition.
	 * 
	 * @param name
	 *            The resources name, or <code>null</code> to omit the name.
	 * @param description
	 *            The resources description, or <code>null</code> to omit the description.
	 * @param weight
	 *            The resources weight, or <code>null</code> to omit the weight.
	 * @param category
	 *            The resources category, or <code>null</code> to omit the category.
	 * 
	 * @return the XML code corresponding to the resource definition.
	 */
	protected String createBasicResourceDefinition( String name , String description , String weight , String category )
	{
		StringBuilder str = new StringBuilder( );

		str.append( "<basic-resource" );

		if ( name != null ) {
			str.append( " name=" ).append( this.quoteString( name ) );
		}
		if ( description != null ) {
			str.append( " description=" ).append( this.quoteString( description ) );
		}
		if ( weight != null ) {
			str.append( " weight=" ).append( this.quoteString( weight ) );
		}
		if ( category != null ) {
			str.append( " category=" ).append( this.quoteString( category ) );
		}

		str.append( " />" );

		return str.toString( );
	}


	/**
	 * Create the a {@link NaturalResource} definition.
	 * 
	 * <p>
	 * Create the XML code that corresponds to a natural resource definition.
	 * 
	 * @param name
	 *            The resources name, or <code>null</code> to omit the name.
	 * @param description
	 *            The resources description, or <code>null</code> to omit the description.
	 * @param weight
	 *            The resources weight, or <code>null</code> to omit the weight.
	 * @param category
	 *            The resources category, or <code>null</code> to omit the category.
	 * @param probability
	 *            The resources presence-probability, or <code>null</code> to omit the probability.
	 * @param quantity
	 *            The XML code representing the {@link ResourceParameter} giving the resources
	 *            quantity average and deviation or <code>null</code> to omit the quantity.
	 * @param difficulty
	 *            The XML code representing the {@link ResourceParameter} giving the resources
	 *            difficulty average and deviation or <code>null</code> to omit the difficulty.
	 * @param recovery
	 *            The XML code representing the {@link ResourceParameter} giving the resources
	 *            recovery average and deviation or <code>null</code> to omit the recovery.
	 * 
	 * @return the XML code corresponding to the natural resource definition
	 */
	protected String createNaturalResourceDefinition( String name , String description , String weight ,
			String category , String probability , String quantity , String difficulty , String recovery )
	{
		StringBuilder str = new StringBuilder( );

		str.append( "<natural-resource" );

		if ( name != null ) {
			str.append( " name=" ).append( this.quoteString( name ) );
		}
		if ( description != null ) {
			str.append( " description=" ).append( this.quoteString( description ) );
		}
		if ( weight != null ) {
			str.append( " weight=" ).append( this.quoteString( weight ) );
		}
		if ( category != null ) {
			str.append( " category=" ).append( this.quoteString( category ) );
		}
		if ( probability != null ) {
			str.append( " presence-probability=" ).append( this.quoteString( probability ) );
		}

		str.append( ">" );

		if ( quantity != null ) {
			str.append( quantity );
		}
		if ( difficulty != null ) {
			str.append( difficulty );
		}
		if ( recovery != null ) {
			str.append( recovery );
		}

		str.append( "</natural-resource>" );

		return str.toString( );
	}


	/**
	 * Create a {@link ResourceParameter} definition
	 * 
	 * <p>
	 * Generates the XML code the corresponds to a resource parameter definition.
	 * 
	 * @param type
	 *            Gives the type of resource parameter to create (quantity/difficulty/recovery)
	 * @param average
	 *            The parameters average.
	 * @param deviation
	 *            The parameters deviation.
	 * 
	 * @return the XML code corresponding to the resource parameter definition
	 */
	protected String createResourceParameterDefinition( String type , String average , String deviation )
	{
		StringBuilder str = new StringBuilder( );

		str.append( "<" );

		if ( type != null ) {
			str.append( type );
		}
		if ( average != null ) {
			str.append( " average=" ).append( this.quoteString( average ) );
		}
		if ( deviation != null ) {
			str.append( " deviation=" ).append( this.quoteString( deviation ) );
		}

		str.append( " />" );

		return str.toString( );
	}


	/**
	 * Creates the XML code for a top-level resource definition element
	 * 
	 * @param resources
	 *            XML definitions of resources
	 * 
	 * @return the top-level elements XML code
	 */
	protected String createTopLevel( String... resources )
	{
		StringBuilder str = new StringBuilder( );
		str.append( "<lw-resources>" );
		for ( String resource : resources ) {
			str.append( resource );
		}
		str.append( "</lw-resources>" );
		return str.toString( );
	}


	/**
	 * Create the necessary XStream instance
	 * 
	 * <p>
	 * Initialise an XStream instance and set it up so it can process resource data definitions.
	 * Unlike {@link ResourceLoader}, this version also registers multiple aliases for
	 * ResourceParameter. This is because ResourceParameter is usually only used inside a
	 * NaturalResource, but when testing may need to be created independently.
	 */
	private XStream createXStreamInstance( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( Resources.class );
		xstream.processAnnotations( NaturalResource.class );
		xstream.alias( "quantity" , ResourceParameter.class );
		xstream.alias( "difficulty" , ResourceParameter.class );
		xstream.alias( "recovery" , ResourceParameter.class );
		xstream.alias( "test" , ResourceParameter.class ); // Just for unit tests
		xstream.alias( "lw-resources" , Resources.class );

		return xstream;
	}


	/**
	 * Create a resource object from its XML code
	 * 
	 * @param xml
	 *            the XML code
	 * @param cls
	 *            the class of the object
	 * 
	 * @return the object that was created from the code
	 * 
	 * @throws ClassCastException
	 *             if the code corresponds to some other type of object
	 * @throws XStreamException
	 *             if some error occurred while deserialising the object
	 */
	protected < T extends ImportableData > T createObject( String xml , Class< T > cls )
			throws ClassCastException , XStreamException
	{
		return cls.cast( this.createXStreamInstance( ).fromXML( xml ) );
	}
}
