package com.deepclone.lw.cli.xmlimport.data.techs;


import com.deepclone.lw.cli.xmlimport.data.ImportableData;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.XStreamException;



/**
 * Base class for testing technology importation
 * 
 * <p>
 * This class is used as a parent for tests of the technology import structures. It includes the
 * code used to actually create these technology definitions, as this can normally only be done
 * through XStream.
 * 
 * @author <a href="mailto:tseeker@legacyworlds.com">E. Benoît</a>
 * 
 */

abstract class BaseTest
{
	/**
	 * Escape &amp;, &lt; and &gt; in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String xmlString( String string )
	{
		return string.replace( "&" , "&amp;" ).replace( "<" , "&lt;" ).replace( ">" , "&gt;" );
	}


	/**
	 * Escape &amp;, &lt; and &gt;, ' and " in XML strings
	 * 
	 * @param string
	 *            the string to escape
	 * 
	 * @return the escaped string
	 */
	private String quoteString( String string )
	{
		return "\"" + this.xmlString( string ).replace( "\"" , "&quot;" ).replace( "'" , "&apos;" ) + "\"";
	}


	/**
	 * Create the XML code for a technology definition
	 * 
	 * <p>
	 * This method generates the XML code for a technology definition, which can then be imported
	 * using XStream into a {@link Technology} instance. It is capable of generating instances with
	 * various fields set to <code>null</code>.
	 * 
	 * @param name
	 *            identifier of the technology's name string
	 * @param category
	 *            identifier of the technology's category string
	 * @param discovery
	 *            identifier of the technology's discovery string
	 * @param description
	 *            identifier of the technology's description string
	 * @param cost
	 *            monetary cost of implementing the technology
	 * @param points
	 *            amount of research points required
	 * @param dependencies
	 *            dependencies as an array of strings
	 * @return
	 */
	protected String createTechnology( String name , String category , String discovery , String description ,
			Long cost , Long points , String... dependencies )
	{
		StringBuilder builder = new StringBuilder( "<technology" );

		if ( name != null ) {
			builder.append( " name=" ).append( this.quoteString( name ) );
		}
		if ( category != null ) {
			builder.append( " category=" ).append( this.quoteString( category ) );
		}
		if ( discovery != null ) {
			builder.append( " discovery=" ).append( this.quoteString( discovery ) );
		}
		if ( description != null ) {
			builder.append( " description=" ).append( this.quoteString( description ) );
		}
		if ( cost != null ) {
			builder.append( " cost=\"" ).append( cost ).append( "\"" );
		}
		if ( points != null ) {
			builder.append( " points=\"" ).append( points ).append( "\"" );
		}

		if ( dependencies.length == 0 ) {
			builder.append( " />" );
		} else {
			builder.append( ">" );
			for ( String dep : dependencies ) {
				builder.append( "<depends-on>" );
				if ( dep != null ) {
					builder.append( this.xmlString( dep ) );
				}
				builder.append( "</depends-on>" );
			}
			builder.append( "</technology>" );
		}

		return builder.toString( );
	}


	/**
	 * Creates the XML code for a top-level technology definition element
	 * 
	 * @param resources
	 *            XML definitions of technologies
	 * 
	 * @return the top-level element's XML code
	 */
	protected String createTopLevel( String... technologies )
	{
		StringBuilder str = new StringBuilder( );
		str.append( "<lw-tech-graph>" );
		for ( String technology : technologies ) {
			str.append( technology );
		}
		str.append( "</lw-tech-graph>" );
		return str.toString( );
	}


	/**
	 * Create the necessary XStream instance
	 * 
	 * <p>
	 * Initialise an XStream instance and set it up so it can process technology data definitions.
	 * This includes annotation processing, of course, but also generating an alias allowing the
	 * &lt;technology&gt; tag to be recognised on its own.
	 */
	private XStream createXStreamInstance( )
	{
		XStream xstream = new XStream( );
		xstream.processAnnotations( Technologies.class );
		xstream.alias( "technology" , Technology.class );
		return xstream;
	}


	/**
	 * Create a resource object from its XML code
	 * 
	 * @param xml
	 *            the XML code
	 * @param cls
	 *            the class of the object
	 * 
	 * @return the object that was created from the code
	 * 
	 * @throws ClassCastException
	 *             if the code corresponds to some other type of object
	 * @throws XStreamException
	 *             if some error occurred while deserialising the object
	 */
	protected < T extends ImportableData > T createObject( String xml , Class< T > cls )
			throws ClassCastException , XStreamException
	{
		return cls.cast( this.createXStreamInstance( ).fromXML( xml ) );
	}
}
