package com.deepclone.lw.web.beans.session;


import java.net.InetAddress;

import org.apache.log4j.Logger;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.remoting.RemoteAccessException;

import com.deepclone.lw.session.Command;
import com.deepclone.lw.session.CommandResponse;
import com.deepclone.lw.session.SessionAccessor;
import com.deepclone.lw.session.SessionException;
import com.deepclone.lw.session.SessionReference;
import com.deepclone.lw.session.SessionResponse;
import com.deepclone.lw.utils.DigestHelper;



public class SessionClientBean
		implements SessionClient , ApplicationContextAware
{
	private Logger logger = Logger.getLogger( SessionClient.class );
	private String sessionService = "sessionSrv";
	private ApplicationContext appContext;


	public void setSessionService( String name )
	{
		this.sessionService = name;
	}


	@Override
	public void setApplicationContext( ApplicationContext applicationContext )
			throws BeansException
	{
		this.appContext = applicationContext;
	}


	private SessionAccessor getSessionService( )
			throws SessionServerException
	{
		try {
			return (SessionAccessor) this.appContext.getBean( this.sessionService );
		} catch ( BeanCreationException e ) {
			this.logger.error( "Could not connect to session manager" , e.getCause( ) );
			throw new SessionServerException( e.getCause( ) );
		}
	}


	private < T extends Session > T makeSession( ClientSessionReference sRef , Class< T > type )
	{
		T session;
		try {
			session = type.newInstance( );
		} catch ( Exception e ) {
			this.logger.error( "Session creation failed" , e );
			throw new RuntimeException( e );
		}
		session.initialise( this , sRef );
		return session;
	}


	@Override
	public < T extends Session > T getSession( ClientSessionReference session , Class< T > type , InetAddress ip )
			throws SessionException , SessionServerException
	{
		if ( !type.isAnnotationPresent( SessionType.class ) ) {
			throw new RuntimeException( "missing SessionType annotation on " + type );
		}

		SessionType t = type.getAnnotation( SessionType.class );
		SessionAccessor sAcc = this.getSessionService( );
		SessionReference sRef;
		try {
			sRef = sAcc.create( t.value( ) , "web" , ip );
		} catch ( RemoteAccessException e ) {
			this.logger.error( "Could not create session" , e );
			throw new SessionServerException( e );
		}

		session.setReference( sRef );
		return this.makeSession( session , type );
	}


	@Override
	public < T extends Session > T restoreSession( ClientSessionReference session , Class< T > type )
	{
		if ( session.isNull( ) ) {
			return null;
		}

		if ( !type.isAnnotationPresent( SessionType.class ) ) {
			throw new RuntimeException( "missing SessionType annotation on " + type );
		}

		SessionType t = type.getAnnotation( SessionType.class );
		if ( !t.value( ).equals( session.getReference( ).type ) ) {
			throw new RuntimeException( "session type mismatch" );
		}

		return this.makeSession( session , type );
	}


	@Override
	public void authenticate( ClientSessionReference session , String identifier , String secret )
			throws SessionException , SessionServerException
	{
		String challenge = session.getReference( ).extra;
		String sha1 = DigestHelper.digest( "sha-1" , challenge + " " + DigestHelper.digest( "sha-1" , secret ) );
		String md5 = DigestHelper.digest( "md5" , challenge + " " + DigestHelper.digest( "md5" , secret ) );

		SessionAccessor sAcc = this.getSessionService( );
		SessionReference ref;
		try {
			ref = sAcc.authenticate( session.getReference( ).identifier , identifier , sha1 , md5 );
		} catch ( RemoteAccessException e ) {
			this.logger.error( "Could not authenticate session" , e );
			throw new SessionServerException( e );
		}
		session.setReference( ref );
	}


	@Override
	public CommandResponse execute( ClientSessionReference reference , Command command )
			throws SessionServerException , SessionException
	{
		SessionAccessor sAcc = this.getSessionService( );
		SessionResponse sRep;
		try {
			sRep = sAcc.executeCommand( reference.getReference( ).identifier , command );
		} catch ( RemoteAccessException e ) {
			this.logger.error( "Could not execute command" , e );
			throw new SessionServerException( e );
		}
		reference.setReference( sRep.session );
		return sRep.data;
	}


	@Override
	public void terminate( ClientSessionReference reference )
			throws SessionServerException , SessionException
	{
		try {
			this.getSessionService( ).terminate( reference.getReference( ).identifier );
		} catch ( RemoteAccessException e ) {
			this.logger.error( "Could not terminate session" , e );
			throw new SessionServerException( e );
		} finally {
			reference.clear( );
		}
	}

}
